#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "aorom.h"
#include "memmap.h"

extern int memfd;

static int mirror;
static unsigned char *prgs;
static unsigned char nprgs;

extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

void InitAOROM(rominfo *ri)
{
    printf("Initializing AOROM mapper...\n");

    prgs = ri->prgrom;
    nprgs = ri->prg_pages >> 1;

    mmap((unsigned char *)0x8000, 0x8000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);

    vrom[0] = malloc(0x400);
    vrom[1] = malloc(0x400);
    vrom[2] = malloc(0x400);
    vrom[3] = malloc(0x400);
    vrom[4] = malloc(0x400);
    vrom[5] = malloc(0x400);
    vrom[6] = malloc(0x400);
    vrom[7] = malloc(0x400);    

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    MM_Mirror_One(nt1);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char AOROM_Read_Mapper(unsigned long addr)
{
    return 0;
}

void AOROM_Write_Mapper(unsigned long addr, unsigned char value)
{
    mmap((unsigned char *)0x8000, 0x8000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd,
	 (unsigned long)prgs + (value & (nprgs - 1))*0x8000);

    if (value & 0x10)
	MM_Mirror_One(nt2);
    else
	MM_Mirror_One(nt1);
}

unsigned char AOROM_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void AOROM_Write_CHR(unsigned long addr, unsigned char value)
{
    vrom[addr>>10][addr&0x3FF] = value;
}
