#include <stdio.h>
#include <unistd.h>
#include "cmdline.h"
#include "osssound.h"
#include "n6502.h"
#include "apu.h"

extern CONTEXTM6502 *n6502thingy;

#define SAMPLERATE         44100
#define BUFLEN             (SAMPLERATE / 60)
#define INPUTCLOCK         ((double)1789772.5)
#define INCVAL             ((double)INPUTCLOCK / (double)SAMPLERATE)
#define FREQ60HZ           ((double)INPUTCLOCK / (double)60)
#define FREQ120HZ           ((double)INPUTCLOCK / (double)120)
#define FREQ240HZ           ((double)INPUTCLOCK / (double)240)

int numevents = 0;

typedef struct
{
    unsigned long timestamp;
    unsigned short addr;
    unsigned char value;
} event;

#define PULSE1   1
#define PULSE2   2
#define TRI      4
#define NOISE    8
#define PCM     16

unsigned char countfunc[32] =
{
    5, 127, 10, 1, 20, 2, 40, 3, 80, 4, 30, 5, 7, 6, 13, 7,
    6, 8, 12, 9, 24, 10, 48, 11, 96, 12, 36, 13, 8, 14, 16, 15
};

double pcm_freq[16] = { 428, 380, 340, 320, 286, 254, 226, 214,
			190, 160, 142, 128, 106,  85,  72,  59 };

event events[100];
extern unsigned long framecount;

unsigned char enable_pulse1 = 0;
unsigned char enable_pulse2 = 0;
unsigned char enable_tri = 0;
unsigned char enable_noise = 0;
unsigned char enable_pcm = 0;

unsigned char mix_buf[735];

extern int dump_fd;

FILE *fp;

void APU_Init(void)
{
    int i;
    printf("BUFLEN: %d\n", BUFLEN);
    printf("INPUTCLOCK: %f\n", INPUTCLOCK);
    printf("INCVAL: %f\n", INCVAL);
    
    memset(mix_buf, 0, 735*2);
    fp = fopen("vrc7dump.txt", "wt");
    
    if (pcm_fd)
	write(pcm_fd, mix_buf, 735);
    if (dump_fd)
	write(dump_fd, mix_buf, 735);

}

unsigned char pulse1_counter  = 0;
unsigned char pulse2_counter  = 0;
unsigned char tri_counter  = 0;
unsigned char noise_counter = 0;

unsigned char Read_APU_Register(unsigned short addr)
{
    unsigned char ret = 0;
    
    if (addr == 0x4015)
    {
	if (pulse1_counter)
	    ret |= PULSE1;
	if (pulse2_counter)
	    ret |= PULSE2;
	if (tri_counter)
	    ret |= TRI;
	if (noise_counter)
	    ret |= NOISE;
    }
    return ret;
}

void Write_APU_Register(unsigned short addr, unsigned char value)
{
    events[numevents].timestamp = framecount + m6502zpGetElapsedTicks(0);
    events[numevents].addr = addr;
    events[numevents].value = value;
    numevents++;
}

int ptr = 0;

unsigned char apu_status = 0;
unsigned char apu_control = 0;

unsigned char nes_duty[] = { 14, 12, 8, 4};

/*
  Neill's noise findings

  freq
  
  0  - silent
  1  - 8h
  2  - 10h
  3  - 20h
  4  - 40h
  5  - 60h
  6  - 80h
  7  - A0h
  8  - 0CAh
  9  - 0FEh
  10 - 17Ch
  11 - 1FCh
  12 - 2FAh
  13 - 3F8h
  14 - 7F2h
  15 - 7FEh

*/

double noise_div[] = { 0, 8, 16, 32, 64, 96, 128, 160, 202, 254,
  380, 508, 762, 1016, 2034, 2046 };

unsigned char noise_buf[4];
unsigned char nb_pos = 0;

/*
double noise_div[] = { 93.2173177,
		       186.43463542,
		       372.86927,
		       736.22891814,
		       1505.2754415,
		       2251.286163,
		       2997.943886,
		       3752.1436,
		       4734.84788359,
		       5946.088,
		       8904.340796,
		       11472.9,
		       22655.3481,
		       33769.29245,
		       66287.87,
		       127840.89285 };
*/

double        counter_phaseacc = 0;
double        counter2_phaseacc = 0;
double        envelope1_phaseacc = 0;
double        envelope2_phaseacc = 0;
double        envelope3_phaseacc = 0;
double        sweep1_phaseacc = 0;
double        sweep2_phaseacc = 0;

unsigned char pulse1_regs[4]  = { 0, 0, 0, 0 };
double        pulse1_phaseacc = 0;
unsigned char pulse1_level    = 0;
unsigned char pulse1_count    = 0;
unsigned char pulse1_envcount = 0;
unsigned char pulse1_vol      = 0;
unsigned long pulse1_freq     = 0;
unsigned char pulse1_sweepcnt = 0;

unsigned char pulse2_regs[4]  = { 0, 0, 0, 0 };
double        pulse2_phaseacc = 0;
unsigned char pulse2_level    = 0;
unsigned char pulse2_count    = 0;
unsigned char pulse2_envcount = 0;
unsigned char pulse2_vol      = 0;
unsigned long pulse2_freq     = 0;
unsigned char pulse2_sweepcnt = 0;

unsigned char noise_regs[4] = { 0, 0, 0, 0 };
unsigned short noise_acc    = 0xFFFF;
double        noise_phaseacc = 0;
unsigned char noise_level = 0;
unsigned char noise_vol = 0;
unsigned char noise_pos = 0;

unsigned char tri_regs[4]  = { 0, 0, 0, 0 };
double        tri_phaseacc = 0;
unsigned char tri_level    = 0;
unsigned char tri_count    = 0;
unsigned char tri_counter2 = 0;

unsigned char pcm_regs[4]  = { 0, 0, 0, 0 };
double        pcm_phaseacc = 0;
unsigned char *pcm_base    = (unsigned char *)0xC000;
unsigned long pcm_pos      = 0;
unsigned long pcm_len      = 0;
unsigned long pcm_addr     = 0;
unsigned char pcm_level    = 0;
unsigned char pcm_mask     = 0;    

unsigned char vrc6_audio = 0;

unsigned char vrc6p1_regs[3] = { 0, 0, 0};
double        vrc6p1_phaseacc = 0;
unsigned char vrc6p1_level    = 0;
unsigned char vrc6p1_count    = 0;

unsigned char vrc6p2_regs[3] = { 0, 0, 0};
double        vrc6p2_phaseacc = 0;
unsigned char vrc6p2_level    = 0;
unsigned char vrc6p2_count    = 0;

unsigned char vrc6s_regs[3] = { 0, 0, 0};
double        vrc6s_phaseacc = 0;
unsigned char vrc6s_level    = 0;
unsigned char vrc6s_phase    = 0;

void update_pulse1(void)
{
    unsigned char b;
    double freq = (double)pulse1_regs[2] + (double)(pulse1_regs[3] & 7) * 256 + 1;

/*    if (pulse1_regs[1] & 0x80)
      freq = pulse1_sweep;*/
    
    pulse1_phaseacc += pulse1_freq + 1;
    if (pulse1_phaseacc < 0)
	pulse1_phaseacc = 0;
    
    pulse1_count++;
    pulse1_count &= 0xF;

    if ((pulse1_freq < 8) || (pulse1_freq > 2047))
	return;
    
    if (pulse1_count < nes_duty[pulse1_regs[0] >> 6])
	b = 0xFF;
    else
	b = 0;

    if (((pulse1_regs[0] & 0x20) || (pulse1_counter)))
    {
	if (pulse1_regs[0] & 0x10)
	    pulse1_level = b & (pulse1_regs[0] & 0xF);
	else
	    pulse1_level = b & (pulse1_vol & 0xF);
    }
}

void update_pulse2(void)
{
    unsigned char b;
    double freq = (double)pulse2_regs[2] + (double)(pulse2_regs[3] & 7) * 256 + 1;

/*    if (pulse2_regs[1] & 0x80)
      freq = pulse2_sweep;*/
    
    pulse2_phaseacc += pulse2_freq + 1;
    if (pulse2_phaseacc < 0)
	pulse2_phaseacc = 0;
    
    pulse2_count++;
    pulse2_count &= 0xF;

    if ((pulse2_freq < 8) || (pulse2_freq > 2047))
	return;
    
    if (pulse2_count < nes_duty[pulse2_regs[0] >> 6])
	b = 0xFF;
    else
	b = 0;

    if (((pulse2_regs[0] & 0x20) || (pulse2_counter)))
    {
	if (pulse2_regs[0] & 0x10)
	    pulse2_level = b & (pulse2_regs[0] & 0xF);
	else
	    pulse2_level = b & (pulse2_vol & 0xF);
    }
}

void update_tri(void)
{
    double freq = (double)tri_regs[2] + (double)(tri_regs[3] & 7) * 256 + 1;
    
    tri_phaseacc += freq;
    if (tri_phaseacc < 0)
	tri_phaseacc = 0;

    if (freq < 8)
	return;
    
    if (((tri_regs[0] & 0x80) || ((tri_counter) && (tri_counter2))))
    {
	tri_count++;
	tri_count &= 0x1F;

	if (tri_count & 0x10)
	    tri_level = (tri_count ^ 0x1F);
	else
	    tri_level = tri_count & 0xF;
    }
}

unsigned long noise_seed1 = 0x523ED313;
unsigned long noise_seed2 = 0xDEADBEEF;

/*

  for 8 and 10h, he uses a sliding window average of 4 noise
  samples and 2 noise samples respectively
  
  2 modes, bit 7 $400E

  if set, 5-bit LFSR, shifts right each time
  shifted in bit is bit 0 xor bit 2

  if clear, 9-bit LFSR
  bit shifted in is bit 0 xor bit 4

*/

void update_noise(void)
{
    unsigned long tmp;
    unsigned long tmp2;
    unsigned char bit0, bit2, bit4;
    
    int i;

    noise_phaseacc += noise_div[noise_regs[2] & 0xF] + 1;
	
    if (((noise_regs[0] & 0x20) || (noise_counter)))
    {
	if (noise_acc & 1)
	    noise_level = noise_regs[0] & 0xF;
	else
	    noise_level = 0;
    }
    
    //  printf("noise_acc before: $%x    ---     ", noise_acc);
	
    bit0 = noise_acc & 1;
    bit2 = (noise_acc >> 2) & 1;
    bit4 = (noise_acc >> 4) & 1;
    
    if (noise_regs[2] & 0x80)
    {
	noise_acc &= ~0x20;
	noise_acc >>= 1;
	if (bit0 ^ bit2)
	    noise_acc |= 0x10;
    }
    else
    {
	noise_acc &= ~0x200;
	noise_acc >>= 1;
	if (bit0 ^ bit4)
	    noise_acc |= 0x100;
    }
    
    if (noise_phaseacc < 0)
	noise_phaseacc = 0;
}

unsigned char old_level = 0;

void update_pcm(void)
{
    pcm_phaseacc += pcm_freq[pcm_regs[0] & 0xF];

    if (pcm_phaseacc < 0)
	pcm_phaseacc = 0;
    
    if (pcm_pos < pcm_len)
    {
	if (!pcm_mask)
	{
	    pcm_pos++;
	    pcm_mask = 0x01;
	}
	
	if (*(unsigned char *)(pcm_addr + pcm_pos) & pcm_mask)
	{
	    if (pcm_level < 0x7F)
		pcm_level++;
	}
	else
	{
	    if (pcm_level)
		pcm_level--;
	}
	pcm_mask <<= 1;
    }
    else
    {
	if (pcm_regs[0] & 0x40)
	{
	    pcm_pos = 0;
	}
/*	else if (pcm_regs[0] & 0x80)
	printf("SEND IRQ\n");*/
	pcm_mask = 0x01;
    }
}

void update_vrc6p1(void)
{
    double freq = (double)vrc6p1_regs[1] + (double)(vrc6p1_regs[2] & 0xF) * 256 + 1;
    unsigned char b;

    if (vrc6p1_regs[0] & 0x80)
    {
//	printf("Digital mode\n");
	vrc6p1_phaseacc += INCVAL;
    }
    else
	vrc6p1_phaseacc += freq;

    if (vrc6p1_phaseacc < 0)
	vrc6p1_phaseacc = 0;
    
    vrc6p1_count++;
    vrc6p1_count &= 0xF;
    
    if (vrc6p1_regs[2] & 0x80)
    {
	if (vrc6p1_count < ((vrc6p1_regs[0] >> 4) & 7) + 1)
	    b = 0xFF;
	else
	    b = 0;

	if (vrc6p1_regs[0] & 0x80)
	    vrc6p1_level = vrc6p1_regs[0] & 0xF;
	else
	    vrc6p1_level = b & (vrc6p1_regs[0] & 0xF);
//	vrc6p1_level <<= 1;	
    }
    else
	vrc6p1_level = 0;
}

void update_vrc6p2(void)
{
    double freq = (double)vrc6p2_regs[1] + (double)(vrc6p2_regs[2] & 0xF) * 256 + 1;
    unsigned char b;

    if (vrc6p2_regs[0] & 0x80)
    {
//	printf("Digital mode\n");
	vrc6p2_phaseacc += INCVAL;
    }
    else
	vrc6p2_phaseacc += freq;

    if (vrc6p2_phaseacc < 0)
	vrc6p2_phaseacc = 0;
    
    vrc6p2_count++;
    vrc6p2_count &= 0xF;

    if (vrc6p2_regs[2] & 0x80)
    {
	if (vrc6p2_count < ((vrc6p2_regs[0] >> 4) & 7) + 1)
	    b = 0xFF;
	else
	    b = 0;

	if (vrc6p2_regs[0] & 0x80)
	    vrc6p2_level = vrc6p2_regs[0] & 0xF;
	else
	    vrc6p2_level = b & (vrc6p2_regs[0] & 0xF);
//	vrc6p2_level <<= 1;	
    }
    else
	vrc6p2_level = 0;
}

void update_vrc6s(void)
{
    double freq = (double)vrc6s_regs[1] + (double)(vrc6s_regs[2] & 0xF) * 256 + 1;
    static unsigned char b = 0;

    vrc6s_phaseacc += freq * 2;

    vrc6s_phase += vrc6s_regs[0] & 0x7F;
    b++;
    
    if (b == 7)
    {
	b = 0;
	vrc6s_phase = 0;
    }
    
    if (vrc6s_regs[2] & 0x80)
    {
	vrc6s_level = (vrc6s_phase >> 3) & 0x1F;
    }
}

void update_envelope1(void)
{
//    envelope1_phaseacc += 131072;
    envelope1_phaseacc += ((pulse1_regs[0] & 0xF) + 1) * 8192;
    
    if (~pulse1_regs[0] & 0x10)
    {
	if (pulse1_vol)
	    pulse1_vol--;
	else if (pulse1_regs[0] & 0x20)
	{
//	    printf("Env 1 loop\n");
	    pulse1_vol = 0xF;
	}
    }
}

void update_envelope2(void)
{
//    envelope2_phaseacc += 111860.78125;
    envelope2_phaseacc += ((pulse2_regs[0] & 0xF) + 1) * 8192;
    
    if (~pulse2_regs[0] & 0x10)
    {
	if (pulse2_vol)
	    pulse2_vol--;
	else if (pulse2_regs[0] & 0x20)
	{
//	    printf("Env 2 loop\n");
	    pulse2_vol = 0xF;
	}
    }
}

void update_envelope3(void)
{
//    envelope2_phaseacc += 111860.78125;
    envelope3_phaseacc += ((noise_regs[0] & 0xF) + 1) * 8192;
    
    if (~noise_regs[0] & 0x10)
    {
	if (noise_vol)
	    noise_vol--;
	else if (noise_regs[0] & 0x20)
	{
//	    printf("Env 3 loop\n");
	    noise_vol = 0xF;
	}
    }
}

void update_sweep1(void)
{
    // speed 0 = 16384
    // speed 2 = 32768
    sweep1_phaseacc += 8192 + (((pulse1_regs[1] >> 4) & 0x7) + 1) * 8192;

    if ((pulse1_regs[1] & 0x80) && (pulse1_regs[1] & 0x7))
    {
	if (pulse1_regs[1] & 0x08)
	{
	    pulse1_freq -= pulse1_freq >> (pulse1_regs[1] & 7);
	}
	else
	{
	    pulse1_freq += pulse1_freq >> (pulse1_regs[1] & 7);
	}
	if ((pulse1_freq > 2047) || (pulse1_freq < 8))
	    pulse1_regs[1] &= ~0x80;
    }
}

void update_sweep2(void)
{
//    sweep2_phaseacc += 65536;
    sweep2_phaseacc += 8192 + (((pulse2_regs[1] >> 4) & 0x7) + 1) * 8192;
//    sweep2_phaseacc += 13982.59765625 * (((pulse2_regs[1] >> 4) & 7) + 1);

    if ((pulse2_regs[1] & 0x80) && (pulse2_regs[1] & 0x7))
    {
	if (pulse2_regs[1] & 0x08)
	{
	    pulse2_freq -= pulse2_freq >> (pulse2_regs[1] & 7);
	}
	else
	{
	    pulse2_freq += pulse2_freq >> (pulse2_regs[1] & 7);
	}
	if ((pulse2_freq > 2047) || (pulse2_freq < 8))
	    pulse2_regs[1] &= ~0x80;	
    }
}

void update_counter(void)
{
    counter_phaseacc += FREQ60HZ;

    if (pulse1_counter)
	pulse1_counter--;

    if (pulse2_counter)
	pulse2_counter--;

    if (tri_counter)
	tri_counter--;    

    if (noise_counter)
	noise_counter--;
}

void update_counter2(void)
{
    counter2_phaseacc += FREQ240HZ;

    if (tri_counter2)
	tri_counter2--;
}

void Dump_sweep(int i, unsigned char v)
{
    printf("Sweep %d: %s speed %d %s shift %d\n", i,
	   (v & 0x80) ? "enabled" : "disabled",
	   (v >> 4) & 7, (v & 0x08) ? "up" : "down", v & 7);
}

void APU_Update(void)
{
    event *ev = events;
    double cycle = INCVAL;
    unsigned short tmp = 0;
    ptr = 0;

    fprintf(fp, "frame\n");
    
    events[numevents].timestamp = 0xFFFFFFFF;

    while (ptr < 735)
    {
	counter_phaseacc -= INCVAL;
	if (counter_phaseacc < 0)
	    update_counter();
	
	counter2_phaseacc -= INCVAL;
	if (counter2_phaseacc < 0)
	    update_counter2();

	envelope1_phaseacc -= INCVAL;
	if (envelope1_phaseacc < 0)
	    update_envelope1();

	envelope2_phaseacc -= INCVAL;
	if (envelope2_phaseacc < 0)
	    update_envelope2();

	envelope3_phaseacc -= INCVAL;
	if (envelope3_phaseacc < 0)
	    update_envelope3();	
	
	sweep1_phaseacc -= INCVAL;
	if (sweep1_phaseacc < 0)
	    update_sweep1();

	sweep2_phaseacc -= INCVAL;
	if (sweep2_phaseacc < 0)
	    update_sweep2();	
	
	pulse1_phaseacc -= INCVAL;
	if (pulse1_phaseacc < 0)
	    update_pulse1();

	pulse2_phaseacc -= INCVAL;
	if (pulse2_phaseacc < 0)
	    update_pulse2();

	tri_phaseacc -= INCVAL;
	if (tri_phaseacc < 0)
	    update_tri();

	pcm_phaseacc -= INCVAL;
	if (pcm_phaseacc < 0)
	    update_pcm();	

	noise_phaseacc -= INCVAL;
	if (noise_phaseacc < 0)
	    update_noise();

	if (vrc6_audio)
	{
	    vrc6p1_phaseacc -= INCVAL;
	    if (vrc6p1_phaseacc < 0)
		update_vrc6p1();
	    
	    vrc6p2_phaseacc -= INCVAL;
	    if (vrc6p2_phaseacc < 0)
		update_vrc6p2();
	    
	    vrc6s_phaseacc -= INCVAL;
	    if (vrc6s_phaseacc < 0)
		update_vrc6s();	
	}
	
	mix_buf[ptr] = 0;

	if (apu_control & PULSE1)
	{
	    tmp = pulse1_level;
	    mix_buf[ptr] += tmp;
	}

	if (apu_control & PULSE2)
	{	
	    tmp = pulse2_level;
	    mix_buf[ptr] += tmp;
	}

	if (apu_control & TRI)
	{	
	    tmp = tri_level;
	    mix_buf[ptr] += tmp;
	}
	
	tmp = vrc6p1_level;
	mix_buf[ptr] += tmp;

	tmp = vrc6p2_level;
	mix_buf[ptr] += tmp;

	tmp = vrc6s_level;
	mix_buf[ptr] += tmp;

	if (apu_control & NOISE)
	{	
	    tmp = noise_level;
	    mix_buf[ptr] += tmp;
	}
	
	tmp = pcm_level;
	mix_buf[ptr] += tmp;

	while ((ev->timestamp < 0xFFFFFFFF) && (ev->timestamp < cycle))
	{
	    switch (ev->addr)
	    {
	    case 0x4000:
		pulse1_regs[0] = ev->value;
		if (~ev->value & 0x10)
		{
//		    printf("Pulse 1 envelope enabled speed %d\n", ev->value & 0xF);
		    pulse1_vol = 0xF;
		}
		break;
	    case 0x4001:
		pulse1_regs[1] = ev->value;
		pulse1_freq    = (double)pulse1_regs[2] + (double)(pulse1_regs[3] & 7) * 256;
/*		if (ev->value & 0x80)
		Dump_sweep(1, ev->value);*/
		break;
	    case 0x4002:
		pulse1_regs[2] = ev->value;
		pulse1_freq    = (double)pulse1_regs[2] + (double)(pulse1_regs[3] & 7) * 256;
		break;
	    case 0x4003:
		pulse1_regs[3] = ev->value;
		pulse1_counter = countfunc[ev->value >> 3];
		pulse1_freq    = (double)pulse1_regs[2] + (double)(pulse1_regs[3] & 7) * 256;
		pulse1_vol     = 0xF;
		break;

	    case 0x4004:
		pulse2_regs[0] = ev->value;
		if (~ev->value & 0x10)
		{
//		    printf("Pulse 2 envelope enabled speed %d\n", ev->value & 0xF);
		    pulse2_vol = 0xF;
		}
		break;
	    case 0x4005:
		pulse2_regs[1] = ev->value;
		pulse2_freq    = (double)pulse2_regs[2] + (double)(pulse2_regs[3] & 7) * 256;
/*		if (ev->value & 0x80)		
		Dump_sweep(2, ev->value);*/
		break;
	    case 0x4006:
		pulse2_regs[2] = ev->value;
		pulse2_freq    = (double)pulse2_regs[2] + (double)(pulse2_regs[3] & 7) * 256;
		break;
	    case 0x4007:
		pulse2_regs[3] = ev->value;
		pulse2_counter = countfunc[ev->value >> 3];
		pulse2_freq    = (double)pulse2_regs[2] + (double)(pulse2_regs[3] & 7) * 256;
		break;
		
	    case 0x4008:
		tri_regs[0] = ev->value;
		break;
	    case 0x400A:
		tri_regs[2] = ev->value;
		break;
	    case 0x400B:
		tri_regs[3] = ev->value;
		tri_counter = countfunc[ev->value >> 3];
		tri_counter2 = tri_regs[0] & 0x7F;
		break;

	    case 0x400C:
		noise_regs[0] = ev->value;
/*		if (ev->value & 0xC0)
		printf("Write to unused bits in $400C\n");*/
//		printf("%x to $400C\n", ev->value);
		if (~ev->value & 0x10)
		{
//		    printf("Pulse 1 envelope enabled speed %d\n", ev->value & 0xF);
		    noise_vol = 0xF;
		}
		break;
	    case 0x400D:
		noise_regs[1] = ev->value;
/*		if (ev->value)
		printf("Write to unused bits in $400D\n");*/
//		printf("%x to $400D\n", ev->value);
		break;		
	    case 0x400E:
/*		if (ev->value & 0x70)
		printf("Write to unused bits in $400E\n");*/
		noise_regs[2] = ev->value;
		noise_vol = 0xF;
//		noise_phaseacc = noise_div[ev->value & 0xF];
//		printf("%x to $400E\n", ev->value);
		break;		
	    case 0x400F:
/*		if (ev->value & 0x07)
		printf("Write to unused bits in $400F\n");*/
		noise_regs[3] = ev->value;
		noise_counter = countfunc[ev->value >> 3];
//		printf("%x to $400F\n", ev->value);
		break;
		
	    case 0x4010:
		pcm_regs[0] = ev->value;
//		printf("%x to $4010\n", ev->value);
		break;
	    case 0x4011:
//		printf("%x to $4011\n", ev->value & 0x7F);
		pcm_regs[1] = ev->value & 0x7F;
		pcm_level = ev->value & 0x7F;
		break;
	    case 0x4012:
//		printf("%x to $4012\n", ev->value);
		pcm_regs[2] = ev->value;
		pcm_addr = 0xC000 + (ev->value << 6);
		break;
	    case 0x4013:
//		printf("%x to $4013\n", ev->value);
		pcm_len = (ev->value << 4) + 1;
		pcm_pos = 0;
		pcm_mask = 0x01;
		pcm_regs[3] = ev->value;
		break;
		
	    case 0x4015:
		apu_control = ev->value;

		break;

/** VRC6 **/
		
	    case 0x9000:
		vrc6p1_regs[0] = ev->value;
		break;
	    case 0x9001:
		vrc6p1_regs[1] = ev->value;
		break;
	    case 0x9002:
		vrc6p1_regs[2] = ev->value;
		break;

	    case 0xA000:
		vrc6p2_regs[0] = ev->value;
		break;
	    case 0xA001:
		vrc6p2_regs[1] = ev->value;
		break;
	    case 0xA002:
		vrc6p2_regs[2] = ev->value;
		break;

	    case 0xB000:
		vrc6s_regs[0] = ev->value;
		break;
	    case 0xB001:
		vrc6s_regs[1] = ev->value;
		break;
	    case 0xB002:
		vrc6s_regs[2] = ev->value;
		break;

/** VRC7 **/
	    case 0x9010:
	    case 0x9030:		
		fprintf(fp, "%4X: %2X\n", ev->addr, ev->value);
	    }
	    ev++;
	}

	cycle += INCVAL;
	
	ptr++;
    }

    if (pcm_fd)
	write(pcm_fd, mix_buf, 735);

    if (dump_fd)
	write(dump_fd, mix_buf, 735);
    
    numevents = 0;
}
