#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "cnrom.h"
#include "memmap.h"

extern int memfd;

static int mirror;
static unsigned char *chrs;
static unsigned char npages;

extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

void InitCNROM(rominfo *ri)
{
    printf("Initializing CNROM mapper...\n");

    mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)ri->prgrom);

    if (ri->prg_pages > 1)
	mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)ri->prgrom + 0x4000);
    else
	mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)ri->prgrom);

    chrs = ri->chrrom;
    vrom[0] = ri->chrrom;
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = vrom[0] + 0x800;
    vrom[3] = vrom[0] + 0xC00;
    vrom[4] = vrom[0] + 0x1000;
    vrom[5] = vrom[0] + 0x1400;
    vrom[6] = vrom[0] + 0x1800;
    vrom[7] = vrom[0] + 0x1C00;
    npages = ri->chr_pages;

    mirror = ri->f.f.mirror;

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (mirror)
    {
	nametable[0] = nt1;
	nametable[1] = nt2;
	nametable[2] = nt1;
	nametable[3] = nt2;
    }
    else
    {
	nametable[0] = nt1;
	nametable[1] = nt1;
	nametable[2] = nt2;
	nametable[3] = nt2; 
    }

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char CNROM_Read_Mapper(unsigned long addr)
{
    return 0;
}

void CNROM_Write_Mapper(unsigned long addr, unsigned char value)
{
#ifdef DEBUG
    if (value > npages)
    {
	printf("Bank switch to page %d / %d\n", value, npages);
    }
#endif
    if (addr >= 0x8000)
    {
	vrom[0] = chrs + (value%npages)*0x2000;
	vrom[1] = vrom[0] + 0x400;
        vrom[2] = vrom[0] + 0x800;
	vrom[3] = vrom[0] + 0xC00;
	vrom[4] = vrom[0] + 0x1000;
        vrom[5] = vrom[0] + 0x1400;
	vrom[6] = vrom[0] + 0x1800;
	vrom[7] = vrom[0] + 0x1C00;
    }

}

unsigned char CNROM_Read_CHR(unsigned long addr)
{
    return vrom[addr>>10][addr & 0x3FF];
}

void CNROM_Write_CHR(unsigned long addr, unsigned char value)
{
}
