#include <stdio.h>
#include <stdlib.h>
#include "rom.h"
#include "memmap.h"
#include "cpu.h"
#include "newppu.h"
#include "apu.h"
#include "cmdline.h"
#include "snss.h"

#include "xlibio.h"

#include "n6502.h"

CONTEXTM6502 *n6502thingy = NULL;
extern struct MemoryReadByte NES_Read[];
extern struct MemoryWriteByte NES_Write[];

extern int memfd;
extern int zerofd;
extern rominfo *ri;
unsigned long framecount;
short VCount;

void (*Renderline)(unsigned char line) = PPU_NewRenderline;

int quitemu = 0;

void InitCPU(void)
{
    if (old_renderline)
	Renderline = PPU_OldRenderline;

    printf("NES_Read: %p\n", NES_Read);
    printf("NES_Write: %p\n", NES_Write);
    
    n6502thingy = malloc(m6502zpGetContextSize());
    memset(n6502thingy, 0, m6502zpGetContextSize());
    n6502thingy->m6502Base = ram;
    n6502thingy->m6502MemoryRead = NES_Read;
    n6502thingy->m6502MemoryWrite = NES_Write;
    m6502zpSetContext(n6502thingy);
    m6502zpreset();
    
    VCount = 0;

}

unsigned long delay = 0;

void VRC6_Cycle(void);
int VRC6_Check(void);

extern unsigned long frames;

void MainLoop(void)
{
    int nmi, irq;
    unsigned long cycles = 0;

    framecount = 0;
    m6502zpexec(NES_CYCLES_PER_SCANLINE);
    printf("ran first scanline\n");
    while (!quitemu)
    {
	nmi = 0;
	irq = 0;
	cycles += NES_CYCLES_PER_SCANLINE;

	VRC6_Cycle();

	if ((VCount >= NES_FIRST_VISIBLE)&&(VCount <= NES_LAST_VISIBLE))
	{
	    Renderline(VCount);
	}

	if (VCount == NES_VBLANK_START)
	{
	    APU_Update();
	    framecount = 0;
	    PPU_Start_Vblank();
	    Process_IO(0);
	}
	if (VCount == NES_NMI)
	{
	    nmi = PPU_Check_NMI();
	}

	if (VCount % 3 != 0)
	    cycles++;
	
	if (VCount == NES_MAX_NTSC_COUNTER)
	{
/*		if (savestate)
		{
		SNSS_SaveState(savestate);
		savestate = 0;
		}
		if (loadstate)
		{
		SNSS_LoadState(loadstate);
		loadstate = 0;
		}*/
	    VCount = -1;
	    PPU_End_Vblank();
	}

	VCount++;
	
	if (cur_mapper->Check_IRQ)
	    irq = cur_mapper->Check_IRQ();
	else
	    irq = 0;

	if (irq)
	    m6502zpint(0);
	if (nmi)
	    m6502zpnmi();
	
	if (delay)
	{
	    if (delay >= cycles)
	    {
		delay -= cycles;
		framecount += cycles;
		cycles = 0;
	    }
	    else
	    {
		framecount += delay;
		m6502zpexec(cycles - delay);

		cycles -= delay;
		cycles -= m6502zpGetElapsedTicks(0);

		framecount += m6502zpGetElapsedTicks(1);		
		delay = 0;
	    }
	}
	else
	{
	    m6502zpexec(cycles);

	    cycles -= m6502zpGetElapsedTicks(0);
	    if (delay)
		cycles = 0;
	    framecount += m6502zpGetElapsedTicks(1);
	    framecount += delay;
	}
    }
}
