#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include "rom.h"
#include "fme7.h"
#include "memmap.h"

extern int memfd;

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char *sram;
static unsigned char nchrs;
static unsigned char nprgs;
static unsigned char prgmask;
static unsigned char chrmask;
static unsigned char command;
static unsigned char regs[4];
static unsigned char chrswitch[8];
static unsigned char irqen = 0;

extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

void InitFME7(rominfo *ri)
{
    unsigned char *p;
    
    printf("Initializing Sunsoft FME-7 mapper...\n");

    prgs = ri->prgrom;
    chrs = ri->chrrom;
    sram = ri->sram;

    nchrs = ri->chr_pages;
    nprgs = ri->prg_pages;    
    
    nchrs <<= 3;
    nprgs <<= 1;

    prgmask = nprgs - 1;
    chrmask = nchrs - 1;
    
    printf("%d 8K PRG pages\n%d 1K CHR pages\n", nprgs, nchrs);

    mirror = ri->f.f.mirror;

    p = mmap((unsigned char *)0x6000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$6000 MMAP FAIL\n");

    p = mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$8000 MMAP FAIL\n");

    p = mmap((unsigned char *)0xA000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$A000 MMAP FAIL\n");

    p = mmap((unsigned char *)0xC000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$C000 MMAP FAIL\n");

    p = mmap((unsigned char *)0xE000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1)*0x2000);
    if (p == (unsigned char *)-1)
	printf("$E000 MMAP FAIL\n");

    regs[0] = 0;
    regs[1] = 0;
    regs[2] = 0;
    regs[3] = 0;

    chrswitch[0] = 0;
    chrswitch[1] = 1;
    chrswitch[2] = 2;
    chrswitch[3] = 3;
    chrswitch[4] = 4;
    chrswitch[5] = 5;
    chrswitch[6] = 6;
    chrswitch[7] = 7;

    vrom[0] = chrs;
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = vrom[0] + 0x800;
    vrom[3] = vrom[0] + 0xC00;
    vrom[4] = vrom[0] + 0x1000;
    vrom[5] = vrom[0] + 0x1400;
    vrom[6] = vrom[0] + 0x1800;
    vrom[7] = vrom[0] + 0x1C00;

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (ri->f.f.mirror)
	MM_Mirror_Horiz(nt1, nt2);
    else
	MM_Mirror_Vert(nt1, nt2);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char FME7_Read_Mapper(unsigned long addr)
{
    return 0;
}

void Update_FME7(void)
{
    vrom[0] = chrs + ((chrswitch[0] & chrmask) * 0x400);
    vrom[1] = chrs + ((chrswitch[1] & chrmask) * 0x400);
    vrom[2] = chrs + ((chrswitch[2] & chrmask) * 0x400);
    vrom[3] = chrs + ((chrswitch[3] & chrmask) * 0x400);
    vrom[4] = chrs + ((chrswitch[4] & chrmask) * 0x400);
    vrom[5] = chrs + ((chrswitch[5] & chrmask) * 0x400);
    vrom[6] = chrs + ((chrswitch[6] & chrmask) * 0x400);
    vrom[7] = chrs + ((chrswitch[7] & chrmask) * 0x400);

    if (regs[0] & 0x40)
	mmap((unsigned char *)0x6000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)sram);
    else
	mmap((unsigned char *)0x6000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + ((regs[0] & prgmask)*0x2000));

    switch(mirror)
    {
    case 0: MM_Mirror_Vert(nt1, nt2); break;
    case 1: MM_Mirror_Horiz(nt1, nt2); break;
    case 2: MM_Mirror_One(nt1); break;
    case 3: MM_Mirror_One(nt2); break;
    }
    
    mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + ((regs[1] & prgmask)*0x2000));
    mmap((unsigned char *)0xA000, 0x2000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + ((regs[2] & prgmask)*0x2000));
    mmap((unsigned char *)0xC000, 0x2000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + ((regs[3] & prgmask)*0x2000));
}

void FME7_Write_Mapper(unsigned long addr, unsigned char value)
{
    if (addr == 0x8000)
	command = value;
    else if (addr == 0xA000)
	switch (command & 0xF)
	{
	case 0: chrswitch[0] = value; break;
	case 1: chrswitch[1] = value; break;
	case 2: chrswitch[2] = value; break;
	case 3: chrswitch[3] = value; break;
	case 4: chrswitch[4] = value; break;
	case 5: chrswitch[5] = value; break;
	case 6: chrswitch[6] = value; break;
	case 7: chrswitch[7] = value; break;
	case 8: regs[0] = value; break;
	case 9: regs[1] = value; break;
	case 10: regs[2] = value; break;
	case 11: regs[3] = value; break;
	case 12: mirror = value; break;
	case 13: irqen = value; break;
	case 14: printf("Counter low? write %x\n", value); break;
	case 15: printf("Counter high? write %x\n", value); break;
	    /* Xod says this is a straight cycle counter */
	}
    else
	printf("Unknown FME7 write %x\n", value);
    Update_FME7();
}

unsigned char FME7_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void FME7_Write_CHR(unsigned long addr, unsigned char value)
{
    return;
}
