#include <stdio.h>
#include <stdlib.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include "xlibgfx.h"
//#include "xlibio.h"

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

unsigned char *vrom;
unsigned char *nt1;
unsigned char *at1;
unsigned char *pal;
int quitemu = 0;

unsigned int timediff(struct timeval *a, struct timeval *b)
{
    unsigned int t;
    t = a->tv_sec*1000000 + a->tv_usec;
    t -= (b->tv_sec*1000000) + b->tv_usec;
    return t;
}

extern unsigned short palconv[68];

void BlitTile(unsigned char *tilebuf, unsigned short x, unsigned short y)
{
    unsigned short *lineptr;
    int i, j;
    
    for (i = 0; i < 8; i++)
    {
	lineptr = (unsigned short *)((unsigned long)framebuffer + (y+i)*bytes_per_line + x*2);
	for (j = 0; j < 8; j++)
	    lineptr[j] = palconv[tilebuf[i*8 + j]];
    }
}

void RenderTile(unsigned char tile, unsigned char attrib, unsigned char bank,
		unsigned char *buf)
{
    int i;
    unsigned char *p, *p2, *p3;
    unsigned char b, romline[2];
    unsigned short tilebase;
    unsigned char mask = 0x80;
    
    tilebase = (tile << 4) + (bank << 12);

    memset(buf, 0, 64);
    p = buf;
    p2 = vrom + tilebase;
    p3 = vrom + tilebase + 8;
    i = 8;
    while (i--)
    {
	romline[0] = *p2++;
	romline[1] = *p3++;
	mask = 0x80;

	while (mask)
	{
	    b = attrib;
	    if (romline[0] & mask)
		b |= 1;
	    if (romline[1] & mask)
		b |= 2;
	    
//	    b = 64 + attrib;
	    if (b)
		*p = pal[b];
	    p++;

	    mask >>= 1;
	}
    }
}

void RenderTileLine(unsigned char tile, unsigned char line, unsigned char att,
		    unsigned char bank, unsigned char *buf)
{
    unsigned short tilebase = (tile << 4) + (bank << 12) + line;
    unsigned char b;
    unsigned char romline[2];
    unsigned char mask = 0x80;
    
//    printf("Tilebase: %x\n",tilebase);
    
    romline[0] = vrom[tilebase];
    romline[1] = vrom[tilebase + 8];

    while (mask)
    {
	b = att;
	if (romline[0] & mask)
	    b |= 1;
	if (romline[1] & mask)
	    b |= 2;
	if (b)
	    *buf++ = pal[b];
	else
	    *buf++ = 64;
	
	mask >>= 1;
    }
}

void BlitLine(unsigned char *src, unsigned short *dst, unsigned char ofs)
{
    int i = ofs;
    
    while (i--)
    {
	*dst++ = palconv[*src++];
    }
}

int PPU_vscroll = 20;
int PPU_hscroll = 0;
unsigned char table = 0;

unsigned char *nametable[4];
unsigned char *attribtable[4];

void RenderLine(unsigned char line)
{
    unsigned char ofs;
    unsigned short *linebuf = (unsigned short *)framebuffer + (line << 8);
    unsigned char *at;
    unsigned char *nt;
    unsigned char tilebuf[8];
    unsigned char tbl = table;
    unsigned char tiley, tilex;
    unsigned char tcount;
    unsigned char tofs;
    unsigned char att;
    unsigned char tile;
    unsigned char attbase, attshift;
    unsigned short namebase;
    
    tilex = PPU_hscroll >> 3;
    tcount = 32;
    if ((tofs = (PPU_hscroll & 7)) != 0)
	tcount++;
    tiley = (line + PPU_vscroll) >> 3;
    ofs = (line + PPU_vscroll) & 0x7;
    if (tiley >= 30)
    {
	tbl ^= 2;
	tiley -= 30;
    }
    at = attribtable[tbl];
    nt = nametable[tbl];
    attbase = ((tiley << 1) & 0xF8);
    namebase = (tiley << 5) + tilex;
    
    while (tcount)
    {
	if (tilex & 0x20)
	{
	    tbl ^= 1;
	    at = attribtable[tbl];
	    nt = nametable[tbl];
	    tilex = 0;
	    namebase = tiley << 5;
	}
	memset(tilebuf, 0, 8);
	attshift = ((tiley << 1) & 0x4) | (tilex & 0x2);
	att = ((at[attbase + (tilex >> 2)] >> attshift) & 0x3) << 2;
	
	tile = nt[namebase++];

//	RenderTile(tile, att, 1, tilebuf);
	RenderTileLine(tile, ofs, att, 1, tilebuf);
	if ((tcount == 33) && (tofs))
	{
	    BlitLine(tilebuf + tofs, linebuf, 8 - tofs);
	    linebuf += 8 - tofs;
	    tofs = 0;
	}
	else if ((tcount == 1) && (tofs))
	{
	    BlitLine(tilebuf, linebuf, tofs);
	}
	else
	{
	    BlitLine(tilebuf, linebuf, 8);
	    linebuf += 8;
	}
	tilex++;
	tcount--;
    }

//    Expandline(line, linebuf);
}

void RenderFrame(void)
{
#if 1
    int i = 0;

    for (i = 0; i < 240; i++)
    {
	RenderLine(i);
    }
#endif
#if 0
    int i, j;
    int k = 0;
    unsigned char tilebuf[64];
    
    for (i = 0; i < 30; i++)
    {
	for (j = 0; j < 32; j++)
	{
	    unsigned char attindex = ((i << 1) & 0xF8) + (j >> 2);
	    unsigned char attshift = (((i << 1) & 0x4) | (j & 0x2));
	    unsigned char att = ((at[attindex] >> attshift) & 0x3) << 2;
	
	    RenderTile(nt[(i << 5) + j],
		       att,
		       1, tilebuf);
//	    RenderTile(k++, 1, 1, tilebuf);
	    BlitTile(tilebuf, j << 3, i << 3);
	}
    }
#endif
}

void fillbuf(void)
{
    unsigned short *buf = (unsigned short *)framebuffer;
    int i, j;
    for (j = 0; j < 224; j++)
	for (i = 0; i < 256; i++)
	{
	    buf[(j * bytes_per_line/2) + i] = palconv[i >> 2];
	}
    updateimage();
    XSync(xdisp, False);
}

int main(void)
{
    FILE *fp;
    int i;
    XEvent event;
    struct timeval old = {0};
    struct timeval now;
    struct timeval next1 = {0};
    unsigned int tottime;
    
    fp = fopen("test.pal", "rb");
    pal = malloc(16);
    fread(pal, 1, 16, fp);
    fclose(fp);

    fp = fopen("test.gfx", "rb");
    nt1 = malloc(0x400);
    at1 = &nt1[0x3C0];
    fread(nt1, 1, 0x400, fp);
    fclose(fp);

    fp = fopen("test.pat", "rb");
    vrom = malloc(0x2000);
    fread(vrom, 1, 0x2000, fp);
    fclose(fp);

    for (i = 0; i < 4; i++)
    {
	nametable[i] = nt1;
	attribtable[i] = at1;
    }
    
    init_XDisplay();

    printf("Testing pallete...\n");
    getchar();
    fillbuf();
    getchar();
    
    while (!quitemu)
    {
	tottime = 0;
	if (next1.tv_sec == 0)
	{
	    while (gettimeofday(&next1, NULL) < 0) ;
	}
	next1.tv_usec += 16667;
	printf("Drawing 500 frames...\n");
	for (i = 0; i < 500; i++)
	{
	    while (gettimeofday(&old, NULL) < 0) ;

	    if (timercmp(&next1, &now, >))
	    {
		do
		{
		    while (gettimeofday(&now, NULL) < 0) ;
		} while (timercmp(&next1, &now, >));
	    }

	    if (next1.tv_usec >= 1000000)
	    {
		next1.tv_sec += next1.tv_usec / 1000000;
		next1.tv_usec %= 1000000;
	    }
	    next1.tv_usec += 16667;
	    RenderFrame();
	    PPU_hscroll+=4;
//	    PPU_vscroll++;
	    if (PPU_hscroll >= 256)
		PPU_hscroll = 0;
	    if (PPU_vscroll >= 240)
		PPU_vscroll = 0;
	    while (gettimeofday(&now, NULL) < 0) ;
	    updateimage();
	    while (XPending(xdisp)) XNextEvent(xdisp, &event);
	    tottime += timediff(&now, &old);
	}
	printf("%f frames/sec\n", 500000000/(float)tottime);
	i = getchar();
	if (i == 'q')
	    quitemu = 1;
    }
    return 0;
}

