#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "gnrom.h"
#include "memmap.h"

extern int memfd;

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char nprgs;
static unsigned char nchrs;
static unsigned char prgmask;
static unsigned char chrmask;

extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

void InitGNROM(rominfo *ri)
{
    printf("Initializing GNROM mapper...\n");

    mmap((unsigned char *)0x8000, 0x8000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)ri->prgrom);

    prgs = ri->prgrom;
    chrs = ri->chrrom;

    nprgs = ri->prg_pages;
    nchrs = ri->chr_pages;

    chrmask = nchrs - 1;
    prgmask = nprgs - 1;
    
    vrom[0] = chrs;
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = vrom[0] + 0x800;
    vrom[3] = vrom[0] + 0xC00;
    vrom[4] = vrom[0] + 0x1000;
    vrom[5] = vrom[0] + 0x1400;
    vrom[6] = vrom[0] + 0x1800;
    vrom[7] = vrom[0] + 0x1C00;

    mirror = ri->f.f.mirror;

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (mirror)
	MM_Mirror_Horiz(nt1, nt2);
    else
	MM_Mirror_Vert(nt1, nt2);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char GNROM_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void GNROM_Write_CHR(unsigned long addr, unsigned char value)
{
}

unsigned char GNROM_Read_Mapper(unsigned long addr)
{
    return 0;
}

void GNROM_Write_Mapper(unsigned long addr, unsigned char value)
{
//    printf("GNROM write %x\n", value);
    
    vrom[0] = chrs + (value & chrmask)*0x2000;
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = vrom[0] + 0x800;
    vrom[3] = vrom[0] + 0xC00;
    vrom[4] = vrom[0] + 0x1000;
    vrom[5] = vrom[0] + 0x1400;
    vrom[6] = vrom[0] + 0x1800;
    vrom[7] = vrom[0] + 0x1C00;

    mmap((unsigned char *)0x8000, 0x8000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + ((value >> 4) & prgmask)*0x8000);
}
