#include <stdio.h>
#include <string.h>

#include <X11/X.h>
#include <X11/Xlib.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include "cmdline.h"
#include "xlibgfx.h"
#include "xlibio.h"
#include "rom.h"
#include "memmap.h"
#include "cpu.h"
#include "apu.h"
#include "newppu.h"
#include "osssound.h"
#include "nsf.h"

#define MAJ_VERSION 0
#define MIN_VERSION 1

rominfo *ri;

int     zerofd;
int     memfd;

struct timeval start = {0};
struct timeval end = {0};

extern unsigned short palconv[64];
extern FILE *trifp;

void fillbuf(void)
{
    unsigned short *buf = (unsigned short *)framebuffer;
    int i, j;
    for (j = 0; j < 224; j++)
	for (i = 0; i < 256; i++)
	{
	    buf[(j * bytes_per_line/2) + i] = palconv[i >> 2];
	}
    updateimage();
    XSync(xdisp, False);
}

//void RenderDebug(void);
int updatedebug(void);
int updatedebug2(void);

void Clobber_Memmap(void)
{
    int i;
    unsigned char *p = 0;

    memset(p, 0, 0x1000);
    if (memcmp(p, 0x1000, 0x1000))
	printf("Ram mirror busted\n");
    
/*    printf("Testing SRAM mirror...\n");
    p = 0x6000;
    i = p[0];
    p[0] = 0xFF;
    if (p[0] != ri->sram[0])
    printf("SRAM mirror apparantly busted\n");*/
}

extern unsigned long frames;
extern unsigned long usec;
extern unsigned long sec;
extern unsigned char asm_active;

int main(int argc, char **argv)
{
    char *romfile;
    FILE *fp;
    
    printf("StankNES v%d.%d\n\n", MAJ_VERSION, MIN_VERSION);

    if (argc < 2)
    {
	usage();
	return 0;
    }

    printf("Assembly procs are %sABLED\n", asm_active ? "EN" : "DIS");
    
    printf("Parsing command line...\n");
    
    romfile = parse_args(argc, argv);
    zerofd = open("/dev/zero", O_RDWR);
    memfd = open("/proc/self/mem", O_RDWR);
    
    printf("Initializing Graphics\n");
    init_XDisplay();
    
    printf("Initializing Sound\n");
    Init_OSS();

    APU_Init();
    
    Init_PPU();
    
    printf("Loading rom image %s...\n", romfile);
    
    ri = loadrom(romfile);

    InitMemmap(ri);

    printf("Initializing 6502...\n");
    
    InitCPU();

    if (ri->nsf)
    {
	printf("Setting up CPU for NSF playback\n");
	InitNSF_CPU(ri);
    }
    
    printf("Running pallete/tile diagnostics\n");

    Clobber_Memmap();

//    getchar();
    
    printf("Starting emulation...\n");

    while (gettimeofday(&start, NULL) < 0) ;

    Process_IO(0);
    
    if (ri->nsf)
    {
	printf("Playing NSF\n");
	NSF_MainLoop(ri);
    }
    else
    {
	MainLoop();
    }

    while (gettimeofday(&end, NULL) < 0) ;
    
    printf("Avg fps: %f\n", (double)frames / (double)(end.tv_sec - start.tv_sec));

    if (ri->f.f.sram)
    {
	int i;
	
	i = strlen(romfile);
	while (romfile[i] != '.')
	    i--;
	romfile[i+1] = 0;
	strcat(romfile, "srm");
	fp = fopen(romfile, "wb");
	if (!fp)
	    printf("Could not open sram save %s!\n", romfile);
	else
	{
	    printf("Saving SRAM area...\n");
	    fwrite(ri->sram, 0x2000, 1, fp);
	    fclose(fp);
	}
    }
    
    Kill_OSS();
    
    kill_XDisplay();
    
    return 0;
}
