#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "memmap.h"
#include "io.h"
#include "newppu.h"
#include "apu.h"
#include "direct.h"
#include "unrom.h"
#include "cnrom.h"
#include "aorom.h"
#include "gnrom.h"
#include "mmc1.h"
#include "mmc2.h"
#include "mmc3.h"
#include "rambo1.h"
#include "vrc6.h"
#include "vrc7.h"
#include "fme7.h"
#include "sun4.h"
#include "nsf.h"
#include "n6502.h"

extern int zerofd;
extern int memfd;

unsigned char *nametable[4];
mapper *cur_mapper;
unsigned char *ram;
unsigned char *sram;
unsigned char *palram;
unsigned char *palette;
unsigned char *sprpal;
static unsigned char *unusedarea;
unsigned char paldirty = 0;
unsigned char *vrom[8];

void (*Updatepal)(unsigned char *data);

void MM_Mirror_Vert(unsigned char *nt1, unsigned char *nt2)
{
    nametable[0] = nt1;
    nametable[1] = nt2;
    nametable[2] = nt1;
    nametable[3] = nt2;
}

void MM_Mirror_Horiz(unsigned char *nt1, unsigned char *nt2)
{
    nametable[0] = nt1;
    nametable[1] = nt1;
    nametable[2] = nt2;
    nametable[3] = nt2;
}

void MM_Mirror_One(unsigned char *nt)
{
    nametable[0] = nt;
    nametable[1] = nt;
    nametable[2] = nt;
    nametable[3] = nt;
}

unsigned char Dummy_Read(unsigned long addr)
{
    return 0;
}

void Ram_Write(unsigned long addr, unsigned char v)
{
    ram[addr] = v;
    ram[addr ^ 0x800] = v;
}

void Dummy_Write(unsigned long addr, unsigned char v)
{
}

mapper Direct = { InitDirect, Direct_Read_Mapper, Direct_Write_Mapper, Direct_Read_CHR,
			Direct_Write_CHR, NULL, NULL, NULL, NULL };

mapper MMC1 = { InitMMC1, MMC1_Read_Mapper, MMC1_Write_Mapper, MMC1_Read_CHR,
	      MMC1_Write_CHR, NULL, NULL, NULL, NULL };

mapper UNROM = { InitUNROM, UNROM_Read_Mapper, UNROM_Write_Mapper, UNROM_Read_CHR,
	       UNROM_Write_CHR, NULL, NULL, NULL, NULL };

mapper CNROM = { InitCNROM, CNROM_Read_Mapper, CNROM_Write_Mapper, CNROM_Read_CHR,
	       CNROM_Write_CHR, NULL, NULL, NULL, NULL };
mapper MMC3 = { InitMMC3, MMC3_Read_Mapper, MMC3_Write_Mapper, MMC3_Read_CHR,
	      MMC3_Write_CHR, NULL, MMC3_Scanline_Hook, MMC3_Check_IRQ,
	      MMC3_Vblank_Hook };

mapper AOROM = { InitAOROM, AOROM_Read_Mapper, AOROM_Write_Mapper, AOROM_Read_CHR,
	       AOROM_Write_CHR, NULL, NULL, NULL, NULL };

mapper MMC2 = { InitMMC2, MMC2_Read_Mapper, MMC2_Write_Mapper, MMC2_Read_CHR,
	      MMC2_Write_CHR, MMC2_CHR_Latch, NULL, NULL, NULL };

mapper RAMBO1 = { InitRAMBO1, RAMBO1_Read_Mapper, RAMBO1_Write_Mapper, RAMBO1_Read_CHR,
	      RAMBO1_Write_CHR, NULL, RAMBO1_Scanline_Hook, RAMBO1_Check_IRQ,
		  RAMBO1_Vblank_Hook };

mapper GNROM = { InitGNROM, GNROM_Read_Mapper, GNROM_Write_Mapper, GNROM_Read_CHR,
		 GNROM_Write_CHR, NULL, NULL, NULL, NULL };

mapper VRC6 = { InitVRC6, VRC6_Read_Mapper, VRC6_Write_Mapper, VRC6_Read_CHR,
	       VRC6_Write_CHR, NULL, NULL, VRC6_Check, NULL };

mapper VRC7 = { InitVRC7, VRC7_Read_Mapper, VRC7_Write_Mapper, VRC7_Read_CHR,
	       VRC7_Write_CHR, NULL, NULL, NULL, NULL };

mapper SUN4 = { InitSUN4, SUN4_Read_Mapper, SUN4_Write_Mapper, SUN4_Read_CHR,
		SUN4_Write_CHR, NULL, NULL, NULL, NULL };

mapper FME7 = { InitFME7, FME7_Read_Mapper, FME7_Write_Mapper, FME7_Read_CHR,
		FME7_Write_CHR, NULL, NULL, NULL, NULL };

mapper NSF = { InitNSF, NSF_Read_Mapper, NSF_Write_Mapper, NULL,
	       NULL, NULL, NULL, NULL, NULL };

mapper Unimplemented = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };

struct mapent
{
    unsigned char num;
    mapper *map;
};

struct mapent mappertable[] =
{
    {0, &Direct},
    {1, &MMC1},
    {2, &UNROM},
    {3, &CNROM},
    {4, &MMC3},
    {5, &Unimplemented},
    {7, &AOROM},
    {9, &MMC2},
    {24, &VRC6},
    {64, &RAMBO1},
    {66, &GNROM},
    {68, &SUN4},
    {69, &FME7},
    {140, &VRC7},
    {254, &NSF},
    {0xFF, &Unimplemented}
};

struct MemoryReadByte NES_Read[] =
{
    {0x2000, 0x4FFF, (void *)Read_Register, NULL},
    {0x5000, 0x5FFF, (void *)Dummy_Read, NULL},
    {(unsigned long)-1, (unsigned long)-1, NULL}
};

struct MemoryWriteByte NES_Write[] =
{
    {0x0000, 0x1FFF, (void *)Ram_Write, NULL},
    {0x2000, 0x4FFF, (void *)Write_Register, NULL},
    {0x5000, 0x5FFF, (void *)Dummy_Write, NULL},
    {0x8000, 0xFFFF, (void *)Dummy_Write, NULL},
    {(unsigned long) -1, (unsigned long) -1, NULL}
};

void InitMemmap(rominfo *ri)
{
    int i;
    
    i = 0;
    while (mappertable[i].num != 0xFF)
    {
	if (mappertable[i].num == ri->mapper)
	    break;
	i++;
    }

    cur_mapper = mappertable[i].map;

    if (cur_mapper->InitMapper == NULL)
    {
	printf("Unimplemented mapper\n");
	return;
    }

    mmap(0, 65536 + 4096, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_FIXED, zerofd, 0);
    
    ram = mmap((unsigned char *)0, 4096, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_FIXED,
	       zerofd, 0);
    mmap((unsigned char *)4096, 4096, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_FIXED,
	       memfd, 0);    
    sram = mmap((unsigned char *)0x6000, 8192, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_FIXED,
		memfd, (unsigned long)ri->sram);
    palram = malloc(0x20);
    palette = malloc(0x20);
    sprpal = palram + 0x10;
    unusedarea = malloc(0xF00);
    
    memset(ram, 0, 0x800);
    memset(palram, 0, 0x20);
    memset(unusedarea, 0, 0xF00);
    
    printf("Initializing memory mapper...\n");

    cur_mapper->InitMapper(ri);
    NES_Read[1].memoryCall = (void *)cur_mapper->Read_Mapper;
    NES_Write[2].memoryCall = (void *)cur_mapper->Write_Mapper;
    NES_Write[3].memoryCall = (void *)cur_mapper->Write_Mapper;
}

void Write_CPU(unsigned long addr, unsigned char value)
{
    if (addr < 0x2000)
    {
	ram[addr] = value;
	ram[addr ^ 0x800] = value;
    }
    else if (addr <= 0x4017)
	Write_Register(addr, value);
    else if (addr < 0x6000)
	cur_mapper->Write_Mapper(addr, value);
    else if (addr < 0x8000)
	ram[addr] = value;
    else
	cur_mapper->Write_Mapper(addr, value);
}

unsigned char Read_CPU(unsigned long addr)
{
    if ((addr >= 0x2000) && (addr < 0x5000))
	return Read_Register(addr);
    else if ((addr >= 0x5000) && (addr < 0x6000))
	return cur_mapper->Read_Mapper(addr);		      
    return ram[addr];
}

void Write_PPU(unsigned long addr, unsigned char value)
{
    addr &= 0x3FFF;

/*    if (cur_mapper->CHR_Latch)
      cur_mapper->CHR_Latch(addr);*/

    if ((addr >= 0x3000) && (addr < 0x3F00))
    {
	addr = 0x2000 + (addr & 0xEFF);
    }
    
    if (addr < 0x2000)
	cur_mapper->Write_CHR(addr, value);
    else if (addr < 0x2400)
	nametable[0][addr & 0x3FF] = value;
    else if (addr < 0x2800)
	nametable[1][addr & 0x3FF] = value;
    else if (addr < 0x2C00)
	nametable[2][addr & 0x3FF] = value;
    else if (addr < 0x3000)
	nametable[3][addr & 0x3FF] = value;
    else if (addr < 0x3F00)
    {
	printf("unused area write\n");
	unusedarea[addr & 0x7FF] = value;
    }
    else if (addr <= 0x3FFF)
    {
	paldirty = 1;

	switch (addr & 0x1F)
	{
	case 0x00:
	case 0x10:
	    palette[0] = value;
	    palette[16] = value;
	    palram[0] = value;
	    palram[4] = value;
	    palram[8] = value;
	    palram[12] = value;
	    palram[16] = value;
	    palram[20] = value;
	    palram[24] = value;
	    palram[28] = value;
	    break;
	case 0x04:
	case 0x14:
	    palette[4] = value;
	    palette[20] = value;
	    break;
	case 0x08:
	case 0x18:
	    palette[8] = value;
	    palette[24] = value;
	    break;
	case 0x0C:
	case 0x1C:
	    palette[12] = value;
	    palette[28] = value;
	    break;
	case 0x01:
	case 0x11:
	case 0x02:
	case 0x12:
	case 0x03:
	case 0x13:
	case 0x05:
	case 0x15:
	case 0x06:
	case 0x16:
	case 0x07:
	case 0x17:
	case 0x09:
	case 0x19:
	case 0x0A:
	case 0x1A:
	case 0x0B:
	case 0x1B:
	case 0x0D:
	case 0x1D:
	case 0x0E:
	case 0x1E:
	case 0x0F:
	case 0x1F:
	    palette[addr & 0x1F] = value;
	    palram[addr & 0x1F] = value;
	    break;
	}
//	Updatepal(palram);
    }
}

unsigned char Read_PPU(unsigned long addr)
{
    addr &= 0x3FFF;

    if (addr >= 0x3F20)
    {
	addr &= 0x1F;
	addr += 0x3F00;
    }
    if ((addr >= 0x3000) && (addr < 0x3F00))
    {
	addr = 0x2000 + (addr & 0xEFF);
    }
    if (addr < 0x2000)
	return cur_mapper->Read_CHR(addr);
    else if (addr < 0x2400)
	return nametable[0][addr & 0x3FF];
    else if (addr < 0x2800)
	return nametable[1][addr & 0x3FF];
    else if (addr < 0x2C00)
	return nametable[2][addr & 0x3FF];
    else if (addr < 0x3000)
	return nametable[3][addr & 0x3FF];
    else if (addr < 0x3F00)
    {
	printf("unused area write\n");
	return unusedarea[addr & 0x7FF];
    }
    else if (addr <= 0x3FFF)
    {
	printf("palette read\n");
	return palette[addr & 0x1F];
    }
    return 0;
}

unsigned char Read_Register(unsigned long addr)
{
    if (addr < 0x4000)
    {
	return Read_PPU_Register((addr & 7) + 0x2000);
    }
    else if (addr <= 0x4015)
	return Read_APU_Register(addr);
    else if ((addr == 0x4016)||(addr == 0x4017))
	return Read_IO_Register(addr);
    else
    {
//	printf("CPU Read from %lx\n", addr);
	return 0;
    }
}

void Write_Register(unsigned long addr, unsigned char value)
{
    if (addr < 0x4000)
	Write_PPU_Register((addr & 7) + 0x2000, value);
    else if (addr == 0x4014)
	Write_PPU_Register(addr, value);
    else if (addr < 0x4016)
	Write_APU_Register(addr, value);
    else if (addr <= 0x4017)
	Write_IO_Register(addr, value);
    else
	printf("CPU Write %x to %lx\n", value, addr);
}
