#include <stdio.h>
#include <stdlib.h>

#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include "rom.h"
#include "mmc1.h"
#include "memmap.h"

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char nprgs;
static unsigned char nchrs;
static unsigned char chrmask;

extern unsigned char *nametable[4];
//extern unsigned char *attribtable[4];
unsigned char *nt1;
unsigned char *nt2;

static unsigned char count = 0;
static unsigned char latch = 0;
static unsigned char regs[4];
static unsigned long lastaddr = 0;

extern int memfd;
extern int zerofd;

void Update_MMC1()
{
    if (regs[0] & 0x2)
	if (regs[0] & 0x1)
	{
	    MM_Mirror_Horiz(nt1, nt2);
	}
	else
	{
	    MM_Mirror_Vert(nt1, nt2);
	}
    else
    {
	if (regs[0] & 0x1)
	    MM_Mirror_One(nt1);
	else
	    MM_Mirror_One(nt2);
    }

    if (regs[0] & 0x08) /* 16K switch */
    {
	if (regs[0] & 0x04) /* 16K switch at 0x8000 */
	{
	    mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1) *0x4000);
	    mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd,
		 (unsigned long)prgs + (regs[3] & (nprgs - 1))*0x4000);
	}
	else /* 16K switch at 0xC000 */
	{
	    mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
	    mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd,
		 (unsigned long)prgs + (regs[3] & (nprgs - 1))*0x4000);
	}
    }
    else
    {
	mmap((unsigned char *)0x8000, 0x8000, PROT_READ, MAP_PRIVATE|MAP_FIXED,
	     memfd, (unsigned long)prgs + ((regs[3] & 0x1E) & (nprgs - 1))*0x4000);
    }

    if (nchrs)
    {

	if (regs[0] & 0x10)
	{
	    vrom[0] = chrs + ((regs[1]) & chrmask) * 0x1000;
	    vrom[1] = vrom[0] + 0x400;
	    vrom[2] = vrom[0] + 0x800;
	    vrom[3] = vrom[0] + 0xC00;	    
	    vrom[4] = chrs + ((regs[2]) & chrmask)*0x1000;
	    vrom[5] = vrom[4] + 0x400;
	    vrom[6] = vrom[4] + 0x800;
	    vrom[7] = vrom[4] + 0xC00;
	}
	else
	{
	    vrom[0] = chrs + (regs[1] & chrmask)*0x1000;
	    vrom[1] = vrom[0] + 0x400;
	    vrom[2] = vrom[0] + 0x800;
	    vrom[3] = vrom[0] + 0xC00;
	    vrom[4] = vrom[0] + 0x1000;
	    vrom[5] = vrom[0] + 0x1400;
	    vrom[6] = vrom[0] + 0x1800;
	    vrom[7] = vrom[0] + 0x1C00;
	}
    }
}

void InitMMC1(rominfo *ri)
{
    int i;
    unsigned char *p;
    
    printf("Initializing MMC1 mapper...\n");

    printf("%d 16K PRG pages\n%d 8K CHR pages\n",
	   ri->prg_pages, ri->chr_pages);
    
    prgs = ri->prgrom;
    chrs = ri->chrrom;

    chrmask = 1;

    nchrs = ri->chr_pages << 1;
    nprgs = ri->prg_pages;

    while (nchrs - 1> chrmask)
    {
	chrmask <<= 1;
	chrmask |= 1;
    }

    printf("Old CHR mask: %x\n", nchrs - 1);
    printf("CHR mask: %x\n", chrmask);
    
    mirror = ri->f.f.mirror;

    printf("Setting $8000 to page 0 (%p), $C000 to page %d (%p)\n", prgs,
	   nprgs - 1, prgs + (nprgs - 1)*0x4000);
    
    p = mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);

    if ((unsigned long)p == -1)
    {
	printf("0x8000 MMAP Fail\n");
	switch (errno)
	{
	case EBADF: printf("EBADF\n"); break;
	case EACCES: printf("EACCES\n"); break;
	case EINVAL: printf("EINVAL\n"); break;
	case EAGAIN: printf("EAGAIN\n"); break;
	case ENOMEM: printf("ENOMEM\n"); break;
	}
	perror("mmap");
    }
    p = mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1)*0x4000);
    if ((unsigned long)p == -1)
	printf("0xC000 MMAP Fail\n");
    getchar();

    if (nchrs)
    {
	vrom[0] = chrs;
	vrom[1] = vrom[0] + 0x400;
	vrom[2] = vrom[0] + 0x800;
	vrom[3] = vrom[0] + 0xC00;
	vrom[4] = chrs + 0x1000;
	vrom[5] = vrom[4] + 0x400;
	vrom[6] = vrom[4] + 0x800;
	vrom[7] = vrom[4] + 0xC00;
    }
    else
    {
	vrom[0] = malloc(0x400);
	vrom[1] = malloc(0x400);
	vrom[2] = malloc(0x400);
	vrom[3] = malloc(0x400);	
	vrom[4] = malloc(0x400);
    	vrom[5] = malloc(0x400);
	vrom[6] = malloc(0x400);
	vrom[7] = malloc(0x400);
    }
    
    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (ri->f.f.mirror)
	MM_Mirror_Vert(nt1, nt2);
    else
	MM_Mirror_Horiz(nt1, nt2);
    
    regs[0] = 0x1F;
    regs[1] = 0;
    regs[2] = 0;
    regs[3] = 0;

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char MMC1_Read_Mapper(unsigned long addr)
{
    return 0;
}

void MMC1_Write_Mapper(unsigned long addr, unsigned char value)
{
//    printf("MMC1 write: $%x to $%x\n", value, addr);
/*    if (!count)
	lastaddr = addr;
    else if (lastaddr != addr)
    {
//	printf("addr mismatch\n");
	count = 0;
	lastaddr = addr;
	}*/
    
    if (value & 0x80)
    {
	regs[0] |= 0x0c;
	latch = regs[(addr >> 13) & 3];
	count = 0;
//	printf("MMC1 Latch Reset\n");
    }
    else
    {
	latch |= (value & 1) << count;
	count++;
    }

    if (count == 5)
    {
/*	printf("Write of 0x%x to register %d\n",
	latch, (addr >> 13) & 3);*/
	regs[(addr >> 13) & 3] = latch;
	latch = 0;
	count = 0;
	Update_MMC1();
    }
}

unsigned char MMC1_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void MMC1_Write_CHR(unsigned long addr, unsigned char value)
{
    if (nchrs)
	return;
    vrom[addr >> 10][addr&0x3FF] = value;
//    printf("Write to VRAM\n");
}
