#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "mmc2.h"
#include "memmap.h"

static int mirror;
static int flip1 = 0;
static int flip2 = 0;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char nprgs;
static unsigned char nchrs;
static unsigned char *tiles;

extern unsigned char *nametable[4];
unsigned char *nt1;
unsigned char *nt2;

static unsigned char regs[4];
unsigned char latch[2];

extern int memfd;
extern int zerofd;

void Update_MMC2(void)
{
    if (latch[0] == 0xFD)
    {
	vrom[0] = chrs + regs[0]*0x1000;
	vrom[1] = vrom[0] + 0x400;
	vrom[2] = vrom[0] + 0x800;
	vrom[3] = vrom[0] + 0xC00;	
    }
    else if (latch[0] == 0xFE)
    {
	vrom[0] = chrs + regs[1]*0x1000;
	vrom[1] = vrom[0] + 0x400;
	vrom[2] = vrom[0] + 0x800;
	vrom[3] = vrom[0] + 0xC00;
    }

    if (latch[1] == 0xFD)
    {
	vrom[4] = chrs + regs[2]*0x1000;	
	vrom[5] = vrom[4] + 0x400;
	vrom[6] = vrom[4] + 0x800;
	vrom[7] = vrom[4] + 0xC00;
    }
    else if (latch[1] == 0xFE)
    {
	vrom[4] = chrs + regs[3]*0x1000;    
	vrom[5] = vrom[4] + 0x400;
	vrom[6] = vrom[4] + 0x800;
	vrom[7] = vrom[4] + 0xC00;
    }
}

void InitMMC2(rominfo *ri)
{
    int i;
    unsigned char *p;

    printf("Initializing MMC2 mapper...\n");

    prgs = ri->prgrom;

    nchrs = ri->chr_pages << 1;
    nprgs = ri->prg_pages << 1;
    
    printf("%d 8K PRG pages\n%d 4K CHR pages\n",
	   nprgs, nchrs);
    
    chrs = ri->chrrom;

    mirror = ri->f.f.mirror;

    mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    mmap((unsigned char *)0xA000, 0x6000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 3)*0x2000);
    
    vrom[0] = chrs;
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = vrom[0] + 0x800;
    vrom[3] = vrom[0] + 0xC00;    
    vrom[4] = chrs + 0x1000;
    vrom[5] = vrom[4] + 0x400;
    vrom[6] = vrom[4] + 0x800;
    vrom[7] = vrom[4] + 0xC00;    

    latch[0] = 0xFE;
    latch[1] = 0xFE;
    
    nt1 = malloc(0x400);
    nt2 = malloc(0x400);
    
    if (ri->f.f.mirror)
	MM_Mirror_Vert(nt1, nt2);
    else
	MM_Mirror_Horiz(nt1, nt2);

    regs[0] = 0;
    regs[1] = 0;
    regs[2] = 1;
    regs[3] = 1;

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);    
}

unsigned char MMC2_Read_Mapper(unsigned long addr)
{
    return 0;
}

void MMC2_Write_Mapper(unsigned long addr, unsigned char value)
{
    switch (addr & 0xF000)
    {
    case 0xA000:
	mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd,
	     (unsigned long)prgs + (value & (nprgs - 1))*0x2000);
	break;
    case 0xB000:
	regs[0] = value & (nchrs - 1);
	break;
    case 0xC000:
	regs[1] = value & (nchrs - 1);
	break;
    case 0xD000:
	regs[2] = value & (nchrs - 1);
	break;
    case 0xE000:
	regs[3] = value & (nchrs - 1);
	break;	
    case 0xF000:
	if (value & 1)
	    MM_Mirror_Horiz(nt1, nt2);
	else
	    MM_Mirror_Vert(nt1, nt2);
    }
    Update_MMC2();
}

unsigned char MMC2_Read_CHR(unsigned long addr)
{
    MMC2_CHR_Latch(addr >> 12, addr >> 4);
    
    return vrom[addr >> 10][addr&0x3FF];
}

void MMC2_Write_CHR(unsigned long addr, unsigned char v)
{
    MMC2_CHR_Latch(addr >> 12, addr >> 4);
}

void MMC2_CHR_Latch(unsigned char bank, unsigned long tile)
{
    if (bank)
	latch[1] = tile;
    else
	latch[0] = tile;
    
    Update_MMC2();
}
