#include <stdio.h>
#include <stdlib.h>

#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include "rom.h"
#include "mmc3.h"
#include "memmap.h"

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char nprgs;
static unsigned short nchrs;
static unsigned char *tiles;
static unsigned char command = 0;
static unsigned char xorval = 0;
static unsigned char prgswitch = 0;
static unsigned char regs[2];
static unsigned long hardwire[2];
static unsigned char sramen = 0;
static unsigned char irqcounter = 0;
static unsigned char irqlatch = 0;
static unsigned char irqen = 0;
static unsigned char irqfire = 0;
static unsigned char chrmask = 0;
static unsigned char prgmask = 0;
static unsigned char chrswitch[6];
static unsigned char *sram;
static unsigned char *ffpage;

extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

extern int memfd;
extern int zerofd;

unsigned char chrdirty = 0;
unsigned char prgdirty = 0;

void MMC3_Update_ROM(void)
{
    if (command & 0x40)
    {
	mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 2)*0x2000);
	mmap((unsigned char *)0xA000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + regs[1]*0x2000);
	mmap((unsigned char *)0xC000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + regs[0]*0x2000);
	mmap((unsigned char *)0xE000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1)*0x2000);	
    }
    else
    {
	mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + regs[0]*0x2000);
	mmap((unsigned char *)0xA000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + regs[1]*0x2000);
	mmap((unsigned char *)0xC000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 2)*0x2000);
	mmap((unsigned char *)0xE000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1)*0x2000);	
    }
    prgdirty = 0;
}

void MMC3_Update_CHR(void)
{
    if (!nchrs)
	return;

    if (command & 0x80)
    {
	vrom[4] = chrs + (chrswitch[0])*0x400;
	vrom[5] = chrs + (chrswitch[0] + 1)*0x400;
	vrom[6] = chrs + (chrswitch[1])*0x400;
	vrom[7] = chrs + (chrswitch[1] + 1)*0x400;
	vrom[0] = chrs + (chrswitch[2])*0x400;
	vrom[1] = chrs + (chrswitch[3])*0x400;
	vrom[2] = chrs + (chrswitch[4])*0x400;
	vrom[3] = chrs + (chrswitch[5])*0x400;
    }
    else
    {
	vrom[0] = chrs + (chrswitch[0]) * 0x400;
	vrom[1] = chrs + (chrswitch[0] + 1)*0x400;
	vrom[2] = chrs + (chrswitch[1])*0x400;
	vrom[3] = chrs + (chrswitch[1] + 1)*0x400;
	vrom[4] = chrs + (chrswitch[2])*0x400;
	vrom[5] = chrs + (chrswitch[3])*0x400;
	vrom[6] = chrs + (chrswitch[4])*0x400;
	vrom[7] = chrs + (chrswitch[5])*0x400;
    }
    chrdirty = 0;
}

void InitMMC3(rominfo *ri)
{
    int i;
    unsigned char *p;

    printf("Initializing MMC3 mapper...\n");

    prgs = ri->prgrom;
    chrs = ri->chrrom;
    sram = ri->sram;
    ffpage = mmap(0, 0x2000, PROT_READ|PROT_WRITE, MAP_PRIVATE, zerofd, 0);
    memset(ffpage, 0xFF, 0x2000);
    
    nchrs = ri->chr_pages << 3;
    nprgs = ri->prg_pages << 1;
    chrmask = 1;

    while (nchrs - 1 > chrmask)
    {
	chrmask <<= 1;
	chrmask |= 1;
    }
    prgmask = nprgs - 1;
    
    printf("%d 8K PRG pages\n%d 1K CHR pages\n",
	   nprgs, nchrs);

    mirror = ri->f.f.mirror;

    printf("Setting $8000 to page 0 (%p), $A000 to page 1 (%p), \n"
	   "$C000 to page %d (%p), $E000 to page %d (%p)\n", prgs, prgs + 0x2000,
	   nprgs - 2, prgs + (nprgs - 2)*0x2000, nprgs - 1, prgs + (nprgs - 1)*0x2000);

    hardwire[0] = (unsigned long)prgs + (nprgs - 2)*0x2000;
    hardwire[1] = (unsigned long)prgs + (nprgs - 1)*0x2000;
    
    p = mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$8000 MMAP FAIL\n");
    
    p = mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 2)*0x2000);
    if (p == (unsigned char *)-1)
	printf("$C000 MMAP FAIL\n");
    
    regs[0] = 0;
    regs[1] = 1;

    if (nchrs)
    {
	vrom[0] = chrs;
	vrom[1] = vrom[0] + 0x400;
	vrom[2] = vrom[0] + 0x800;
	vrom[3] = vrom[0] + 0xC00;
	vrom[4] = chrs + 0x1000;
	vrom[5] = vrom[4] + 0x400;
	vrom[6] = vrom[4] + 0x800;
	vrom[7] = vrom[4] + 0xC00;
    }
    else
    {
	vrom[0] = malloc(0x400);
	vrom[1] = malloc(0x400);
	vrom[2] = malloc(0x400);
	vrom[3] = malloc(0x400);
	vrom[4] = malloc(0x400);
	vrom[5] = malloc(0x400);
	vrom[6] = malloc(0x400);
	vrom[7] = malloc(0x400);
    }

    chrswitch[0] = 0;
    chrswitch[1] = 2;
    chrswitch[2] = 4;
    chrswitch[3] = 5;
    chrswitch[4] = 6;
    chrswitch[5] = 7;
    
    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (ri->f.f.mirror)
	MM_Mirror_Vert(nt1, nt2);
    else
	MM_Mirror_Horiz(nt1, nt2);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char MMC3_Read_Mapper(unsigned long addr)
{
    return 0;
}

void MMC3_Write_Mapper(unsigned long addr, unsigned char value)
{
    if (addr == 0x8000)
	command = value;
    else if (addr == 0x8001)
	switch (command & 7)
	{
	case 0:
	    value &= chrmask;
	    chrswitch[0] = value;
	    chrdirty |= 1;
	    break;
	case 1:
	    value &= chrmask;
	    chrswitch[1] = value;
	    chrdirty |= 2;   
	    break;	    
	case 2:
	    value &= chrmask;
	    chrswitch[2] = value;	    
	    chrdirty |= 4;
	    break;
	case 3:
	    value &= chrmask;
	    chrswitch[3] = value;	    
	    chrdirty |= 8;
	    break;	    
	case 4:
	    value &= chrmask;
	    chrswitch[4] = value;	    
	    chrdirty |= 16;
	    break;	    
	case 5:
	    value &= chrmask;
	    chrswitch[5] = value;	    
	    chrdirty |= 32;
	    break;	    
	case 6:
	    value &= prgmask;
	    regs[0] = value;
	    prgdirty++;
	    break;
	case 7:
	    value &= prgmask;
	    regs[1] = value;
	    prgdirty++;
	    break;
	}
    else if (addr == 0xA000)
    {
	mirror = value & 1;
	if (value & 0x40)
	{
	    printf("1-screen\n");
	    nametable[0] = nt2;
	    nametable[1] = nt2;
	    nametable[2] = nt2;
	    nametable[3] = nt2;	    
	}
	else
	    if (mirror)
		MM_Mirror_Horiz(nt1, nt2);
	    else
		MM_Mirror_Vert(nt1, nt2);
    }
    else if (addr == 0xA001)
    {
	if (value & 0x80)
	    mmap((unsigned char *)0x6000, 0x2000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)sram);
	else
	    mmap((unsigned char *)0x6000, 0x2000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)ffpage);
    }
    else if (addr == 0xC000)
	irqcounter = value;
    else if (addr == 0xC001)
	irqlatch = value;
    else if (addr == 0xE000)
    {
	irqen = 0;
	irqcounter = irqlatch;
    }
    else if (addr == 0xE001)
	irqen = 1;
    if (prgdirty)
	MMC3_Update_ROM();
    if (chrdirty)
	MMC3_Update_CHR();
}

unsigned char MMC3_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void MMC3_Write_CHR(unsigned long addr, unsigned char value)
{
    if (nchrs)
	return;
    vrom[addr >> 10][addr & 0x3FF] = value;
}

void MMC3_Scanline_Hook(void)
{
    if (irqen)
	irqcounter--;
    if ((!irqcounter) && (irqen))
	irqfire++;
}

int MMC3_Check_IRQ(void)
{
    if (irqfire)
    {
	irqfire = 0;
	return 1;
    }
    return 0;
}

void MMC3_Vblank_Hook(void)
{
    irqcounter = irqlatch;
}
