#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "n6502.h"

#include "rom.h"
#include "memmap.h"
#include "newppu.h"

#include "cmdline.h"

#include "apu.h"

#include "mathops.h"

extern int zerofd;
extern int memfd;

unsigned char dumptables = 0;

extern unsigned char paldirty;

unsigned char sprram[0x100];

typedef struct
{
    unsigned char sy;
    unsigned char s;
} sprsortent;

sprsortent sprtable[64];

void (*Expandline)(unsigned char line, unsigned char *data);
void (*Decode_5bit)(unsigned char line, unsigned char shift, unsigned char *p0,
		    unsigned char *p1, unsigned char *p2, unsigned char *p3, unsigned char *p4);

void (*Clearline)(unsigned char line);
void (*Updatepal)(unsigned char *pal);
void updateimage();

unsigned long frames = 0;
unsigned char hitline = 0;

static unsigned char con_1;
static unsigned char con_2;
static unsigned char vram_buffer;
static unsigned char spraddr;
static unsigned long vram_addr;
static unsigned char status;
static unsigned short tmpscroll;
static unsigned char xfine;
static unsigned char ff;
static unsigned char sbuf[264];
static unsigned char spritestart = 0;
static unsigned char sprdirty = 0;

typedef struct
{
    unsigned char sy;
    unsigned char index;
    struct
    {
	unsigned char att0   : 1;
	unsigned char att1   : 1;
	unsigned char szero  : 1;
	unsigned char        : 2;
	unsigned char b      : 1;
	unsigned char hflip  : 1;
	unsigned char vflip  : 1;
    } f;
    unsigned char sx;
} sprite;

sprite *sprites = (sprite *)sprram;

unsigned char sff = 0;

void Sync()
{
    static struct timeval next1 = {0};
    struct timeval now;

    while (gettimeofday(&now, NULL) < 0) ;

    if (next1.tv_sec == 0)
    {
	next1 = now;
	next1.tv_usec++;
    }
    
    if (timercmp(&next1, &now, >))
    {
	do
	{
	    while (gettimeofday(&now, NULL) < 0) ;
	    
	} while (timercmp(&next1, &now, >));
    }
    else
    {
//	printf("YOU SUCK, MAKE IT FASTER\n");
	next1 = now;
    }

    next1.tv_usec += FRAMETIME;

    if (next1.tv_usec >= 1000000)
    {
	next1.tv_sec += next1.tv_usec / 1000000;
	next1.tv_usec %= 1000000;
    }

    if (look_ugly)
	sff ^= 1;
}

unsigned char Read_PPU_Register(unsigned short addr)
{
    unsigned char ret;

    switch (addr)
    {
    case 0x2000:
    case 0x2001:
    case 0x2002:
    case 0x2003:
    case 0x2004:
    case 0x2005:
    case 0x2006:
	ret = status;
	status &= ~STAT_VBLANK;
	ff = 0;
	return ret;

    case 0x2007:
	if (vram_addr >= 0x3F00)
	{
	    ret = Read_PPU(vram_addr);
	    vram_buffer = ret;
	}
	else
	{
	    ret = vram_buffer;
	    vram_buffer = Read_PPU(vram_addr);
	}
	if (con_1 & CON1_PPUINC)
	    vram_addr += 32;
	else
	    vram_addr++;
	vram_addr &= 0x3FFF;
	return ret;
	
    default:
	printf("Read from register $%x\n", addr);
    }
    return 0;
}

extern unsigned long cyclesRemaining;
extern unsigned long delay;
extern unsigned short VCount;

void Sort_SPRRAM()
{
    unsigned int i = 0;
    unsigned int j = 0;
    
    do
    {
	sprtable[i].sy = sprram[i << 2];
	sprtable[i].s = i;
	i++;
    } while(i < 64);

    for (i = 0; i < 63; i++)
	for (j = 0; j < 63 - i; j++)
	{
	    if (sprtable[j].sy > sprtable[j+1].sy)
	    {
		sprtable[j].sy ^= sprtable[j+1].sy;
		sprtable[j+1].sy ^= sprtable[j].sy;
		sprtable[j].sy ^= sprtable[j+1].sy;

		sprtable[j].s ^= sprtable[j+1].s;
		sprtable[j+1].s ^= sprtable[j].s;
		sprtable[j].s ^= sprtable[j+1].s;
	    }
	}
}

void Write_PPU_Register(unsigned short addr, unsigned char value)
{
    unsigned long i;

    switch (addr)
    {
    case 0x2000:
	con_1 = value;
	tmpscroll &= ~0x0C00;
	tmpscroll |= (value & 3) << 10;
	break;
	
    case 0x2001:
	con_2 = value;
	if (con_2 & 0xE0)
	    printf("Hmm, something's using the color intensify...\n");
	break;
	
    case 0x2003:
	spraddr = value;
	break;
	
    case 0x2004:
	sprram[spraddr++] = value;
	sprdirty = 1;
	break;
	
    case 0x2005:
	if (ff)
	{
	    tmpscroll &= ~0x73E0;
	    tmpscroll |= (value << 2) & 0x3E0;
	    tmpscroll |= (value & 7) << 12;
	}
	else
	{
	    tmpscroll &= ~0x001F;
	    tmpscroll |= (value >> 3);
	    xfine = (value & 7);
	}
	ff ^= 1;
	break;
	
    case 0x2006:
	if (ff)
	{
	    tmpscroll &= 0xFF00;
	    tmpscroll |= value;
	    vram_addr = tmpscroll;
	}
	else
	{
	    tmpscroll &= 0x00FF;
	    tmpscroll |= (value << 8) & 0x3F00;
	}
	ff ^= 1;
	break;
	
    case 0x2007:
	Write_PPU(vram_addr, value);
	if (con_1 & CON1_PPUINC)
	    vram_addr += 32;
	else
	    vram_addr++;
	vram_addr &= 0x3FFF;
	break;
	
    case 0x4014:
	i = value << 8;
	memcpy(sprram, (unsigned char *)i, 256);
	Sort_SPRRAM();
	delay = 513 - cyclesRemaining;
	cyclesRemaining = 0;
	break;
	
    default:
	printf("Write $%x to $%x\n", value, addr);
    }
}

void Dump_GFX_Tables()
{
    FILE *fp = fopen("screendump.dmp", "wb");

    /* Format: palram, spr ram, vrom, name/attrib tables, PPU regs in CON1, CON2, VRAM_ADDR */

    fwrite(palram, 1, 32, fp);
    fwrite(sprram, 1, 256, fp);
    
    fwrite(vrom[0], 1, 0x400, fp);
    fwrite(vrom[1], 1, 0x400, fp);
    fwrite(vrom[2], 1, 0x400, fp);
    fwrite(vrom[3], 1, 0x400, fp);
    fwrite(vrom[4], 1, 0x400, fp);
    fwrite(vrom[5], 1, 0x400, fp);
    fwrite(vrom[6], 1, 0x400, fp);
    fwrite(vrom[7], 1, 0x400, fp); 

    fwrite(nametable[0], 1, 0x400, fp);
    fwrite(nametable[1], 1, 0x400, fp);
    fwrite(nametable[2], 1, 0x400, fp);
    fwrite(nametable[3], 1, 0x400, fp);

    fwrite(&con_1, 1, 1, fp);
    fwrite(&con_2, 1, 1, fp);
    fwrite(&vram_addr, 4, 1, fp);

    fclose(fp);
}

void PPU_Start_Vblank()
{
    unsigned char draw = 1;

    if (sff)
    {
	draw = 0;
    }

    Sync();

    frames++;
    spritestart = 0;
    
    if (draw)
	updateimage();

    status |= STAT_VBLANK;
//    status = STAT_VBLANK;
    
    if (cur_mapper->Vblank_Hook)
	cur_mapper->Vblank_Hook();
}

unsigned char PPU_Check_NMI()
{
    return (con_1 & CON1_NMI);
}

void PPU_End_Vblank()
{
//    status &= ~STAT_HIT;
    status = 0; // hmm, nyef says it works for bigfoot...
    
    if (con_2 & (CON2_SDRAW | CON2_BDRAW))
    {
	vram_addr = tmpscroll;
    }
    
    if (dumptables)
    {
	dumptables = 0;
	Dump_GFX_Tables();
    }
    
}

extern unsigned short VCount;

typedef union
{
    unsigned char b[36];
    unsigned long l[9];
} plane;

plane s;
plane p3;
plane p2;
plane p1;
plane p0;

plane bp0;
plane bp1;
plane ba0;
plane ba1;

plane p;
plane z;
plane sp0;
plane sp1;
plane sa0;
plane sa1;

unsigned long decoderlut[16] = {

    0x00000000,
    0x01000000,
    0x00010000,
    0x01010000,
    0x00000100,
    0x01000100,
    0x00010100,
    0x01010100,
    0x00000001,
    0x01000001,
    0x00010001,
    0x01010001,
    0x00000101,
    0x01000101,
    0x00010101,
    0x01010101
};

void Unpack_Planes(void)
{
    unsigned long acc;
    unsigned char mask = 0x0F;
    unsigned char shift = 4;
    unsigned char i = 0;
    unsigned char stop = 32;
    unsigned long *sb = (unsigned long *)sbuf;

    if (xfine)
	stop++;

    while (i < stop)
    {
	acc = decoderlut[s.b[i] >> 4];
	acc <<= 1;
	acc |= decoderlut[p3.b[i] >> 4];
	acc <<= 1;
	acc |= decoderlut[p2.b[i] >> 4];
	acc <<= 1;
	acc |= decoderlut[p1.b[i] >> 4];
	acc <<= 1;
	acc |= decoderlut[p0.b[i] >> 4];
	
	*sb++ = acc;
	
	acc = decoderlut[s.b[i] & 0xF];
	acc <<= 1;
	acc |= decoderlut[p3.b[i] & 0xF];
	acc <<= 1;
	acc |= decoderlut[p2.b[i] & 0xF];
	acc <<= 1;
	acc |= decoderlut[p1.b[i] & 0xF];
	acc <<= 1;
	acc |= decoderlut[p0.b[i] & 0xF];
	*sb++ = acc;	
	i++;
    }
}

void Shuffle_Planes()
{
    unsigned long hitflag = 0;
    unsigned long t, u;
    unsigned char i = 0;

    for (i = 0; i < 9; i++)
    {
	t = sp0.l[i] | sp1.l[i];
	u = bp0.l[i] | bp1.l[i];
	hitflag |= z.l[i] & (t & u);
	
	s.l[i] = ((~p.l[i]) & t) | (p.l[i] & (~u));
	t = s.l[i];
	u = ~t;
	p1.l[i] = (t & sp1.l[i]) | (u & bp1.l[i]);
	p0.l[i] = (t & sp0.l[i]) | (u & bp0.l[i]);
	p3.l[i] = (t & sa1.l[i]) | (u & ba1.l[i]);
	p2.l[i] = (t & sa0.l[i]) | (u & ba0.l[i]);
    }
    if (hitflag)
	status |= STAT_HIT;
}

unsigned char attlut[2] = { 0x00, 0xFF };
unsigned char newattlut[4][2] = { { 0x00, 0x00 }, { 0xFF, 0x00 }, { 0x00, 0xFF }, { 0xFF, 0xFF }};
unsigned char *vromb[4];

void Fetch_BG()
{
    int i;
    unsigned long tile;              // temp storage of tile index, att byte, and low 5 bits of vram addr
    unsigned char bank = (con_1 & CON1_BPAT) >> 2;
    unsigned char table = (vram_addr >> 10) & 3;
    unsigned char *nt = nametable[table];
    unsigned long atb = 0x3C0 + ((vram_addr >> 4) & 0x38);
    unsigned char mask = 1 << ((vram_addr >> 4) & 4);
    unsigned long sva = (vram_addr & (~0x3FF) ^ 0x400);
    unsigned char *b0 = bp0.b, *b1 = bp1.b, *a0 = ba0.b, *a1 = ba1.b;
    
    vromb[0] = vrom[bank] + (vram_addr >> 12);
    vromb[1] = vrom[bank+1] + (vram_addr >> 12);
    vromb[2] = vrom[bank+2] + (vram_addr >> 12);
    vromb[3] = vrom[bank+3] + (vram_addr >> 12);
    
    vram_addr &= 0x3FF;
    
    for (i = 33; i; i--)
    {
	tile = nt[atb + ((vram_addr >> 2) & 7)];
	tile >>= (vram_addr & 2);
	
	*a0++ = ((tile & mask) == 0) - 1;
	*a1++ = ((tile & (mask << 1)) == 0) - 1;

	tile = nt[vram_addr];

	nt = vromb[tile >> 6];
	
	if (cur_mapper->CHR_Latch)
	    cur_mapper->CHR_Latch(bank, tile);

	tile <<= 4;
	tile &= 0x3FF;
	
	*b0++ = nt[tile];
    	*b1++ = nt[tile | 8];


	tile = vram_addr & 0x1F;
	tile++;
	
	vram_addr &= ~0x1F;
	vram_addr |= tile & 0x1F;
	table ^= tile >>= 5;
	nt = nametable[table];
    }

    vram_addr |= sva;
}

unsigned char nsprites;

unsigned char pat0[8];
unsigned char pat1[8];
unsigned char att0[8];
unsigned char att1[8];
unsigned char hflip[8];
unsigned char pri[8];
unsigned char xofs[8];
unsigned char szero[8];

void Shuffle_Sprites()
{
    unsigned char idx;
    unsigned char mask;
    unsigned char nmask;
    unsigned char shift;
        
    while (nsprites--)
    {
	idx = xofs[nsprites] >> 3;
	shift = xofs[nsprites] & 7;
	shift += xfine;
	idx += shift >> 3;
	shift &= 7;

	if (hflip[nsprites])
	{
	    pat0[nsprites] = flip(pat0[nsprites]);
	    pat1[nsprites] = flip(pat1[nsprites]);
	}

	mask = pat0[nsprites] | pat1[nsprites];
	mask = ror(mask, shift);
	nmask = (~(0xFF >> shift)) & mask;
	mask &= 0xFF >> shift;
	
	pat0[nsprites] = ror(pat0[nsprites], shift);
	pat1[nsprites] = ror(pat1[nsprites], shift);

	sp0.b[idx] &= ~mask;
	sp1.b[idx] &= ~mask;
	sa0.b[idx] &= ~mask;
	sa1.b[idx] &= ~mask;
	z.b[idx] &= ~mask;
	p.b[idx] &= ~mask;
	
	sp0.b[idx] |= (pat0[nsprites] & mask);
	sp1.b[idx] |= (pat1[nsprites] & mask);
	sa0.b[idx] |= (att0[nsprites] & mask);
	sa1.b[idx] |= (att1[nsprites] & mask);
	z.b[idx] |= (szero[nsprites] & mask);
	p.b[idx] |= (pri[nsprites] & mask);

	idx++;
	
	sp0.b[idx] &= ~nmask;
	sp1.b[idx] &= ~nmask;
	sa0.b[idx] &= ~nmask;
	sa1.b[idx] &= ~nmask;
	z.b[idx] &= ~nmask;
	p.b[idx] &= ~nmask;
	
	sp0.b[idx] |= (pat0[nsprites] & nmask);
	sp1.b[idx] |= (pat1[nsprites] & nmask);
	sa0.b[idx] |= (att0[nsprites] & nmask);
	sa1.b[idx] |= (att1[nsprites] & nmask);
	z.b[idx] |= (szero[nsprites] & nmask);
	p.b[idx] |= (pri[nsprites] & nmask);	
    }
}

void Fetch_Sprites(unsigned char line)
{
    unsigned char sh;
    unsigned char i;
    unsigned char ofs;
    unsigned long bank;
    unsigned long tile;
    unsigned char ninc;
    
    sprite *spr = sprites;
    sprsortent splist[8];
    
    sprsortent *sp;

    spr->f.szero = 1;
    
    /* Clear out bitplanes -- FIXME try to eliminate these calls */

    memset(sp0.b, 0, 33);
    memset(sp1.b, 0, 33);

    sh = 8 + ((con_1 & CON1_SPRSIZE) >> 2);
    
    nsprites = 0;
    i = spritestart;
    sp  = &sprtable[spritestart];
    // Grabs the data for the first 8 sprites, data is sorted in order of SPR RAM index

    while ((sp->sy + sh < line) && (i < 64))
    {
	i++;
	sp++;
	spritestart++;
    }
    
    while ((nsprites < 8) && (sp->sy < line) && (i++ < 64))
	splist[nsprites++] = *sp++;

    for (ofs = 0; ofs < nsprites - 1; ofs++)
	for (i = 0; i < nsprites - 1 - ofs; i++)
	{
	    if (splist[i].s > splist[i+1].s)
	    {
		splist[i].sy ^= splist[i+1].sy;
		splist[i+1].sy ^= splist[i].sy;
		splist[i].sy ^= splist[i+1].sy;

		splist[i].s ^= splist[i+1].s;
		splist[i+1].s ^= splist[i].s;
		splist[i].s ^= splist[i+1].s;
	    }
	}

    for (i = 0; i < nsprites; i++)
    {
	ofs = (line - (splist[i].sy + 1)) & (sh - 1);
	spr = &sprites[splist[i].s];
	
	if (spr->f.vflip)
	    ofs = (sh - 1 - ofs);
	
	tile = spr->index;
	tile &= ~(sh >> 4);
	tile |= ofs >> 3;
	ofs &= 7;
	
	bank = ((sh >> 4) & spr->index) | ((sh & con_1 & CON1_SPAT) >> 3);

	xofs[i] = spr->sx;

	szero[i] = attlut[spr->f.szero]; // HACK, there has GOT to be a way to get rid of this if
	    
	att0[i] = attlut[spr->f.att0];
	att1[i] = attlut[spr->f.att1];
	hflip[i] = spr->f.hflip;
	pri[i] = attlut[spr->f.b];
	
	tile += bank << 8;
	
	if (cur_mapper->CHR_Latch)
	    cur_mapper->CHR_Latch(bank, tile);
	
	tile <<= 4;
	tile += ofs;
	
	pat0[i] = vrom[tile >> 10][tile & 0x3FF];
	pat1[i] = vrom[tile >> 10][(tile & 0x3FF) + 8];
    }
    
}

void PPU_NewRenderline(unsigned char line)
{
    /* Update frame skipper ff */

    if (sff)
    {
	if (line == hitline)
	    status |= STAT_HIT;
	return;
    }

    if (sprdirty)
    {
	Sort_SPRRAM();
	sprdirty = 0;
    }
    
    /* Kick the backend, update the color decoder lut */
    
    if ((paldirty) && (con_2 & (CON2_BDRAW | CON2_SDRAW)))
    {
	paldirty = 0;
	Updatepal(palram);
    }

    /* Reset back to the start of the nametable line */
    
    if (con_2 & (CON2_SDRAW | CON2_BDRAW))
    {
	vram_addr &= ~0x041F;
	vram_addr |= tmpscroll & 0x041F;
    }

    if (con_2 & CON2_BDRAW)
    {
	Fetch_BG();
    }
    else
    {
	memset(bp0.b, 0, 33);
	memset(bp1.b, 0, 33);
    }

    if (con_2 & CON2_SDRAW)
    {
	Shuffle_Sprites();
    }
    
    if ((con_2 & (CON2_BDRAW | CON2_SDRAW)) && (line >= 8) && (line <= 232))
    {
	unsigned char mask = 0xFF >> xfine;
	
	if (~con_2 & CON2_BCLIP)
	{
	    bp0.b[0] &= ~mask;
	    bp1.b[0] &= ~mask;
	    bp0.b[1] &= mask;
	    bp1.b[1] &= mask;
	}

	if (~con_2 & CON2_SCLIP)
	{
	    sp0.b[0] &= ~mask;
	    sp1.b[0] &= ~mask;
	    sp0.b[1] &= mask;
	    sp1.b[1] &= mask;
	}
	Shuffle_Planes();
//	Unpack_Planes();
//	Expandline(line, sbuf + xfine);
	Decode_5bit(line, xfine, p0.b, p1.b, p2.b, p3.b, s.b);
    }
    else
    {
	Clearline(line);
    }

    if (con_2 & CON2_SDRAW)
    {
	Fetch_Sprites(line + 1);
    }
    else
    {
	memset(sp0.b, 0, 33);
	memset(sp1.b, 0, 33);
	memset(z.b, 0, 33);
    }
    
    /* Inc to the next line */
    
    if (con_2 & (CON2_SDRAW | CON2_BDRAW))
    {
	vram_addr += 0x1000;
	if (vram_addr & 0x8000)
	{
	    unsigned short tmp;
	    
	    vram_addr &= 0x0FFF;
	    tmp = vram_addr & 0x03E0;

	    if (tmp == 0x3A0)
	    {
		vram_addr ^= 0x800;
		tmp = 0;
	    }
	    else
	    {
		tmp += 0x20;
	    }
	    
	    vram_addr &= ~0x03E0;
	    vram_addr |= tmp & 0x03E0;
	}

	if (cur_mapper->Scanline_Hook)
	    cur_mapper->Scanline_Hook();	
    }
}

void PPU_OldRenderline(unsigned char line)
{
}

void Init_PPU(void)
{
    memset(sprram, 0, 256);
}

