#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "n6502.h"

#include "rom.h"
#include "nsf.h"
#include "memmap.h"

CONTEXTM6502 *n6502thingy;

unsigned char *prg;
unsigned char cursong;
int switched = 0;
extern int memfd;

unsigned char haltcode[] = { 0xEA, 0xEA, 0xEA, 0xEA,
			     0x80, 0x00, 0x5F, 0xEA,
			     0xEA, 0xEA, 0xEA, 0xEA }; // STA $5F00

void InitNSF(rominfo *ri)
{
    int i;
    
    for (i = 0; i < 8; i++)
	if (ri->switchinit[i] != 0)
	    switched = 1;

    prg = ri->prgrom;
    
    if (switched)
    {
	printf("Switched\n");
	for (i = 0; i < 8; i++)
	{
	    mmap((unsigned char *)0x8000 + i*0x1000, 0x1000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd,
		 (unsigned long)ri->prgrom + ri->switchinit[i] * 0x1000);
	}
    }
    else
    {
	printf("Loading at page %x, %d bytes\n", ri->loadaddr & 0xF000, 0x10000 - (ri->loadaddr & 0xF000));
	
	mmap((unsigned char *)(ri->loadaddr & 0xF000), 0x10000 - (ri->loadaddr & 0xF000),
	     PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)ri->prgrom);
	
	printf("%x %x %x %x\n", ram[ri->loadaddr], ram[ri->loadaddr + 1],
	       ram[ri->loadaddr + 2], ram[ri->loadaddr + 3]);
    }
    
    memcpy(&ram[0x6000], haltcode, sizeof(haltcode));
    return;
}

void InitNSF_CPU(rominfo *ri)
{
    memset(ram, 0, 0x1000);
    Write_APU_Register(0x4015, 0xF);

    m6502zpGetContext(n6502thingy);
    n6502thingy->m6502af &= 0xFF00;
    n6502thingy->m6502af |= ri->startsong - 1;
    n6502thingy->m6502x = 0;
    n6502thingy->m6502s = 0xFD;
    ram[0x1FF] = 0x60;
    ram[0x1FE] = 0x00;
    n6502thingy->m6502pc = ri->initaddr;
    m6502zpSetContext(n6502thingy);
    m6502zpexec(4096);
    
    cursong = ri->startsong - 1;
}

unsigned char NSF_Read_Mapper(unsigned long addr)
{
    return 0;
}

extern unsigned long cyclesRemaining;

void NSF_Write_Mapper(unsigned long addr, unsigned char value)
{
    if (switched)
	if ((addr & 0xFFF8) == 0x5FF8)
	{
	    printf("NSF switch\n");
	    switch (addr & 7)
	    {
	    case 0:
		mmap((unsigned char *)0x8000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;
	    case 1:
		mmap((unsigned char *)0x9000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    case 2:
		mmap((unsigned char *)0xA000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    case 3:
		mmap((unsigned char *)0xB000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    case 4:
		mmap((unsigned char *)0xC000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    case 5:
		mmap((unsigned char *)0xD000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    case 6:
		mmap((unsigned char *)0xE000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    case 7:
		mmap((unsigned char *)0xF000, 0x1000, PROT_READ|PROT_WRITE,
		     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prg + value * 0x1000);
		break;	    
	    }
	}
    
    if (addr == 0x5F00)
    {
	printf("NSF halt\n");
	cyclesRemaining = 0;	
    }
}

extern unsigned char quitemu;
extern unsigned char changesong;
void Sync(void);

void NSF_MainLoop(rominfo *ri)
{

    while (!quitemu)
    {
	while (!changesong)
	{
	    m6502zpGetContext(n6502thingy);
	    n6502thingy->m6502s = 0xFD;
	    ram[0x1FF] = 0x60;
	    ram[0x1FE] = 0x00;
	    n6502thingy->m6502pc = ri->playaddr;
	    m6502zpSetContext(n6502thingy);
	    
	    m6502zpexec(65536);
	    Sync();
	    Process_IO(0);
	    APU_Update();
	    
	}

	cursong %= ri->maxsongs;
	changesong = 0;
	m6502zpGetContext(n6502thingy);
	n6502thingy->m6502s = 0xFD;
	n6502thingy->m6502af &= 0xFF00;
	n6502thingy->m6502af |= cursong;
	n6502thingy->m6502x = 0;
	ram[0x1FF] = 0x60;
	ram[0x1FE] = 0x00;
	n6502thingy->m6502pc = ri->playaddr;
	m6502zpSetContext(n6502thingy);
    }
}
