#include <stdio.h>
#include <unistd.h>
#include "cmdline.h"
#include "osssound.h"
#include "n6502.h"
#include "apu.h"

extern CONTEXTM6502 *n6502thingy;

int numevents = 0;

typedef struct
{
    unsigned long timestamp;
    unsigned short addr;
    unsigned char value;
} event;

/* MAGIC formula: MAGIC = nes magic / sample rate */

#define MAGIC 0xA2F968E1
#define SAWMAGIC 0xA9875B11

unsigned long vrc6_sawtime[7] = { 0x09249249, 0x12492492, 0x1B6DB6DB, 0x24924924,
				  0x2DB6DB6D, 0x36DB6DB6, 0x40000000 };
				  
unsigned long vrc6_duty[8] = { 0x04000000, 0x08000000, 0x0C000000, 0x10000000,
			       0x14000000, 0x18000000, 0x1C000000, 0x20000000 };
unsigned long pulse_duty[4] = { 0x08000000, 0x10000000, 0x20000000, 0x30000000 };
unsigned short noise_count[16] = { 44100 / 19200, 44100 / 9600, 44100 / 4800, 44100 / 2431,
				   44100 / 1189, 44100 / 795, 44100 / 597, 44100 / 477,
				   44100 / 378, 44100 / 301, 44100 / 201, 44100 / 156,
				   44100 / 79, 44100 / 53, 44100 / 27, 44100 / 14 };
unsigned long pcm_count[16] = { 0x1AC, 0x17C, 0x154, 0x140, 0x11E, 0x0FE, 0x0E2, 0x0D6,
				0x0BE, 0x0A0, 0x08E, 0x080, 0x06A, 0x055, 0x048, 0x036 };

unsigned char noise_loop[92];
event events[10000];
extern unsigned long framecount;
unsigned short noise_reg;

/* 735 cycles/refresh */

unsigned char output[2048];
int wavptr = 0;
int wavflag = 0;

unsigned char mixbuf[735];
unsigned char pulse1_buf[735];
unsigned char pulse2_buf[735];
unsigned char tri_buf[735];
unsigned char noise_buf[735];
unsigned char pcm_buf[735];
unsigned char vrc6_buf[735];

unsigned char enable_pulse1 = 1;
unsigned char enable_pulse2 = 1;
unsigned char enable_tri = 1;
unsigned char enable_noise = 1;
unsigned char enable_pcm = 1;

unsigned long pulse1_duty = 0x08000000;
unsigned long pulse2_duty = 0x08000000;

int pulse1_ptr;
int pulse2_ptr;
int tri_ptr;
int noise_ptr;
int pcm_ptr;

FILE *fp;

int tickfuse = 60;

void APU_Init(void)
{
    memset(pulse1_buf, 0, 735);
    memset(pulse2_buf, 0, 735);
    memset(tri_buf, 0, 735);
    memset(pcm_buf, 0, 735);
    memset(noise_buf, 0, 735);
    memset(vrc6_buf, 0, 735);
    memset(mixbuf, 0, 735);
    fp = fopen("tridump", "wt");
    if (pcm_fd)
      write(pcm_fd, pulse1_buf, 735);
}

#define PULSE1_D   1
#define PULSE2_D   2
#define TRI_D      4
#define NOISE_D    8
#define PCM_D     16

unsigned char apu_status = 0;
unsigned char apu_control = 0;
unsigned long pulse1_counter = 0;
unsigned long pulse2_counter = 0;
unsigned long tri_counter = 0;
unsigned long pcm_counter = 0;
unsigned long noise_counter = 0;

unsigned char Read_APU_Register(unsigned short addr)
{
    if (addr == 0x4015)
    {
	apu_status = 0;
       
	if (pulse1_counter)
	    apu_status |= PULSE1_D;
	if (pulse2_counter)
	    apu_status |= PULSE2_D;
	if (tri_counter)
	    apu_status |= TRI_D;
	if (noise_counter)
	    apu_status |= NOISE_D;
	if (pcm_counter < 0xFFFFFFFF)
	    apu_status |= PCM_D;
    }

    return apu_status;
}

void Write_APU_Register(unsigned short addr, unsigned char value)
{
    events[numevents].timestamp = m6502zpGetElapsedTicks(0);
    events[numevents].addr = addr;
    events[numevents].value = value;
/*    if ((addr >= 0x400B) && (addr <= 0x400B))
      printf("%lx: Write %x to %x\n", events[numevents].timestamp, value, addr);*/

    numevents++;
}

unsigned char countfunc[32] =
{
    5, 128, 10, 1, 20, 2, 40, 3, 80, 4, 30, 5, 7, 6, 13, 7,
    6, 8, 12, 9, 24, 10, 48, 11, 96, 12, 36, 13, 8, 14, 16, 15
};

int ptr = 0;

unsigned long pulse1_step = 0;
unsigned long pulse1_pos = 0;
unsigned char pulse1_reg[4];
unsigned char pulse1_vol;
unsigned char pulse1_envrate = 0;
unsigned char pulse1_loop = 0;
unsigned char pulse1_env = 0;
unsigned char pulse1_sweep = 0;
unsigned char pulse1_sweepdir = 0;
unsigned char pulse1_sweepdelta = 0;
unsigned char pulse1_sweepspeed = 0;
unsigned long pulse1_sweepfuse = 0;
unsigned char pulse1_envpos = 0;
unsigned long pulse1_freq = 0;
unsigned long pulse1_playfreq = 0;

unsigned long pulse2_step = 0;
unsigned long pulse2_pos = 0;
unsigned char pulse2_reg[4];
unsigned char pulse2_vol;
unsigned char pulse2_envrate = 0;
unsigned char pulse2_loop = 0;
unsigned char pulse2_env = 0;
unsigned char pulse2_sweep = 0;
unsigned char pulse2_sweepdir = 0;
unsigned char pulse2_sweepdelta = 0;
unsigned char pulse2_sweepspeed = 0;
unsigned long pulse2_sweepfuse = 0;
unsigned char pulse2_envpos = 0;
unsigned long pulse2_freq = 0;
unsigned long pulse2_playfreq = 0;

unsigned long tri_step = 0;
unsigned long tri_pos = 0;
unsigned long tri_playfreq = 0;
unsigned long tri_counter2 = 0;
unsigned long tri_counter2latch = 0;
unsigned char tri_fired = 0;
unsigned char tri_hold = 0;
unsigned long tri_fuse = 735 >> 2;

unsigned long noise_timer = 0;
unsigned long noise_freq = 0;
unsigned char noise_mode = 0;
unsigned char noise_vol = 0;
unsigned char noise_pos = 0;

unsigned char pcm_vol = 0;
unsigned char pcm_setvol = 0;
unsigned char pcm_rate = 0;
unsigned char pcm_irq = 0;
unsigned char pcm_loop = 0;
unsigned long pcm_addr = 0;
unsigned long pcm_len = 0;
unsigned long pcm_pos = 0;
unsigned char pcm_byte = 0;
unsigned char pcm_mask = 0;
unsigned char *pcm_base = (unsigned char *)0xC000;

unsigned long vrc6_pulse1_duty = 0;
unsigned long vrc6_pulse1_pos = 0;
unsigned char vrc6_pulse1_vol = 0;
unsigned char vrc6_pulse1_gate = 0;
unsigned long vrc6_pulse1_freq = 0;
unsigned char vrc6_pulse1_enable = 0;
unsigned long vrc6_pulse1_step = 0;

unsigned long vrc6_pulse2_duty = 0;
unsigned long vrc6_pulse2_pos = 0;
unsigned char vrc6_pulse2_vol = 0;
unsigned char vrc6_pulse2_gate = 0;
unsigned long vrc6_pulse2_freq = 0;
unsigned char vrc6_pulse2_enable = 0;
unsigned long vrc6_pulse2_step = 0;

unsigned char vrc6_phase_in = 0;
unsigned long vrc6_phase_freq = 0;
unsigned long vrc6_phase_enable = 0;
unsigned long vrc6_phase_step = 0;
unsigned long vrc6_phase_pos = 0;
unsigned char vrc6_phase_state = 0;
unsigned char vrc6_phase_acc = 0;

void update_vrc6()
{
    vrc6_buf[ptr] = 0;
    if (vrc6_pulse1_enable)
    {
	if (vrc6_pulse1_gate)
	    vrc6_buf[ptr] += vrc6_pulse1_vol;
	else
	{
	    vrc6_pulse1_pos += vrc6_pulse1_step;
	    vrc6_pulse1_pos &= 0x3FFFFFFF;

	    if (vrc6_pulse1_pos < vrc6_pulse1_duty)
		vrc6_buf[ptr] += vrc6_pulse1_vol;
	}
    }
    if (vrc6_pulse2_enable)
    {
	if (vrc6_pulse2_gate)
	    vrc6_buf[ptr] += vrc6_pulse2_vol;
	else
	{
	    vrc6_pulse2_pos += vrc6_pulse2_step;
	    vrc6_pulse2_pos &= 0x3FFFFFFF;

	    if (vrc6_pulse2_pos < vrc6_pulse2_duty)
		vrc6_buf[ptr] += vrc6_pulse2_vol;
	}
    }
    if (vrc6_phase_enable)
    {
	vrc6_phase_pos += vrc6_phase_step;
	while (vrc6_phase_pos > vrc6_sawtime[vrc6_phase_state])
	{
	    vrc6_phase_state++;
	    vrc6_phase_acc += vrc6_phase_in;
	    if (vrc6_phase_state == 7)
	    {
		vrc6_phase_acc = 0;
		vrc6_phase_state = 0;
		vrc6_phase_pos -= 0x40000000;
	    }
	}
	vrc6_buf[ptr] += vrc6_phase_acc >> 3;
    }
}

int pulse1_check()
{
    if (!apu_control & PULSE1_D)
	return 0;
    
    if (!pulse1_counter)
	return 0;
    
    if ((pulse1_playfreq < 8) || (pulse1_playfreq > 0x7FF))
	return 0;
    
    if ((pulse1_sweepdir == 0) && (pulse1_playfreq & 0x400))
	return 0;
    return 1;
}

void update_pulse1()
{
    if (pulse1_check())
    {
	pulse1_step = MAGIC / (pulse1_playfreq + 1);
    
	pulse1_pos += pulse1_step;
	pulse1_pos &= 0x3FFFFFFF;
	
	if (pulse1_pos < pulse1_duty)
	    pulse1_buf[ptr] = pulse1_vol;
	else
	    pulse1_buf[ptr] = 0;
    }
    else
	pulse1_buf[ptr] = 0;
}

int pulse2_check()
{
    if (!apu_control & PULSE2_D)
	return 0;
    
    if (!pulse2_counter)
	return 0;
    
    if ((pulse2_playfreq < 8) || (pulse2_playfreq > 0x7FF))
	return 0;
    
    if ((pulse2_sweepdir == 0) && (pulse2_playfreq & 0x400))
	return 0;
    return 1;
}

void update_pulse2()
{
    if (pulse2_check())
    {
	pulse2_step = MAGIC / (pulse2_playfreq + 1);
    
	pulse2_pos += pulse2_step;
	pulse2_pos &= 0x3FFFFFFF;
	
	if (pulse2_pos < pulse2_duty)
	    pulse2_buf[ptr] = pulse2_vol;
	else
	    pulse2_buf[ptr] = 0;
    }
    else
	pulse2_buf[ptr] = 0;
}

int tri_check()
{
    if (~apu_control & TRI_D)
	return 0;
    
    if (!tri_counter)
	return 0;

    if (!tri_counter2)
	return 0;
    
    if ((tri_playfreq < 8) || (tri_playfreq > 0x7FF))
	return 0;
    
    return 1;
}

unsigned char tri_level;

void update_tri()
{
    if (tri_check())
    {
	tri_step = MAGIC / ((tri_playfreq) << 1);
    
	tri_pos += tri_step;
	tri_pos &= 0x3FFFFFFF;
	
	tri_level = (tri_pos & 0x1E000000) >> 25;
	if (tri_pos & 0x20000000)
	    tri_level = (~tri_level & 0x0F);
	tri_buf[ptr] = tri_level;
    }
    else
	tri_buf[ptr] = tri_level;

}

int noise_check()
{
    if (~apu_control & NOISE_D)
	return 0;

    if (!noise_counter)
	return 0;
    
    return 1;
}

void update_noise()
{
    if (noise_check())
    {
	if (!noise_timer)
	{
	    noise_timer = noise_count[noise_freq];
	    noise_reg += 511;

	    noise_loop[noise_pos++] = (noise_reg & 0xFF);
	    if (noise_pos > 91)
		noise_pos = 0;
	}
	else
	    noise_timer--;
	
/*	if (noise_mode)
	    noise_buf[ptr] = noise_loop[noise_pos];
	    else*/
	    noise_buf[ptr] = rand() & 0xFF;
    }
    else
	noise_buf[ptr] = 0;
}

unsigned long pcm_end = 0;
void update_pcm()
{
    if (~apu_control & PCM_D)
    {
	pcm_buf[ptr] = 0;
	return;
    }

//    printf("pcm_pos: %d\n", pcm_pos);
    
    if (pcm_pos <= pcm_len)
    {
	if (pcm_counter >= 41)
	    pcm_counter -= 41;
	else
	    pcm_counter = 0;
	pcm_end = 0;
    }
    else
    {
	pcm_end++;
	pcm_counter = 0xFFFFFFFF;
    }

/*    if (pcm_end == 1)
      printf("pcm_endvol: %d\n", pcm_vol);*/
    
    if (pcm_counter == 0)
    {
	if (!pcm_mask)
	{
	    pcm_byte = *(pcm_base + pcm_addr + pcm_pos);
	    pcm_mask = 0x80;
	    pcm_pos++;
	    if (pcm_pos == pcm_len)
	    {
		if (pcm_loop)
		    pcm_pos = 0;
		/* FIXME - IRQ */
	    }
	}
	if (pcm_byte & pcm_mask)
	{
	    if (pcm_setvol < 0x7F)
		pcm_setvol++;
	}
	else
	{
	    if (pcm_setvol)
		pcm_setvol--;
	}
	pcm_mask >>= 1;
	pcm_counter = pcm_count[pcm_rate];
    }
    if (pcm_vol < pcm_setvol - 1)
	pcm_vol += 2;
    else if (pcm_vol < pcm_setvol)
	pcm_vol++;
    else if (pcm_vol > pcm_setvol + 1)
	pcm_vol -= 2;
    else if (pcm_vol > pcm_setvol)
	pcm_vol--;
    
    pcm_buf[ptr] = pcm_vol;    
}

void APU_Update(void)
{
    event *ev = events;
    int i = 0;
    int nev = 0;
    int cycle = 41;
    int chanflag = 0;
    int lastswp = 0;
    
    ptr = 0;
    events[numevents].timestamp = -1;
    pulse1_envpos = 1;

//    fprintf(fp, "Frame\n");
    
    while (ptr < 735)
    {
	update_pulse1();
	update_pulse2();
	update_tri();
	update_pcm();
	update_vrc6();
//	update_noise();
	
	nev = 0;	

	while ((ev->timestamp != -1) && (ev->timestamp < cycle))
	{
	    switch (ev->addr)
	    {
	    case 0x4000:
		pulse1_loop = (ev->value >> 5) & 1;
		if (ev->value & 0x10)
		{
		    pulse1_vol = (ev->value & 0xF);
		    pulse1_envrate = 0;
		    pulse1_env = 0;
		}   
		else
		{
//		    pulse1_envrate = (ev->value & 0xF);
//		    if (pulse1_env == 0)
			pulse1_vol = 0xF;
		    pulse1_env = 1;
		}
		pulse1_duty = pulse_duty[ev->value >> 6];
		break;
	    case 0x4001:
		pulse1_sweep = ev->value >> 7;
		pulse1_sweepdir = (ev->value >> 3) & 1;
		pulse1_sweepdelta = ev->value & 7;
		pulse1_sweepspeed = ((ev->value >> 4) & 7) + 1;
		pulse1_sweepfuse = 735 / (pulse1_sweepspeed);
		break;
	    case 0x4002:
		pulse1_freq &= ~0xFF;
		pulse1_freq |= ev->value;
		pulse1_playfreq = pulse1_freq;
		break;
	    case 0x4003:
		pulse1_counter = countfunc[ev->value >> 3];
		pulse1_freq &= ~0x700;
		pulse1_freq |= (ev->value & 7) << 8;
		pulse1_playfreq = pulse1_freq;
		break;
	    case 0x4004:
		pulse2_loop = (ev->value >> 5) & 1;
		if (ev->value & 0x10)
		{
		    pulse2_vol = (ev->value & 0xF);
		    pulse2_envrate = 0;
		    pulse2_env = 0;
		}
		else
		{
//		    pulse2_envrate = (ev->value & 0xF);
		    pulse2_envrate = 1;
//		    if (pulse2_env == 0)
			pulse2_vol = 0xF;
		    pulse2_env = 1;
		}
		pulse2_duty = pulse_duty[ev->value >> 6];
//		printf("Write: %x to 4004\n", ev->value);
		break;
	    case 0x4005:
		pulse2_sweep = ev->value >> 7;
		pulse2_sweepdir = (ev->value >> 3) & 1;
		pulse2_sweepdelta = ev->value & 7;
		pulse2_sweepspeed = ((ev->value >> 4) & 7) + 1;
		pulse2_sweepfuse = 735 / (pulse2_sweepspeed);
		break;
	    case 0x4006:
		pulse2_freq &= ~0xFF;
		pulse2_freq |= ev->value;
		pulse2_playfreq = pulse2_freq;
		break;
	    case 0x4007:
		pulse2_counter = countfunc[ev->value >> 3];
		pulse2_freq &= ~0x700;
		pulse2_freq |= (ev->value & 7) << 8;
		pulse2_playfreq = pulse2_freq;
		break;		
	    case 0x4008:
		tri_hold = ev->value >> 7;
		tri_counter2latch = ev->value & 0x7F;
		break;
	    case 0x400A:
		tri_playfreq &= ~0xFF;
		tri_playfreq |= ev->value;
		break;
	    case 0x400B:
		tri_playfreq &= ~0x700;
		tri_playfreq |= (ev->value & 7) << 8;
		tri_counter = countfunc[ev->value >> 3];
		tri_counter2 = tri_counter2latch;
		break;
	    case 0x400C:
		noise_vol = (ev->value & 0xF) << 3;
//		printf("Write %x to 400C\n", ev->value);		
		break;
	    case 0x400E:
		noise_mode = (ev->value) >> 7;
		noise_freq = (ev->value & 0xF);
		noise_timer = noise_count[noise_freq];
//		printf("Write %x to 400E\n", ev->value);		
		break;
	    case 0x400F:
		noise_counter = countfunc[(ev->value >> 3)];
//		printf("Write %x to 400F\n", ev->value);
		break;
	    case 0x4010:
		pcm_irq = (ev->value >> 7);
		pcm_loop = (ev->value >> 6) & 1;
		pcm_rate = (ev->value & 0xF);
		pcm_counter = pcm_count[pcm_rate];
//		printf("pcm_counter: %d\n", pcm_counter);
		break;
	    case 0x4011:
		pcm_setvol = (ev->value >> 1);
//		printf("pcm_setvol: %d\n", pcm_setvol);
		break;
	    case 0x4012:
		pcm_addr = (ev->value << 6);
		pcm_mask = 0;
//		printf("pcm_addr: %x\n", pcm_addr);
		break;
	    case 0x4013:
		pcm_len = (ev->value << 4) + 1;
		pcm_pos = 0;
//		printf("pcm_len: %d\n", pcm_len);
		break;
	    case 0x4015:
		apu_control = ev->value;
		break;
		
	    case 0x9000: // VRC6 from here on
		vrc6_pulse1_gate = ev->value >> 7;
		vrc6_pulse1_duty = vrc6_duty[(ev->value >> 4) & 7];
		vrc6_pulse1_vol = ev->value & 0xF;		
		break;
	    case 0x9001:
		vrc6_pulse1_freq &= ~0xFF;
		vrc6_pulse1_freq |= ev->value;
		vrc6_pulse1_step = MAGIC / (vrc6_pulse1_freq + 1);
		break;
	    case 0x9002:
		vrc6_pulse1_enable = ev->value >> 7;
		vrc6_pulse1_freq &= ~0xF00;
		vrc6_pulse1_freq |= (ev->value & 0xF) << 8;
		vrc6_pulse1_step = MAGIC / (vrc6_pulse1_freq + 1);
		break;
		
	    case 0xA000:
		vrc6_pulse2_gate = ev->value >> 7;
		vrc6_pulse2_duty = vrc6_duty[(ev->value >> 4) & 7];
		vrc6_pulse2_vol = ev->value & 0xF;
		break;
	    case 0xA001:
		vrc6_pulse2_freq &= ~0xFF;
		vrc6_pulse2_freq |= 0xFF;
		vrc6_pulse2_step = MAGIC / (vrc6_pulse2_freq + 1);
		break;		
	    case 0xA002:
		vrc6_pulse2_enable = ev->value >> 7;
		vrc6_pulse2_freq &= ~0xF00;
		vrc6_pulse2_freq |= (ev->value & 0xF) << 8;
		vrc6_pulse2_step = MAGIC / (vrc6_pulse2_freq + 1);
		break;
		
	    case 0xB000:
		vrc6_phase_in = ev->value & 0x3F;
		break;
	    case 0xB001:
		vrc6_phase_freq &= ~0xFF;
		vrc6_phase_freq |= 0xFF;
		vrc6_phase_step = SAWMAGIC / (vrc6_phase_freq + 1);
		break;
	    case 0xB002:
		vrc6_phase_enable = ev->value >> 7;
		vrc6_phase_freq &= ~0xF00;
		vrc6_phase_freq |= (ev->value & 0xF) << 8;
		vrc6_phase_step = SAWMAGIC / (vrc6_phase_freq + 1);
		break;
	    }
	    ev++;
	}
	
	cycle += 41;

	nev = 0;

	if (tri_fuse)
	    tri_fuse--;
	
	if (!tri_hold)
	    tri_fired = 1;

	if (!tri_fuse)
	    tri_fuse = 735 >> 2;
	
	ptr++;
    }

    for (ptr = 0; ptr < 735; ptr++)
    {
//	output[wavptr] = (pulse1_buf[ptr] + pulse2_buf[ptr]) >> 1;
//	mixbuf[ptr] = (pulse1_buf[ptr] + pulse2_buf[ptr] + tri_buf[ptr] + pcm_buf[ptr]);// >> 1;
//	mixbuf[ptr] = pulse1_buf[ptr];
	mixbuf[ptr] = 0;
	if (enable_pulse1)
	    mixbuf[ptr] += pulse1_buf[ptr];
	if (enable_pulse2)
	    mixbuf[ptr] += pulse2_buf[ptr];
	if (enable_tri)
	    mixbuf[ptr] += tri_buf[ptr];
	if (enable_pcm)
	mixbuf[ptr] += pcm_buf[ptr];
	mixbuf[ptr] += vrc6_buf[ptr];
/*	mixbuf[ptr] >>= 1;
	if (enable_noise)
	mixbuf[ptr] += noise_buf[ptr] >> 1;*/
	
/*	wavptr++;
	if (wavptr == 2048)
	{
	    wavptr = 0;
	    wavflag = 2;
	}
	else if (wavptr = 1024)
	{
	    wavflag = 1;
	    }*/
    }
    
    if (pcm_fd)
      write(pcm_fd, mixbuf, 735);
    
/*    if (wavflag)
    {
	if (pcm_fd)
	{
	    write(pcm_fd, &output[(wavflag - 1) << 10], 1024);
	}
	wavflag = 0;
	}*/

	if (pulse1_envrate > pulse1_vol)
	{
	    if (pulse1_loop)
	    {
		pulse1_vol = 0xF;
	    }
	    else
	    {
		pulse1_vol = 0;
		pulse1_envrate = 0;
	    }
	}
	else
	    pulse1_vol -= pulse1_envrate;

	if (pulse2_envrate > pulse2_vol)
	{
	    if (pulse2_loop)
	    {
		pulse1_vol = 0xF;
	    }
	    else
	    {
		pulse2_vol = 0;
		pulse2_envrate = 0;
	    }
	}
	else
	    pulse2_vol -= pulse2_envrate;
	
	
    
    if (pulse1_counter)
	pulse1_counter--;
    
    if (pulse2_counter)
	pulse2_counter--;

    if ((!tri_hold) && (tri_counter))
	tri_counter--;

    if (noise_counter)
	noise_counter--;
    
    if (!tri_hold)
    {    
	if (tri_counter2 >= 4)
	    tri_counter2 -= 4;
	else
	    tri_counter2 = 0;
    }
    
    if (tickfuse)
	tickfuse--;

    if ((pulse1_counter) && (pulse1_sweep) && (pulse1_sweepdelta))
    {
	unsigned long f;
	
	f = pulse1_playfreq << pulse1_sweepdelta;
	if (pulse1_sweepdir)
	    f -= pulse1_playfreq;
	else
	    f += pulse1_playfreq;
	
	f >>= pulse1_sweepdelta;
	
	if ((f > 2047) || (f < 8))
	{
	    pulse1_sweep = 0;
	}
	
	pulse1_playfreq = f;
    }
    
    if ((pulse2_counter) && (pulse2_sweep) && (pulse2_sweepdelta))
    {
	unsigned long f;
	
	f = pulse2_playfreq << pulse2_sweepdelta;
	if (pulse2_sweepdir)
	    f -= pulse2_playfreq;
	else
	    f += pulse2_playfreq;
	
	f >>= pulse2_sweepdelta;
	
	if ((f > 2047) || (f < 8))
	{
	    pulse2_sweep = 0;
	}
	
	pulse2_playfreq = f;
    }
    
    if (!tickfuse)
    {

	tickfuse = 60;
    }
    
    numevents = 0;
}
