#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "rom.h"
#include "memmap.h"
#include "ppu.h"

static unsigned char readbuffer;
static unsigned char PPU_control_1;
static unsigned char PPU_control_2;
static unsigned char PPU_status;
static unsigned char PPU_SPR_addr;
static unsigned short PPU_tmpscroll;
static unsigned char PPU_xfine;
static unsigned short PPU_VRAM_addr;
static unsigned short addsave;
static unsigned char PPU_ff;
static unsigned char PPU_hscroll;
static unsigned char PPU_vscroll;
static unsigned char table;
static unsigned char state;

#define PPU_STATE_INACTIVE 0
#define PPU_STATE_DRAWING 1
#define PPU_STATE_VBLANK 2

unsigned char SPR_ram[0x100];

extern unsigned short *framebuffer;

void (*Expandline)(unsigned char line, unsigned char *data);
void updateimage();
void updatedebug();
void updatedebug2();

extern unsigned short *debugbuffer;
extern int dbytes_per_line;
extern unsigned short *debugbuffer2;
extern int d2bytes_per_line;
extern unsigned short palconv[68];

unsigned int debugdirtyflags = 0xFFFFFFFF;
unsigned int maindirty = 1;
unsigned int debugdirty = 1;
unsigned int debug2dirty = 1;

typedef struct
{
    unsigned char sy;
    unsigned char index;
    struct
    {
	unsigned char att : 2;
	unsigned char szero : 1;
	unsigned char     : 2;
	unsigned char b : 1;
	unsigned char hflip : 1;
	unsigned char vflip : 1;
    } f;
    unsigned char sx;
} sprite;

typedef struct
{
    unsigned char x;
    unsigned char tile0;
    unsigned char tile1;
    unsigned char att;
    struct
    {
	unsigned char b : 1;
	unsigned char v : 1;
	unsigned char h : 1;
    } f;
} scache;

sprite *sprites = (sprite *)SPR_ram;

unsigned long frames = 0;
unsigned long usec = 0;
unsigned long sec = 0;

unsigned char nbuf[256];

void Sync()
{
    static struct timeval next1 = {0};
    struct timeval now;
    
    while (gettimeofday (&now, NULL) < 0) ;
    if (next1.tv_sec == 0)
    {
	next1 = now;
	next1.tv_usec++;
    }

    if (timercmp(&next1, &now, >))
    {
	do
	{
	    while (gettimeofday(&now, NULL) < 0) ;
	} while (timercmp(&next1, &now, >));
    }
    else
    {
//	printf("SKIP NEEDED\n");
/*	Expandline(0, nbuf);
	Expandline(1, nbuf);
	Expandline(2, nbuf);
	Expandline(3, nbuf);
	Expandline(4, nbuf);
	Expandline(5, nbuf);
	Expandline(6, nbuf);
	Expandline(7, nbuf);*/
	next1 = now;
    }
    next1.tv_usec += FRAMETIME;

    if (next1.tv_usec >= 1000000)
    {
	next1.tv_sec += next1.tv_usec / 1000000;
	next1.tv_usec %= 1000000;
    }
}

void BlitTile(unsigned char *tilebuf, unsigned short x, unsigned short y)
{
    unsigned short *lineptr;
    int i, j;
    
    for (i = 0; i < 8; i++)
    {
	lineptr = (unsigned short *)((unsigned long)debugbuffer + (y+i)*dbytes_per_line + x*2);
	for (j = 0; j < 8; j++)
	    lineptr[j] = palconv[tilebuf[i*8 + j]];
    }
}

void BlitTile2(unsigned char *tilebuf, unsigned short x, unsigned short y)
{
    unsigned short *lineptr;
    int i, j;
    
    for (i = 0; i < 8; i++)
    {
	lineptr = (unsigned short *)((unsigned long)debugbuffer2 + (y+i)*d2bytes_per_line + x*2);
	for (j = 0; j < 8; j++)
	    lineptr[j] = palconv[tilebuf[i*8 + j]];
    }
}

void RenderTile(unsigned char tile, unsigned char attrib, unsigned char vrom,
		unsigned char pal, unsigned char *buf)
{
    int i, j;
    unsigned char pallete[16];
    unsigned char romline[2];
    unsigned short tilebase;
    unsigned char mask = 0x80;
    
    if (pal)
	for (i = 0; i < 16; i++)
	    pallete[i] = Read_PPU(0x3F10 + i);
    else
	for (i = 0; i < 16; i++)
	    pallete[i] = Read_PPU(0x3F00 + i);

    tilebase = (tile << 4) + (vrom << 12);
    
    memset(buf, 64, 64);
    for (i = 0; i < 8; i++)
    {
	romline[0] = Read_PPU(tilebase + i);
	romline[1] = Read_PPU(tilebase + i + 8);
	mask = 0x80;

	for (j = 0; j < 8; j++)
	{
	    if (romline[0] & mask)
		buf[i*8+j] |= 1;
	    if (romline[1] & mask)
		buf[i*8+j] |= 2;
	    buf[i*8+j] |= attrib << 2;
	    buf[i*8+j] = pallete[buf[i*8+j]];
	    mask >>= 1;
	}
    }
}

extern unsigned char *nt1;
extern unsigned char *nt2;
extern unsigned char *at1;
extern unsigned char *at2;

void ProcessDirty()
{
    unsigned long redraw = 0;

    if (debugdirtyflags & DIRTY_PAT0)
    {
	redraw |= REDRAW_PAT0;
	if (~PPU_control_1 & CON1_BPAT)
	    redraw |= REDRAW_NT;
    }
    if (debugdirtyflags & DIRTY_PAT1)
    {
	if (PPU_control_1 & CON1_BPAT)
	    redraw |= REDRAW_NT;
	redraw |= REDRAW_PAT1;
    }
    if (debugdirtyflags & DIRTY_NT0)
	redraw |= REDRAW_NT0;
    if (debugdirtyflags & DIRTY_NT1)
	redraw |= REDRAW_NT1;
    if (debugdirtyflags & DIRTY_NT2)
	redraw |= REDRAW_NT2;
    if (debugdirtyflags & DIRTY_NT3)
	redraw |= REDRAW_NT3;
    if (debugdirtyflags & DIRTY_AT0)
	redraw |= REDRAW_AT0;
    if (debugdirtyflags & DIRTY_AT1)
	redraw |= REDRAW_AT1;
    if (debugdirtyflags & DIRTY_AT2)
	redraw |= REDRAW_AT2;
    if (debugdirtyflags & DIRTY_AT3)
	redraw |= REDRAW_AT3;
    if (debugdirtyflags & DIRTY_PAL0)
	redraw |= REDRAW_PAL0;
    if (debugdirtyflags & DIRTY_PAL1)
	redraw |= REDRAW_PAL1;
    debugdirtyflags = redraw;
}

void PPU_Drawpal()
{
    unsigned char buf[256];
    int i = 0;
    
    while (i < 256)
    {
	buf[i] = palram[i>>4];
	i++;
    }

    for (i = 0; i < 8; i++)
	Expandline(i, buf);
}

void RenderDebug()
{
    int i, j, k;
    unsigned short x, y;
    unsigned int patmask = REDRAW_PAT0_0;
    unsigned char tilebuf[64];

    ProcessDirty();
    
    x = 0;
    y = 0;
    for (k = 0; k < 2; k++)
    {
	for (j = 0; j < 4; j++)
	{
	    if (debugdirtyflags & (patmask << j))
	    {
		x = 0;
		y = 0;
		for (i = 0; i < 256; i++)
		{
		    RenderTile(i, j, k, 0, tilebuf);
		    BlitTile(tilebuf, x*8 + j*128, y*8 + k*128);
		    x++;
		    if (x == 16)
		    {
			y++;
			x = 0;
		    }
		}
		debugdirty++;
	    }
	}
	patmask <<= 4;
    }


    for (i = 0; i < 4; i++)
    {
	if (debugdirtyflags & (REDRAW_NT0 << i))
	{
	    for (y = 0; y < 30; y++)
	    {
		for (x = 0; x < 32; x++)
		{
		    unsigned char attindex = ((y << 1) & 0xF8) + (x >> 2);
		    unsigned char attshift = ((y & 0x2) | ((x & 0x2) >> 1)) << 1;
		    
		    RenderTile(nametable[i][y*32 + x],
			       (attribtable[i][attindex] & (0x3 << attshift)) >> attshift,
			       (PPU_control_1 & CON1_BPAT) >> 4, 0, tilebuf);
		    BlitTile2(tilebuf, (256*(i&1)) + x*8, (120*(i&2)) + y*8);
		}
	    }
	    debug2dirty++;
	}
    }

    debugdirtyflags = 0;
    
    /*
    memset(tilebuf, 0, 64);
    BlitTile(tilebuf, 0, 256);
    memset(tilebuf, 5, 64);
    BlitTile(tilebuf, 8, 256);
    memset(tilebuf, 1, 64);
    BlitTile(tilebuf, 16, 256);
    memset(tilebuf, 9, 64);
    BlitTile(tilebuf, 24, 256);
    memset(tilebuf, 0, 64);
    BlitTile(tilebuf, 32, 256);
    memset(tilebuf, 21, 64);
    BlitTile(tilebuf, 40, 256);
    memset(tilebuf, 17, 64);
    BlitTile(tilebuf, 48, 256);
    memset(tilebuf, 25, 64);
    BlitTile(tilebuf, 56, 256);
    memset(tilebuf, 0, 64);
    BlitTile(tilebuf, 64, 256);
    memset(tilebuf, 37, 64);
    BlitTile(tilebuf, 72, 256);
    memset(tilebuf, 33, 64);
    BlitTile(tilebuf, 80, 256);
    memset(tilebuf, 41, 64);
    BlitTile(tilebuf, 88, 256);
    memset(tilebuf, 0, 64);
    BlitTile(tilebuf, 96, 256);
    memset(tilebuf, 53, 64);
    BlitTile(tilebuf, 104, 256);
    memset(tilebuf, 49, 64);
    BlitTile(tilebuf, 112, 256);
    memset(tilebuf, 58, 64);
    BlitTile(tilebuf, 120, 256);*/
}

unsigned char Read_PPU_Register(unsigned short addr)
{
    unsigned char ret;
    switch (addr)
    {
    case 0x2000:
    case 0x2001:
    case 0x2002:
    case 0x2003:
    case 0x2005:
    case 0x2006:
	ret = PPU_status;
	PPU_status &= !STAT_VBLANK;
	PPU_ff = 0;
	return ret;
    case 0x2004: return SPR_ram[PPU_SPR_addr++];
    case 0x2007:
	if (state == PPU_STATE_DRAWING)
	    printf("Read from $2007 during frame\n");
	ret = readbuffer;
	readbuffer = Read_PPU(PPU_VRAM_addr);

	if (PPU_control_1 & CON1_PPUINC)
	    PPU_VRAM_addr += 32;
	else
	    PPU_VRAM_addr++;
	return ret;
    default:
	printf("Read from register $%x\n", addr);
    }
    return 0;
}

void Write_PPU_Register(unsigned short addr, unsigned char value)
{
    unsigned short old_addr;
    int i;
    
    switch (addr)
    {
    case 0x2000:
	PPU_control_1 = value;
	PPU_tmpscroll &= 0x73FF;
	PPU_tmpscroll |= (value & 3) << 10;
	break;
    case 0x2001:
	PPU_control_2 = value;
	break;
    case 0x2003:
	PPU_SPR_addr = value;
	break;
    case 0x2004:
	SPR_ram[PPU_SPR_addr++] = value;
	break;
    case 0x2005:	
	if (PPU_ff)
	{
	    PPU_vscroll = value;
	    PPU_tmpscroll &= 0x0C1F;
	    PPU_tmpscroll |= (value & 7) << 12;
	    PPU_tmpscroll |= (value << 2) & 0x03E0;
	    PPU_ff = 0;
	}
	else
	{
	    PPU_tmpscroll &= 0x7FE0;
	    PPU_tmpscroll |= value >> 3;
	    PPU_xfine = value & 7;
	    PPU_ff = 1;
	}
	break;	
    case 0x2006:
	if (PPU_ff)
	{
	    PPU_tmpscroll &= 0xFF00;
	    PPU_tmpscroll |= value;
	    PPU_VRAM_addr = PPU_tmpscroll;
	    PPU_ff = 0;
	}
	else
	{
	    addsave = value;
	    PPU_tmpscroll &= 0x00FF;
	    PPU_tmpscroll |= (value & 0x3F) << 8;
	    PPU_ff = 1;
	}
	break;
    case 0x2007:
	Write_PPU(PPU_VRAM_addr, value);
	
	if (PPU_control_1 & CON1_PPUINC)
	    PPU_VRAM_addr += 32;
	else
	    PPU_VRAM_addr++;
	break;
    case 0x4014:
	old_addr = value << 8;
	for (i = 0; i < 256; i++)
	    SPR_ram[i] = Read_CPU(old_addr + i);
	break;
    }
}

void PPU_Start_Vblank()
{
    Sync(); // comment out for profiling and speed hacking
    frames++;
//    if (maindirty)
	updateimage();

    PPU_status |= STAT_VBLANK;
    state = PPU_STATE_VBLANK;
}

unsigned char PPU_Check_NMI()
{
    return (PPU_control_1 & CON1_NMI);
}

void PPU_End_Vblank()
{
    PPU_status &= ~STAT_HIT;
    // Try slightly modified loopyscroll, with bigfootguy-juju
    PPU_VRAM_addr = PPU_tmpscroll;
}

/*
void DecodeLine(unsigned char *buf, unsigned char romline[2],
		unsigned char attrib)
{
    unsigned char i;
    unsigned char mask = 0x80;
    
    for (i = 8; i > 0; i--)
    {
	*buf = attrib;
	if (romline[0] & mask)
	    *buf |= 1;
	if (romline[1] & mask)
	    *buf |= 2;
	buf++;
	mask >>= 1;
    }
}
*/


extern unsigned char *vrom;
unsigned char debugprint = 0;

void TransBackBlit(unsigned char *src, unsigned char *dst, unsigned char ofs)
{
    while (ofs--)
    {
	if (*src & 3)
	    *dst = palram[*src];
	dst++;
	src++;
    }
}

void TransBlitLineH(unsigned char *src, unsigned char *dst, unsigned char ofs)
{
    src += ofs - 1;
    while (ofs--)
    {
	if (*src & 3)
	    *dst = sprpal[*src];
	dst++;
	src--;
    }
}

void TransBlitLine(unsigned char *src, unsigned char *dst, unsigned char ofs)
{
    while (ofs--)
    {
	if (*src & 3)
	    *dst = sprpal[*src];
	dst++;
	src++;
    }
}

void BlitLine8(unsigned char *src, unsigned char *dst,
	       unsigned char transmask)
{
    if (transmask & 0x80)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x40)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x20)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x10)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x08)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x04)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x02)
	*dst = palram[*src];
    dst++; src++;
    if (transmask & 0x01)
	*dst = palram[*src];
    dst++; src++;
}

void BlitLine(unsigned char *src, unsigned char *dst, unsigned char ofs,
	      unsigned char transmask)
{
    switch (ofs)
    {
    case 8:
	if (transmask & 0x80)
	    *dst = palram[*src];
	dst++; src++;
    case 7:
	if (transmask & 0x40)
	    *dst = palram[*src];
	dst++; src++;
    case 6:
	if (transmask & 0x20)
	    *dst = palram[*src];
	dst++; src++;
    case 5:
	if (transmask & 0x10)
	    *dst = palram[*src];
	dst++; src++;
    case 4:
	if (transmask & 0x08)	
	    *dst = palram[*src];
	dst++; src++;
    case 3:
	if (transmask & 0x04)	
	    *dst = palram[*src];
	dst++; src++;
    case 2:
	if (transmask & 0x02)	
	    *dst = palram[*src];
	dst++; src++;
    case 1:
	if (transmask & 0x01)
	    *dst = palram[*src];
	dst++; src++;
    }
}

void BlitLineI(unsigned char *src, unsigned char *dst, unsigned char ofs,
	       unsigned char transmask)
{
    switch (ofs)
    {
    case 1:
	if (transmask & 0x80)
	    *dst = palram[*src];
	dst++; src++;
    case 2:
	if (transmask & 0x40)
	    *dst = palram[*src];
	dst++; src++;
    case 3:
	if (transmask & 0x20)
	    *dst = palram[*src];
	dst++; src++;
    case 4:
	if (transmask & 0x10)
	    *dst = palram[*src];
	dst++; src++;
    case 5:
	if (transmask & 0x08)
	    *dst = palram[*src];
	dst++; src++;
    case 6:
	if (transmask & 0x04)
	    *dst = palram[*src];
	dst++; src++;
    case 7:
	if (transmask & 0x02)
	    *dst = palram[*src];
	dst++; src++;
    case 8:
	if (transmask & 0x01)
	    *dst = palram[*src];
	dst++; src++;
    }
}

void DecodeLine(unsigned char tile0, unsigned char tile1, unsigned char att,
		unsigned char *buf)
{
    unsigned char mask = 0x80;
    unsigned char b;
    
    while (mask)
    {
	b = att;
	if (tile0 & mask)
	    b |= 1;
	if (tile1 & mask)
	    b |= 2;
	*buf++ = b;
	mask >>= 1;
    }
}

int RenderTileLine(unsigned char tile, unsigned char line, unsigned char att,
		    unsigned short bank, unsigned char *buf)
{
    unsigned short tilebase = (tile << 4) + bank + line;
    unsigned char draw = 0;
    unsigned char romline0, romline1;
    unsigned char mask = 0x80;
    
    romline0 = Read_VROM(tilebase);
    romline1 = Read_VROM(tilebase + 8);

    draw = (romline0 | romline1);

    memset(buf, att, 8);
    
    while (mask)
    {
	if (romline0 & mask)
	    *buf |= 1;
	if (romline1 & mask)
	    *buf |= 2;

	buf++;
	mask >>= 1;
    }
    return draw;
}

unsigned char obuf[256];
unsigned char backbuf[8][8];

void PPU_Drawsprites(unsigned char line, unsigned char *buf,
		     sprite *sprlist, unsigned char numsprites,
		     unsigned char sh)
{
    unsigned char ofs;
    unsigned char len;
    unsigned char *pbuf;
    unsigned char d;
    unsigned char tilebuf[8];
    unsigned short bank = 0; // Shut up GCC, I know what I'm doin
    unsigned char n = numsprites;
    sprite *sp = sprlist;

    if (sh == 8)
	bank = (PPU_control_1 & CON1_SPAT) << 9;
	
    while (n)
    {
	if (sp->f.b)
	    memcpy(backbuf[n], obuf + sp->sx, 8);
	sp++;
	n--;
    }
    
    while (numsprites)
    {
	pbuf = buf + sprlist->sx;
	ofs = (line - (sprlist->sy + 1)) & (sh - 1);

	if (sprlist->f.vflip)
	    ofs = (sh - 1 - ofs);
	
	if (sh == 16)
	{
	    bank = (sprlist->index & 1) << 12;
	    if (ofs & 8)
		sprlist->index |= 1;
	    else
		sprlist->index &= 0xFE;
	    ofs &= 7;
	}
	
	memset(tilebuf, 0, 8);
	d = RenderTileLine(sprlist->index, ofs, (sprlist->f.att << 2),
			   bank,
			   tilebuf);

	if (d)
	{
	    if (sprlist->f.szero)
		PPU_status |= STAT_HIT;
	    if (sprlist->sx + 8 > 256)
	    {
		len = 256 - sprlist->sx;
	    }
	    else
		len = 8;
	    
	    if (sprlist->f.hflip)
		TransBlitLineH(tilebuf, pbuf, len);
	    else
		TransBlitLine(tilebuf, pbuf, len);
	    if (sprlist->f.b)
		TransBackBlit(backbuf[numsprites], pbuf, len);
	}
	numsprites--;
	sprlist++;
    }
}

void PPU_Drawtiles(unsigned char line, unsigned char *buf)
{
    unsigned char tilex, tiley, tofs, tcount, *at, *nt, tilebuf[8];
    unsigned char att, tile, ofs, attbase, attshift, tbl = table;
    unsigned short namebase;
    unsigned char *sv = obuf, tmask;
    
    tilex = PPU_hscroll >> 3;
    
    tcount = 32;
    
    if ((tofs = (PPU_hscroll & 7)) != 0)
	tcount++;
    
    tiley = (line + PPU_vscroll) >> 3;
    ofs = (line + PPU_vscroll) & 7;
    
    if (tiley >= 30)
    {
	tbl ^= 2;
	tiley -= 30;
    }

    at = attribtable[tbl];
    nt = nametable[tbl];

    attbase = ((tiley << 1) & 0xF8);
    namebase = (tiley << 5) + tilex;
    
    while (tcount)
    {
	if (tilex & 0x20)
	{
	    tbl ^= 1;
	    at = attribtable[tbl];
	    nt = nametable[tbl];
	    tilex = 0;
	    namebase = tiley << 5;
	}

	memset(tilebuf, 0, 8);
	attshift = ((tiley << 1) & 4) | (tilex & 2);
	att = ((at[attbase + (tilex >> 2)] >> attshift) & 3) << 2;

	tile = nt[namebase++];

	tmask = RenderTileLine(tile, ofs, att,
			       (PPU_control_1 & CON1_BPAT) << 8, tilebuf);

	if ((tcount == 33) && (tofs))
	{
	    memcpy(sv, tilebuf + tofs, 8 - tofs);
	    BlitLine(tilebuf + tofs, buf, 8 - tofs, tmask);
	    buf += 8 - tofs;
	    sv += 8 - tofs;
	}
	else if ((tcount == 1) && (tofs))
	{
	    memcpy(sv, tilebuf, tofs);
	    BlitLine(tilebuf, buf, tofs, tmask);
	    buf += tofs;
	    sv += tofs;
	}
	else
	{
	    memcpy(sv, tilebuf, 8);
	    BlitLine(tilebuf, buf, 8, tmask);
	    buf += 8;
	    sv += 8;
	}

	tilex++;
	tcount--;
    }
}

void PPU_DrawBG(unsigned char *buf)
{
    unsigned char tbl = (PPU_VRAM_addr & 0x0C00) >> 10, tcount, tilebuf[8];
    unsigned char *sv = obuf;
    unsigned char ofs = (PPU_VRAM_addr >> 12), att, tile, *at, *nt;
    unsigned short bank = (PPU_control_1 & CON1_BPAT) << 8;
    unsigned char eminusxf = 8 - PPU_xfine;
    unsigned char attshift = (PPU_VRAM_addr & 0x40) >> 4;
    unsigned char tilex = PPU_VRAM_addr & 0x1F;
    unsigned char tiley = (PPU_VRAM_addr >> 5) & 0x1F;
    unsigned char attbase;
    unsigned char tmask;
    
    tcount = 32;

    if (PPU_xfine)
	tcount++;

    at = attribtable[tbl];
    nt = nametable[tbl];

    attbase = ((tiley << 1) & 0xF8);
    
    while (tcount)
    {
	tile = nt[(PPU_VRAM_addr & 0x3E0) + tilex];
	attshift = ((tiley << 1) & 4) | (tilex & 2);
	att = ((at[attbase + (tilex >> 2)] >> attshift) & 3) << 2;

	tmask = RenderTileLine(tile, ofs, att, bank, tilebuf);

	if ((tcount == 33)&&(PPU_xfine))
	{
	    memcpy(sv, tilebuf + PPU_xfine, eminusxf);
	    BlitLineI(tilebuf + PPU_xfine, buf, PPU_xfine, tmask);
	    buf += eminusxf;
	    sv += eminusxf;
	}
	else if ((tcount == 1)&&(PPU_xfine))
	{
	    memcpy(sv, tilebuf, PPU_xfine);
	    BlitLine(tilebuf, buf, PPU_xfine, tmask);
	    buf += PPU_xfine;
	    sv += PPU_xfine;
	}
	else
	{
	    memcpy(sv, tilebuf, 8);
	    BlitLine8(tilebuf, buf, tmask);
	    buf += 8;
	    sv += 8;
	}
	
	tcount--;

	tilex++;
	if (tilex > 31)
	{
	    tilex = 0;
	    nt = nametable[tbl ^ 1];
	    at = attribtable[tbl ^ 1];
	}
    }
}

void PPU_NewRenderline(unsigned char line)
{
    int i, sh;
    unsigned char buf[256];
    sprite sprlist[8];
    sprite *spr = sprites;
    sprite *sp = sprlist;
    int numsprites = 8;

//    if (~PPU_control_2 & CON2_BDRAW)
    memset(buf, palram[0], 256);

    if (PPU_control_2 & (CON2_BDRAW | CON2_SDRAW))
	state = PPU_STATE_DRAWING;
    else
	state = 0;
	
    if (PPU_control_1 & CON1_SPRSIZE)
	sh = 16;
    else
	sh = 8;

    spr[0].f.szero = 1;
    
    for (i = 64; i; i--)
    {
	if ((spr->sy < line) && (spr->sy + sh >= line))
	{
	    *sp++ = *spr;
	    numsprites--;
	}
	spr++;
	if (!numsprites)
	    break;
    }

    if (PPU_control_2 & (CON2_BDRAW | CON2_SDRAW))
    {
	PPU_VRAM_addr &= 0x7BE0;
	PPU_VRAM_addr |= PPU_tmpscroll & 0x041F;
    }
    
    if (!numsprites)
	PPU_status |= STAT_SCOUNT;
    
//    if (PPU_control_2 & CON2_BDRAW)
//	PPU_Drawtiles(line, buf);
//	PPU_DrawBG(buf);
    
    numsprites = 8 - numsprites;

    
    if ((PPU_control_2 & CON2_SDRAW)&&(numsprites))
	PPU_Drawsprites(line, buf, sprlist, numsprites, sh);
//    PPU_status |= STAT_HIT;
    
    Expandline(line, buf);

    if (PPU_control_2 & (CON2_SDRAW | CON2_BDRAW))
    {
	PPU_VRAM_addr += 0x1000;
	if (PPU_VRAM_addr & 0x8000)
	{
	    unsigned short tmp;
	    tmp = (PPU_VRAM_addr & 0x03E0) + 0x20;
	    PPU_VRAM_addr &= 0x7C1F;
	    if (tmp == 0x3C0)
		PPU_VRAM_addr ^= 0x800;
	    else
		PPU_VRAM_addr |= (tmp & 0x03E0);
	}
    }
}

void PPU_OldRenderline(unsigned char line)
{
/*    unsigned char ofs;
    unsigned char *at;
    unsigned char *nt;
    unsigned char tilebuf[8];
    unsigned char tbl = table;
    unsigned char tiley, tilex;
    unsigned char tcount;
    unsigned char tofs;
    unsigned char att;
    unsigned char tile;
    unsigned char attbase, attshift;
    unsigned short namebase;
    unsigned char *p = obuf;
    int i = 256;

    while (i--) // DAMN YOSHI
	*p++ = palram[0];
         // 64 + ((PPU_control_2 & CON2_BGMASK) >> CON2_BGSHIFT);
    p = obuf;
    
    PPU_status |= STAT_HIT;

    if (PPU_control_2 & CON2_SDRAW)
    {
	i = 64;
	scount = 8;
	while ((i--)&&(scount))
	{
	    if (sprites->sy > line)
	    {
		sprites++;
		continue;
	    }
	    if (sprites->sy + sh <= line)
	    {
		sprites++;
		continue;
	    }
	    
	    ofs = line - sprites->sy;
	    namebase = (sprites->index << 4);
	    if (PPU_control_1 & CON1_SPRSIZE)
	    {
		if (ofs > 7)
		{
		    ofs -= 8;
		    namebase += 0x1000;
		}
	    }
	    else if (PPU_control_1 & CON1_SPAT)
		namebase += 0x1000;
	    
	    sp->tile0 = vrom[namebase + ofs];
	    sp->tile1 = vrom[namebase + ofs + 8];
	    if (sp->tile0 | sp->tile1)
	    {
		sp->x = sprites->sx;
		sp->att = (sprites->f.att << 2) | 0x10;
		sp->f.b = sprites->f.background;
		sp->f.h = sprites->f.hflip;
		sp->f.v = sprites->f.vflip;
		sp++;
		scount--;
	    }	    
	}
	
	scount = 8 - scount;
	i = scount;
	sp = cache;
	
	while (i--)
	{
	    if (sp->f.b)
	    {
		DecodeLine(sp->tile0, sp->tile1, sp->att, tilebuf);
		BlitLine(tilebuf, obuf + sp->x, 8);
		}
	    sp++;
//	    scount--;
	}
	p = obuf;
    }

    if (PPU_control_2 & CON2_BDRAW)
    {
	tilex = PPU_hscroll >> 3;

	tcount = 32;
	if (tofs == (PPU_hscroll & 7))
	    tcount++;
	tiley = (line + PPU_vscroll) >> 3;
	ofs = (line + PPU_vscroll) & 0x7;

	if (~PPU_control_2 & CON2_BCLIP)
	{
	    tilex++;
	    linebuf += 8;
	    if (tcount == 32)
		tcount--;
		}
	
	if (tiley >= 30)
	{
	    tbl ^= 2;
	    tiley -= 30;
	}
	at = attribtable[tbl];
	nt = nametable[tbl];
	attbase = ((tiley << 1) & 0xF8);
	namebase = (tiley << 5) + tilex;
	
	while (tcount)
	{
	    if (tilex & 0x20)
	    {
		tbl ^= 1;
		at = attribtable[tbl];
		nt = nametable[tbl];
		tilex = 0;
		namebase = tiley << 5;
	    }
	    memset(tilebuf, 0, 8);
	    attshift = ((tiley << 1) & 0x4) | (tilex & 0x2);
	    att = ((at[attbase + (tilex >> 2)] >> attshift) & 0x3) << 2;
	    
	    tile = nt[namebase++];
	    
	    RenderTileLine(tile, ofs, att, (PPU_control_1 & CON1_BPAT) >> 4, tilebuf);
	    if ((tcount == 33) && (tofs))
	    {

		tcount--;
		BlitLine(tilebuf + tofs, p, 8 - tofs);
		p += 8 - tofs;
	    }
	    else if ((tcount == 1) && (tofs))
	    {
		BlitLine(tilebuf, p, tofs);
		p += tofs;
	    }
	    else
	    {
		BlitLine(tilebuf, p, 8);
		p += 8;
	    }
	    tilex++;
	    tcount--;
	}
	p = obuf;
    }

    if (PPU_control_2 & CON2_SDRAW)
    {
	sp = cache;
	i = scount;
	
	while (i--)
	{
	    if (!sp->f.b)
	    {
		DecodeLine(sp->tile0, sp->tile1, sp->att, tilebuf);
		BlitLine(tilebuf, p + sp->x, 8);
	    }
	    sp++;
	}
	p = obuf;
    }

    Expandline(line, obuf);*/
}

void Init_PPU(void)
{
    int i;
    for (i = 0; i < 4; i++)
    {
	memset(nametable[i], 0, 0x3C0);
	memset(attribtable[i], 0, 0x40);
    }
    memset(nbuf, 64, 256);
    memset(SPR_ram, 0, 256);    
}
