; s6502.asm
;
; Stanknes 6502 core
;
; Expects 00000000h-0000FFFFh to be mapped appropriately
; doesn't support running code in $2-5xxx correctly, will
; most likely result in a protection fault unless they are
; mapped in

; Calling code must also make sure the page at 00010000h is
; mapped, as it will be used for context saving and possibly
; temp storage

; Fetch opcode byte into AL, jump to corebase + AL << log2(opsize)
; opsize is the max. number of bytes for an instruction.
; should be an even multiple of 32 for efficient cache usage

; see 6502corenotes for more

; Execute_6502(unsigned long cycles);
;
; Executes <cycles> 6502 cycles, under some circumstances it 
; might execute up to 6 extra cycles

Execute_6502:

mov ecx, [esp + 4]

coreloop:

; FIXME check for interrupts of any form

test esi, 0E000h
jz rammirror
movzx eax, byte [esi]
inc esi
shl eax, TABLE_SHIFT
add eax, table_base
jmp eax

rammirror:

and esi, 07FFh
movzx eax, byte [esi]
inc esi
shl eax, TABLE_SHIFT
add eax, table_base
jmp eax

%macro PACK_FLAGS 0         ; expects tmp flags in bh, x86 flags in ah
;*****************
mov  dh,    ah              ; 2
shl  dh,    1               ; 2
rcr  bh,    3               ; 3
shr  edx,   2               ; 3
shr  dh,    6               ; 2
shr  edx,   1               ; 2
xchg bh,    dh              ; 2
shr  edx,   6               ; 3
inc  bh                     ; 2  -- 21 bytes
;*****************
%endmacro

table_base:

; 00 BRK
_brk:
inc   esi                   ; 2
mov   bh,  [esp + 4]        ; 3
or    bh,  20h              ; 3
PACK_FLAGS                  ; 21
dec   bl                    ; 2
mov   [ebx], esi            ; 2
dec   bl                    ; 2
mov   [ebx], dl             ; 2
dec   bl                    ; 2
or    [esp + 4], 00100000b  ; 4
mov   esi, [0FFFEh]         ; 6
sub   ebp, 7                ; 3
jmp   end_table             ; 5
                            ; 57

; 01 ORA (Indirect, X)
_ora_indx:

xor   edx, edx              ; 2
xchg  ebx, edi              ; 2
mov   dl,  [esi]            ; 2
inc   esi                   ; 2
add   dl,  cl               ; 2
mov   bl,  [edx]            ; 2
inc   dl                    ; 2
mov   bh,  [edx]            ; 2
xchg  ebx, edi              ; 2
mov   dl,  [edi]            ; 2
sahf                        ; 1
or    al,  dl               ; 2
lahf                        ; 1
sub   ebp, 6                ; 3
jmp   end_table             ; 5
                            ; 32
; 02 nop

; 03 nop

; 04 nop

; 05 ORA Zero Page

_ora_zp:
movzx edi, byte [esi]       ; 2
inc   esi                   ; 2
mov   dl,  [edi]            ; 2
sahf                        ; 1
or    al, dl                ; 2
lahf                        ; 1
sub   ebp, 3                ; 3
jmp end_table               ; 5
                            ; 18

; 06 ASL Zero Page

_asl_zp:
movzx edi, byte [esi]       ; 3
inc   esi                   ; 2
mov   dl,  [edi]            ; 2
sahf                        ; 1
shl   dl,  1                ; 2
lahf                        ; 1
mov   [edi], dl             ; 2
sub   ebp, 5                ; 3
jmp end_table               ; 5
                            ; 21

; 07 nop

; 08 PHP

_php:
mov   bh,  [esp + 4]        ; 3
PACK_FLAGS                  ; 21
mov   [ebx], dl             ; 2
dec   bl                    ; 2
sub   ebp, 3                ; 3
jmp end_table               ; 5
                            ; 37

; 09 ORA Immediate

_ora_imm:
mov   dl,  [esi]            ; 2
inc   esi                   ; 2
sahf                        ; 1
or    al, dl                ; 2
lahf                        ; 1
sub   ebp, 3                ; 3
jmp end_table               ; 5
                            ; 16

; 0A ASL Accumulator

_asl_accum:
sahf                        ; 1
shl   al, 1                 ; 2
lahf                        ; 1
sub   ebp, 2                ; 3
jmp end_table               ; 5
                            ; 11

; 0B nop

; 0C nop

; 0D nop

; 0E nop

; 0F nop

; 10 BPL

_bpl:
mov    edi, 2               ; 3
xor    edx, edx             ; 2
sahf                        ; 1
jns    .branch              ; 2

.endbpl
xchg  esi, ebx              ; 2
sub   ebp, edi              ; 2
jmp end_table               ; 5

.branch:
inc   edi                   ; 2
mov   dl,  byte [esi]       ; 2
inc   esi                   ; 2
xchg  esi, ebx              ; 2
add   bl,  dl               ; 2
jc    short .pagechange     ; 2
jmp   short .endbpl         ; 2

.pagechange:
inc   edi                   ; 2
inc   bh                    ; 2
jmp   short .endbpl         ; 2
                            ; 37

; 11 ORA (Indirect), Y

_ora_indy:
xor   edx, edx              ; 2
xchg  ebx, edi              ; 2
mov   dl, [esi]             ; 2
inc   esi                   ; 2
mov   bl, [edx]             ; 2
inc   dl                    ; 2
mov   bh, [edx]             ; 2
sub   ebp, 5                ; 3
add   bl, ch                ; 2
jc short .pagecross         ; 2

.end_ora_indy:
xchg  ebx, edi              ; 2
mov   dl, [edi]             ; 2
sahf                        ; 1
or    al, dl                ; 2
lahf                        ; 1
jmp end_table               ; 5

.pagecross:
dec   ebp                   ; 2
inc   bh                    ; 2
jmp short .end_ora_indy     ; 2
                            ; 40

; 12 nop

; 13 nop

; 14 nop

; 15 ORA ZP, X

_ora_zpx:
xor   edx, edx              ; 2
mov   dl, [esi]             ; 2
inc   esi                   ; 2
add   dl, ch                ; 2
mov   dl, [edx]             ; 2
sahf                        ; 1
or    al, dl                ; 2
lahf                        ; 1
sub   ebp, 4                ; 3
jmp end_table               ; 5
                            ; 22

; 16 ASL ZP, X

_asl_zpx:
xor   edx, edx              ; 2
mov   dl, [esi]             ; 2
inc   esi                   ; 2
add   dl, ch                ; 2
xchg  edx, edi              ; 2
mov   dl, [edi]             ; 2
sahf                        ; 1
shl   dl, 1                 ; 2
lahf                        ; 1
mov   [edi], dl             ; 2
sub   ebp, 6                ; 3
jmp end_table               ; 5
                            ; 26

; 17 nop

; 18 CLC

_clc:
sahf                        ; 1
clc                         ; 1
lahf                        ; 1
sub   ebp, 2                ; 3
jmp end_table               ; 5
                            ; 11

; 19 ORA Abs, Y

_ora_absy:
movzx edx, word [esi]       ; 2
add   esi, 2                ; 3
sub   ebp, 4                ; 3
add   dl,  cl               ; 2
jc short .pagecross         ; 2

.end_ora
xchg  edx, edi              ; 2
mov   dl, [edi]             ; 2
sahf                        ; 1
or    al, dl                ; 2
lahf                        ; 1
jmp end_table               ; 5

.pagecross
dec   ebp                   ; 2
inc   dh                    ; 1
jmp short .end_ora          ; 2
                            ; 30

; 1A nop

; 1B nop

; 1C nop

; 1D ORA Abs, X

_ora_absy:
movzx edx, word [esi]       ; 2
add   esi, 2                ; 3
sub   ebp, 4                ; 3
add   dl,  ch               ; 2
jc short .pagecross         ; 2

.end_ora
xchg  edx, edi              ; 2
mov   dl, [edi]             ; 2
sahf                        ; 1
or    al, dl                ; 2
lahf                        ; 1
jmp end_table               ; 5

.pagecross
dec   ebp                   ; 2
inc   dh                    ; 1
jmp short .end_ora          ; 2
                            ; 30

; 1E ASL Abs, X

_ora_absy:
movzx edx, word [esi]       ; 2
add   esi, 2                ; 3
sub   ebp, 7                ; 3
add   dl,  cl               ; 2
adc   dh,  0                ; 3
xchg  edx, edi              ; 2
mov   dl, [edi]             ; 2
sahf                        ; 1
shl   dl, 1                 ; 2
lahf                        ; 1
mov   [edi], dl             ; 2
jmp end_table               ; 5

; 1F nop

; 20 JSR