#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include "osssound.h"

#define NUMFRAGS 5
#define LOG_2_FRAG 10
#define FRAGSIZE (1<<LOG_2_FRAG)
#define AUDIO_BUFF_SIZE ((NUMFRAGS + 1) * FRAGSIZE)

int pcm_fd;

void Init_OSS(void)
{
    int tmp;
    
    printf("Opening PCM device...\n");
    pcm_fd = open("/dev/dsp", O_WRONLY | O_NONBLOCK);
    if (pcm_fd < 0)
    {
	perror("open(\"/dev/dsp\")");
	pcm_fd = 0;
	return;
    }
    printf("Setting sample format...\n");
    tmp = AFMT_U8;
    if (ioctl(pcm_fd, SNDCTL_DSP_SETFMT, &tmp) < 0)
    {
	perror("SNDCTL_DSP_SETFMT");
	close(pcm_fd);
	pcm_fd = 0;
	return;
    }
    printf("Setting channels...\n");
    tmp = 0;
    if (ioctl(pcm_fd, SNDCTL_DSP_STEREO, &tmp) < 0)
    {
	perror("SNDCTL_DSP_STEREO");
	close(pcm_fd);
	pcm_fd = 0;
	return;
    }
    printf("Setting sample rate...\n");
    tmp = 44100;
    if (ioctl(pcm_fd, SNDCTL_DSP_SPEED, &tmp) < 0)
    {
	perror("SNDCTL_DSP_SPEED");
	close(pcm_fd);
	pcm_fd = 0;
	return;
    }
    printf("Setting fragment size/count...\n");
    tmp = 0x00080008;
    if (ioctl(pcm_fd, SNDCTL_DSP_SETFRAGMENT, &tmp) < 0)
    {
	perror("SNDCTL_DSP_SETFRAGMENT");
	close(pcm_fd);
	pcm_fd = 0;
	return;
    }
}

void Kill_OSS(void)
{
    if (pcm_fd)
	close(pcm_fd);
}

