#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include <errno.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "n6502.h"

#include "rom.h"
#include "memmap.h"
#include "newppu.h"
#include "luts.h"

#include "cmdline.h"

#include "apu.h"

extern int zerofd;
extern int memfd;

extern unsigned char paldirty;
extern unsigned char showcache;

unsigned char sprram[0x100];
static unsigned char vram_buffer;
static unsigned short vram_addr;
static unsigned char spraddr;
static unsigned char con_1;
static unsigned char con_2;
static unsigned char status;
static unsigned char ff;
static unsigned char xfine;
static unsigned short tmpscroll;
static unsigned char sbuf[288];

unsigned char numsprites = 0;

void (*Expandline)(unsigned char line, unsigned char *data);
void (*Clearline)(unsigned char line);
void (*Updatepal)(unsigned char *pal);
void updateimage();

unsigned long frames = 0;
unsigned long skip = 0;
unsigned char hitline = 0;

typedef struct
{
    unsigned char sy;
    unsigned char index;
    struct
    {
	unsigned char att   : 2;
	unsigned char       : 3;
	unsigned char b     : 1;
	unsigned char hflip : 1;
	unsigned char vflip : 1;
    } f;
    unsigned char sx;
} sprite;

typedef struct
{
    unsigned char *pixels;
    unsigned char hflip;
    unsigned char att;
    unsigned char b;
    unsigned char xofs;
    unsigned char zero;
    unsigned long tile;
} sprent;

sprite *sprites = (sprite *)sprram;
sprent sprcache[8];

unsigned char sff = 0;

void Sync()
{
    static struct timeval next1 = {0};
    struct timeval now;

    while (gettimeofday (&now, NULL) < 0) ;
    if (next1.tv_sec == 0)
    {
	next1 = now;
	next1.tv_usec++;
    }

//    frames++;
    
    if (timercmp(&next1, &now, >))
    {
	do
	{
	    while (gettimeofday(&now, NULL) < 0) ;
	} while (timercmp(&next1, &now, >));
    }
    else
    {
	printf("YOU SUCK, MAKE IT FASTER\n");
	next1 = now;
    }
    next1.tv_usec += FRAMETIME;
    if (next1.tv_usec >= 1000000)
    {
	next1.tv_sec += next1.tv_usec / 1000000;
	next1.tv_usec %= 1000000;
    }

    if (look_ugly)
	sff ^= 1;
}

void Render_APU_Status(void)
{
    memset(sbuf + 40, 0, 216);
    if (enable_pulse1)
	memset(sbuf, 1, 8);
    else
	memset(sbuf, 0, 8);
    if (enable_pulse2)
	memset(sbuf + 8, 1, 8);
    else
	memset(sbuf + 8, 0, 8);

    if (enable_tri)
	memset(sbuf + 16, 1, 8);
    else
	memset(sbuf + 16, 0, 8);

    if (enable_noise)
	memset(sbuf + 24, 1, 8);
    else
	memset(sbuf + 24, 0, 8);

    if (enable_pcm)
	memset(sbuf + 32, 1, 8);
    else
	memset(sbuf + 32, 0, 8);
}

unsigned char Read_PPU_Register(unsigned short addr)
{
    unsigned char ret;
    
    switch (addr)
    {
    case 0x2000:
    case 0x2001:	
    case 0x2002:
    case 0x2003:
    case 0x2004:
    case 0x2005:
    case 0x2006:
	ret = status;
	status &= ~STAT_VBLANK;
	ff = 0;
	return ret;

	printf("Read from $2004\n");
	return sprram[spraddr++];
    case 0x2007:
	if (vram_addr >= 0x3F00)
	{
	    ret = Read_PPU(vram_addr);
	    vram_buffer = ret;
	}
	else
	{
	    ret = vram_buffer;
	    vram_buffer = Read_PPU(vram_addr);
	}
	if (con_1 & CON1_PPUINC)
	    vram_addr += 32;
	else
	    vram_addr++;
	vram_addr &= 0x3FFF;
	return ret;
    default:
	printf("Read from register $%x\n", addr);
    }
    return 0;
}

unsigned char *decoderlut;

void DecodeTile(unsigned char *dest, unsigned long tile)
{
    unsigned long r0, r1;
    unsigned char bank = tile >> 10;
    unsigned char tmp[8];
    
    r0 = vrom[bank][(tile & 0x3FF)];
    r1 = vrom[bank][(tile & 0x3FF) + 8];

    memcpy (tmp, decoderlut + (r1 << 3), 8);
    memcpy (dest, decoderlut + (r0 << 3), 8);

    tmp[0] <<= 1;
    tmp[1] <<= 1;
    tmp[2] <<= 1;
    tmp[3] <<= 1;
    tmp[4] <<= 1;
    tmp[5] <<= 1;
    tmp[6] <<= 1;
    tmp[7] <<= 1;

    dest[0] |= tmp[0];
    dest[1] |= tmp[1];
    dest[2] |= tmp[2];
    dest[3] |= tmp[3];
    dest[4] |= tmp[4];
    dest[5] |= tmp[5];
    dest[6] |= tmp[6];
    dest[7] |= tmp[7];
}

extern unsigned long cyclesRemaining;
extern unsigned long delay;
extern short VCount;

void Write_PPU_Register(unsigned short addr, unsigned char value)
{
    int i;
    
    switch (addr)
    {
    case 0x2000:
	con_1 = value;
	tmpscroll &= ~0x0C00;
	tmpscroll |= (value & 3) << 10;
	break;
    case 0x2001:
	con_2 = value;
	if (con_2 & 0xE0)
	    printf("Hmm, something's using the color intensify...\n");	
	break;
    case 0x2003:
	spraddr = value;
	break;
    case 0x2004:
	sprram[spraddr++] = value;
	break;
    case 0x2005:
	if (ff)
	{
//	    printf("vscroll: %d\n", value);
	    tmpscroll &= ~0x73E0;
	    tmpscroll |= (value << 2) & 0x3E0;
	    tmpscroll |= (value & 7) << 12;
	}
	else
	{
//	    printf("hscroll: %d\n", value);
	    tmpscroll &= ~0x001F;
	    tmpscroll |= (value >> 3);
	    xfine = (value & 7);	    
	}
	ff ^= 1;
	break;
    case 0x2006:
	if (ff)
	{
	    tmpscroll &= 0xFF00;
	    tmpscroll |= value;
	    vram_addr = tmpscroll;
//	    printf("New vram_addr: %x\n", vram_addr);
	}
	else
	{
	    tmpscroll &= 0x00FF;
	    tmpscroll |= (value << 8) & 0x3F00;
	}
/*	if ((VCount >= 0) && (VCount < 240))
	printf("tmpscroll = %x\n", tmpscroll);*/
	ff ^= 1;
	break;
    case 0x2007:
	Write_PPU(vram_addr, value);
	if (con_1 & CON1_PPUINC)
	    vram_addr += 32;
	else
	    vram_addr++;
	vram_addr &= 0x3FFF;
	break;
    case 0x4014:
	addr = value << 8;
	for (i = 0; i < 256; i++)
	{
	    sprram[i] = Read_CPU(addr++);
	}
	delay = 513 - cyclesRemaining;
	cyclesRemaining = 0;
	break;
    default:
	printf("Write $%x to $%x\n", value, addr);
    }
}

void PPU_Decode_Tile(unsigned char *dest, unsigned char *src)
{
    unsigned char r[2];
    unsigned char i;
    unsigned char mask;
    unsigned char b;
    
    for (i = 0; i < 8; i++)
    {
	r[0] = src[0];
	r[1] = src[8];
	
	mask = 0x80;

	while (mask)
	{
	    b = 0;	

	    if (r[0] & mask)
		b |= 1;
	    if (r[1] & mask)
		b |= 2;
	    *dest++ = b;
	    mask >>= 1;
	}
	src++;
    }
}

void PPU_Decode_CHR(unsigned char *dest, unsigned char *src, unsigned long size)
{
    long ntiles = size >> 4;
    long i;

    for (i = 0; i < ntiles; i++)
    {
	PPU_Decode_Tile(dest, src);
	dest += 64;
	src += 16;
    }
}

void PPU_Start_Vblank()
{
    unsigned char draw = 1;
    
    if (sff)
    {
	draw = 0;
    }
    Sync();
    frames++;
    if (draw)
      updateimage();
    
    status |= STAT_VBLANK;    
}

unsigned char PPU_Check_NMI()
{
/*    if (con_1 & CON1_NMI)
      printf("NMI fire\n");*/
    return (con_1 & CON1_NMI);
}

void Redraw_Cache()
{
}

void PPU_End_Vblank()
{
    status &= ~STAT_HIT;
    
    if (con_2 & (CON2_SDRAW | CON2_BDRAW))
    {
	vram_addr = tmpscroll;
    }
    if (cur_mapper->Vblank_Hook)
	cur_mapper->Vblank_Hook();

    if (showcache)
    {
	Redraw_Cache();
	status |= STAT_HIT;
    }
}

unsigned long bgtiles[33];
unsigned char bgatts[33];

void Fetch_BG(void)
{
    unsigned long bank = (con_1 & CON1_BPAT) << 4;
    unsigned char ofs = ((vram_addr >> 12) & 7);
    unsigned char table = (vram_addr >> 10) & 3;
    unsigned long *bt = bgtiles;
    unsigned char *ba = bgatts;
    unsigned char tiley = vram_addr >> 4;
    unsigned char tilex = vram_addr & 0x1F;
    unsigned long tmpaddr = 0x3C0 + (tiley & 0x38);
    unsigned long tmpshift = (tiley & 4);
    unsigned long namebase = vram_addr & 0x3E0;
    unsigned char *nt = nametable[table] + namebase;
    unsigned char *at = nametable[table] + tmpaddr;    
    unsigned long tile = 0;


    int i = 33;
    int n = 0;

    while (i)
    {
	tile = nt[tilex];
	tile += bank;

	if (cur_mapper->CHR_Latch)
	    cur_mapper->CHR_Latch(tile);
	
	tile <<= 4;
	tile += ofs;
	
	*bt++ = tile;

	*ba++ = ((at[((tilex >> 2) & 7)] >> (tmpshift | (tilex & 2))) & 3) << 2;
	
	if (tilex == 0x1F)
	{
	    vram_addr ^= 0x400;
	    table ^= 1;
	    tilex = 0;
	    nt = nametable[table] + namebase;
	    at = nametable[table] + tmpaddr;
	}
	else
	    tilex++;
	
	i--;
    }
    vram_addr &= ~0x1F;
    vram_addr |= tilex;
}

unsigned char tilebuf[8];

void Draw_BG(void)
{
    unsigned long *bt = bgtiles;
    unsigned char *ba = bgatts;
    unsigned char *t;
    unsigned char b;
    unsigned char a;
    unsigned char count = 33;
    unsigned char *sb = sbuf;
    unsigned char i;

    DecodeTile(tilebuf, *bt++);
    t = tilebuf;
    a = (*ba++);
    t += xfine;
    
    switch (xfine)
    {
    case 0: do { *sb++ = a + *t++;
    case 1:      *sb++ = a + *t++;
    case 2:      *sb++ = a + *t++;
    case 3:      *sb++ = a + *t++;
    case 4:      *sb++ = a + *t++;
    case 5:      *sb++ = a + *t++;
    case 6:      *sb++ = a + *t++;
    case 7:      *sb++ = a + *t++;
	DecodeTile(tilebuf, *bt++);
	t = tilebuf;
	a = (*ba++);
	count--;
    } while (count);

    }

}

unsigned char backbuf[64];

unsigned char drawn(unsigned char *src)
{
    int i = 8;
    int d = 0;

    while (i--)
	if (*src++)
	    d++;
    return d;
}

unsigned char drawn_flip(unsigned char *src)
{
    int i = 8;
    int d = 0;

    while (i--)
	if (*src--)
	    d++;
    return d;
}

unsigned char Blit_flip(unsigned char *dst, unsigned char *src, unsigned char att)
{
    int i = 8;
    int d = 0;
    
    src += 7;

    while (i--)
    {
	if (*src)
	{
	    *dst = *src + att;
	    d++;
	}
	src--;
	dst++;
    }
    return d;
}

unsigned char Blit(unsigned char *dst, unsigned char *src, unsigned char att)
{
    int i = 8;
    int d = 0;
    
    while (i--)
    {
	if (*src)
	{
	    *dst = *src + att;
	    d++;
	}
	src++;
	dst++;
    }
    return d;
}

void Blit_back(unsigned char *dst, unsigned char *src)
{
    int i = 8;

    while (i--)
    {
	if (*src & 3)
	    *dst = *src;
	src++;
	dst++;
    }
}

extern short VCount;

void Draw_Sprites()
{
    sprent *spent = sprcache;
    unsigned char *bb = backbuf;
    int n = numsprites;
    
    while (n)
    {
	if (spent->b)
	{
	    memcpy(bb, sbuf + spent->xofs, 8);
	    bb += 8;
	}
	spent++;
	n--;
    }
    spent--;
    
    while (numsprites)
    {
	if (spent->hflip)
	    n = Blit_flip(sbuf + spent->xofs, spent->pixels, spent->att);
	else
	    n = Blit(sbuf + spent->xofs, spent->pixels, spent->att);
	
	if (n && spent->zero)
	{
	    status |= STAT_HIT;
	    hitline = VCount;
	}
	if (spent->b)
	{
	    bb -= 8;
	    Blit_back(sbuf + spent->xofs, bb);
	}

	spent--;
	numsprites--;
    }
}

void Fetch_Sprites(unsigned char line)
{
    sprite *spr = sprites;
    sprent *spent = sprcache;
    unsigned long bank = (con_1 & CON1_SPAT) << 5;
    unsigned long ofs;
    unsigned char sh;
    unsigned long tile;
    int i = 0;
    numsprites = 0;

    if (con_1 & CON1_SPRSIZE)
	sh = 16;
    else
	sh = 8;
    
    while ((numsprites < 8) && (i < 64))
    {
	if ((spr->sy < line) && (spr->sy + sh >= line))
	{
	    ofs = (line - (spr->sy + 1)) & (sh - 1);

	    if (spr->f.vflip)
		ofs = (sh - 1 - ofs);

	    if (sh == 16)
	    {
		bank = (spr->index & 1) << 8;
		if (ofs & 8)
		    tile = spr->index | 1;
		else
		    tile = spr->index & 0xFE;
		ofs &= 7;
	    }
	    else
		tile = spr->index;
	    
	    spent->tile = tile;
	    
	    ofs <<= 3;
	    
	    spent->xofs = spr->sx;
	    spent->att = (spr->f.att << 2) + 0x10;
	    spent->b = spr->f.b;
	    spent->hflip = spr->f.hflip;
	    if (!i)
	    {
		spent->zero = 1;
	    }
	    else
		spent->zero = 0;

	    tile += bank;

	    if (cur_mapper->CHR_Latch)
		cur_mapper->CHR_Latch(tile);
	    
/*	    if (cachestat[tile])
		PPU_Update_Tile(tile);

		tile <<= 6;*/

//	    printf("Tile %x == addr %x\n", spent->tile, tile);
	    
	    spent->pixels = tilebuf;//tilecache + tile + ofs;
	    
	    numsprites++;
	    spent++;
	}
	spr++;
	i++;
    }
}

void PPU_NewRenderline(unsigned char line)
{
    if (sff)
    {
	if (line == hitline)
	    status |= STAT_HIT;
	return;
    }
    
    if ((paldirty) && (con_2 & (CON2_BDRAW | CON2_SDRAW)))
    {
	paldirty = 0;
	Updatepal(palram);
    }

    if (con_2 & (CON2_BDRAW | CON2_SDRAW))
    {
	vram_addr &= ~0x041F;
	vram_addr |= tmpscroll & 0x041F;
    }
    
    if (con_2 & CON2_BDRAW)
    {
	Fetch_BG();
	Draw_BG();
	if (~con_2 & CON2_BCLIP)
	    memset(sbuf, 0, 8);
    }
    else
	memset(sbuf, 0, 8);

    if (con_2 & CON2_SDRAW)
    {
	Draw_Sprites();
    }

    if (!showcache)
    {
	if ((con_2 & (CON2_BDRAW | CON2_SDRAW)) && (line >= 8) && (line <= 232))
	{
	    Expandline(line, sbuf);
	}
	else
	{
/*	    if (line > 232)
	    {
		Render_APU_Status();
		Expandline(line, sbuf);
	    }
	    else*/
		Clearline(line);
	}
    }
    
    if (con_2 & CON2_SDRAW)
    {
	Fetch_Sprites(line + 1);
    }
    else
	numsprites = 0;
    
    if (con_2 & (CON2_SDRAW | CON2_BDRAW))
    {
	vram_addr += 0x1000;
	if (vram_addr & 0x8000)
	{
	    unsigned short tmp;
	    
	    vram_addr &= 0x0FFF;
	    tmp = vram_addr & 0x03E0;

	    if (tmp == 0x3A0)
	    {
		vram_addr ^= 0x800;
		tmp = 0;
	    }
	    else
	    {
		tmp += 0x20;
	    }
	    
	    vram_addr &= ~0x03E0;
	    vram_addr |= tmp & 0x03E0;
	}

	if (cur_mapper->Scanline_Hook)
	    cur_mapper->Scanline_Hook();	
    }
}

void PPU_OldRenderline(unsigned char line)
{
}

void Init_PPU(void)
{
    int i;
    
    memset(sprram, 0, 256);

    decoderlut = malloc(256 * 8);

    for (i = 0; i < 256; i++)
    {
	if (i & 0x80)
	    decoderlut[(i << 3)] = 1;
	if (i & 0x40)
	    decoderlut[(i << 3) + 1] = 1;
	if (i & 0x20)
	    decoderlut[(i << 3) + 2] = 1;
	if (i & 0x10)
	    decoderlut[(i << 3) + 3] = 1;
	if (i & 0x08)
	    decoderlut[(i << 3) + 4] = 1;
	if (i & 0x04)
	    decoderlut[(i << 3) + 5] = 1;
	if (i & 0x02)
	    decoderlut[(i << 3) + 6] = 1;
	if (i & 0x01)
	    decoderlut[(i << 3) + 7] = 1;	
    }
}
