#include <stdio.h>
#include <string.h>
#include <stdlib.h>

char nesmagic[4] = "NES\032";
#define MIRROR_BIT    1
#define SRAM_BIT      2
#define TRAINER_BIT   4
#define VRAM_BIT      8

typedef struct
{
    int hasname;
    char name[128];
    int prg_pages;
    int chr_pages;
    int flags;
    int mapper;
} rominfo;

int main(int argc, char **argv)
{
    unsigned char tmpstr[255];
    FILE *fp;
    rominfo ri;
    int i;
    unsigned short vectors[3];    
    
    argc--;
    argv++;

    if (!argc)
	return 0;
    
    fp = fopen(argv[0],"r");

    fread(tmpstr,1,16,fp);

    if (strncmp(tmpstr, nesmagic, 4))
    {
	printf("Bad magic number\n");
	return 0;
    }

    ri.prg_pages = tmpstr[4];
    ri.chr_pages = tmpstr[5];
    ri.mapper = (tmpstr[6]&0xF0) >> 4;
    ri.mapper |= tmpstr[7]&0xF0;
    ri.flags = tmpstr[6]&0x0F;
    
    printf("%dk PRG ROM (%d pages)\n%dk CHR ROM (%d pages)\n",
	   ri.prg_pages*16, ri.prg_pages, ri.chr_pages*8, ri.chr_pages);
    printf("Mapper %d\n",ri.mapper);
    if (ri.flags & MIRROR_BIT)
	printf("Vertical mirroring\n");
    else
	printf("Horizontal mirroring\n");
    if (ri.flags & SRAM_BIT)
	printf("SRAM\n");
    if (ri.flags & TRAINER_BIT)
	printf("Trainer\n");
    if (ri.flags & VRAM_BIT)
	printf("4 screen VRAM\n");

    if (ri.flags & TRAINER_BIT)
    {
	printf("Skipping 512 byte trainer...\n");
	fseek(fp, 512, SEEK_CUR);
    }

    i = ri.prg_pages;

    printf("Skipping PRG pages...\n");
    while (i--)
	fseek(fp, 16*1024, SEEK_CUR);

    fseek(fp, -6, SEEK_CUR);
    fread(vectors, 2, 3, fp);
    printf("NMI: $%x\nRESET: $%x\nIRQ: $%x\n", vectors[0], vectors[1],
           vectors[2]);

    i = ri.chr_pages;

    if (i)
	printf("Skipping CHR pages...\n");

    while (i--)
	fseek(fp, 8*1024, SEEK_CUR);

    printf("At point %ld in the file...\n", ftell(fp));
    
    printf("Attempting to read name...\n");
    fread(ri.name, 1, 128, fp);
    if (!feof(fp))
	ri.hasname = 1;
    else
	ri.hasname = 0;

    if (ri.hasname)
	printf("Cart name: %s\n", ri.name);

    return 0;
}

