; s6502.asm
;
; Stanknes 6502 core
;
; Expects 00000000h-0000FFFFh to be mapped appropriately
; doesn't support running code in $2-5xxx correctly, will
; most likely result in a protection fault unless they are
; mapped in

; Calling code must also make sure the page at 00010000h is
; mapped, as it will be used for context saving and possibly
; temp storage

; Fetch opcode byte into AL, jump to corebase + AL << log2(opsize)
; opsize is the max. number of bytes for an instruction.
; should be an even multiple of 32 for efficient cache usage

; see 6502corenotes for more

%macro PACK_FLAGS 0         ; expects tmp flags in bh, x86 flags in ah
mov   dh,    ah             ; 2
shl   dh,    1              ; 2
rcr   bh,    3              ; 3
shr   edx,   2              ; 3
shr   dh,    6              ; 2
shr   edx,   1              ; 2
xchg  bh,    dh             ; 2
shr   edx,   6              ; 3
%endmacro                   ; 19

%macro _PUSH 1
movzx ebx, byte [10004h]    ; 7
mov   [ebx], %1             ; 3
dec   bl                    ; 2
mov   [10004h], bl          ; 6
%endmacro                   ; 18

%macro _PUSHW 1             ; takes letter of GP reg to push
movzx ebx, byte [10004h]    ; 7
mov   [ebx], %1h            ; 3
dec   bl                    ; 2
mov   [ebx], %1l            ; 3
dec   bl                    ; 2
mov   [10004h], bl          ; 6
%endmacro                   ; 23

%macro _PUSH24 1            ; 
movzx ebx, byte [10004h]    ; 7
mov   [ebx], %1h            ; 3
dec   bl                    ; 2
mov   [ebx], %1l            ; 3
dec   bl                    ; 2
shr   e%1x, 8               ; 3
mov   [ebx], %1h            ; 3
dec   bl                    ; 2
mov   [10004h], bl          ; 6
%endmacro                   ; 31

%macro _POP 1
movzx ebx, byte [10004h]    ; 7
mov   %1, [ebx]             ; 3
inc   bl                    ; 2
mov   [10004h], bl          ; 6
%endmacro                   ; 18

%macro _FETCHB 1
mov   %1, [esi]             ; 3
inc   esi                   ; 2
%endmacro                   ; 5

%macro _FETCHBX 1
movzx %1, byte [esi]        ; 3
inc   esi                   ; 2
%endmacro                   ; 5

%macro _FETCHW 1
movzx %1, word [esi]        ; 3
add   esi, 2                ; 3
%endmacro                   ; 6

%macro _ZPREADB 1
mov   %1, [ebx]             ; 3
%endmacro                   ; 3

%macro _READB
; FIXME
%endmacro

%macro _WRITEB
; FIXME
%endmacro

%macro AM_IMM
_FETCHB(dl)                 
%endmacro                   ; 3

%macro AM_ZP
_FETCHBX(ebx)               ; 5
%endmacro                   ; 5

%macro AM_ZPX
_FETCHBX(ebx)               ; 5
add   bl, ch                ; 2
%endmacro                   ; 7

%macro AM_ZPY
_FETCHBX(ebx)               ; 5
add   bl, cl                ; 2
%endmacro                   ; 7

%macro AM_ABS
_FETCHW(ebx)                ; 6
%endmacro                   ; 6

%macro AM_ABSX              ; base timing in dh, exits with actual time in dh, addr in ebx
_FETCHW(ebx)                ; 6
add   bl, ch                ; 2
setc  dl                    ; 3
add   dh, dl                ; 2
add   bh, dl                ; 2
%endmacro                   ; 15

%macro AM_ABSY
_FETCHW(ebx)                ; 6
add   bl, cl                ; 2
setc  dl                    ; 3
add   dh, dl                ; 2
add   bh, dl                ; 2
%endmacro                   ; 15

%macro AM_ABSX_FIXED
_FETCHW(ebx)                ; 6
add   bl, ch                ; 2
adc   bh, 0                 ; 3
%endmacro                   ; 11

%macro AM_ABSY_FIXED
_FETCHW(ebx)                ; 6
add   bl, cl                ; 2
adc   bh, 0                 ; 3
%endmacro                   ; 11

%macro AM_REL
; FIXME
%endmacro

%macro AM_INDX
_FETCHBX(edx)               ; 5
add   dl, ch                ; 2
mov   bl, [edx]             ; 2
inc   dl                    ; 2
mov   bh, [edx]             ; 2
%endmacro                   ; 13

%macro AM_INDY              ; base time in edi, exits with addr in ebx, time in edi
_FETCHBX(edx)               ; 5
mov   bl, [edx]             ; 2
inc   dl                    ; 2
mov   bh, [edx]             ; 2
add   bl, cl                ; 2
setc  dl                    ; 3
add   bh, dl                ; 2
shr   dl, 1                 ; 2
adc   edi, 0                ; 3
%endmacro

%macro _EXIT 1
sub   ebp, %1
jmp end_table
%endmacro

%macro _ORA
sahf
or    al, dl
lahf
%endmacro

%macro _EOR
sahf
xor   al, dl
lahf
%endmacro

%macro _ASL 1
sahf
shl   %1, 1
lahf
%endmacro

%macro _LSR 1
sahf
shr   %1, 1
lahf
%endmacro

%macro _AND
sahf
and   al, dl
lahf
%endmacro

%macro _ADC
sahf
adc   al, dl
lahf
seto  dl
shl   dl, 7
mov   dh, [10007h]
and   dh, 7Fh
or    dh, dl
mov   [10007h], dh
%endmacro

%macro _SBC
sahf
sbb   al, dl
lahf
seto  dl
shl   dl, 7
mov   dh, [10007h]
and   dh, 7Fh
or    dh, dl
mov   [10007h], dh
%endmacro

%macro _ROL 1
sahf
rcl   %1, 1
lahf
%endmacro

%macro _ROR 1
sahf
rcr   %1, 1
lahf
%endmacro



; Execute_6502(unsigned long cycles);
;
; Executes <cycles> 6502 cycles, under some circumstances it 
; might execute up to 6 extra cycles

Execute_6502:

mov   ebp, [esp + 4]

coreloop:

movzx ebx, byte [10007h]
push  ebx
; FIXME interrupts, context load

test  esi, 0E000h
jz rammirror
movzx ebx, byte [esi]
inc   esi
shl   ebx, TABLE_SHIFT
add   ebx, table_base
jmp   eax

rammirror:
and   esi, 07FFh
movzx ebx, byte [esi]
inc   esi
shl   eax, TABLE_SHIFT
add   eax, table_base
jmp   eax


table_base:

; ***** BRK

ALIGN 32
_BRK:
inc   esi                 ; 2
lea   edi, [esp + 4]      ; 3
or    [edi], 20h          ; 3
or    [edi], 08h          ; 3
mov   bh, [edi]           ; 2
PACK_FLAGS                ; 19
xchg  edx, esi            ; 2
_PUSHW(d)                 ; 21
mov   edx, esi            ; 2
_PUSH(dl)                 ; 18
movzx esi, word [0FFFEh]  ; 7
sub   ebp, 7              ; 3
jmp end_table             ; 5

; ***** ORA

ALIGN 32
_ORA_IMM:
AM_IMM
_ORA
_EXIT(2)

ALIGN 32
_ORA_ZP:
AM_ZP                     ; 5
_READB
_ORA
_EXIT(3)

ALIGN 32
_ORA_ZPX:
AM_ZPX
_READB
_ORA
_EXIT(4)

ALIGN 32
_ORA_ABS:
AM_ABS
_READB
_ORA
_EXIT(4)

ALIGN 32
_ORA_ABSX:
mov   dh, 4
AM_ABSX
_READB
_ORA
movzx edx, dh
_EXIT(edx)

ALIGN 32
_ORA_ABSY:
mov   dh, 4
AM_ABSY
_READB
_ORA
movzx edx, dh
_EXIT(edx)

ALIGN 32
_ORA_INDX:
AM_INDX                   ; 13
_READB                    ; ?
_ORA
_EXIT(6)

ALIGN 32
_ORA_INDY:
mov   edi, 5
AM_INDY                   ; 13
_READB                    ; ?
_ORA
_EXIT(edi)

; ***** ASL

ALIGN 32
_ASL_ACCUM:
_ASL(al)
_EXIT(2)

ALIGN 32
_ASL_ZP:
AM_ZP                     ; 5
_READB
_ASL(dl)
_WRITEB
_EXIT(5)

ALIGN 32
_ASL_ZPX:
AM_ZPX
_READB
_ASL(dl)
_WRITEB
_EXIT(6)

ALIGN 32
_ASL_ABS:
AM_ABS
_READB
_ASL(dl)
_WRITEB
_EXIT(6)

ALIGN 32
_ASL_ABSX:
AM_ABSX_FIXED
_READB
_ASL(dl)
_WRITEB
_EXIT(7)

; ***** LSR

ALIGN 32
_LSR_ACCUM:
_LSR(al)
_EXIT(2)

ALIGN 32
_LSR_ZP:
AM_ZP                     ; 5
_READB
_LSR(dl)
_WRITEB
_EXIT(5)

ALIGN 32
_LSR_ZPX:
AM_ZPX
_READB
_LSR(dl)
_WRITEB
_EXIT(6)

ALIGN 32
_LSR_ABS:
AM_ABS
_READB
_LSR(dl)
_WRITEB
_EXIT(6)

ALIGN 32
_LSR_ABSX:
AM_ABSX_FIXED
_READB
_LSR(dl)
_WRITEB
_EXIT(7)

; ***** NOP

ALIGN 32
_NOP:
_EXIT(2)

; ***** EOR

ALIGN 32
_EOR_IMM:
AM_IMM
_EOR
_EXIT(2)

ALIGN 32
_EOR_ZP:
AM_ZP                     ; 5
_READB
_EOR
_EXIT(3)

ALIGN 32
_EOR_ZPX:
AM_ZPX
_READB
_EOR
_EXIT(4)

ALIGN 32
_EOR_ABS:
AM_ABS
_READB
_EOR
_EXIT(4)

ALIGN 32
_EOR_ABSX:
mov   dh, 4
AM_ABSX
_READB
_EOR
movzx edx, dh
_EXIT(edx)

ALIGN 32
_EOR_ABSY:
mov   dh, 4
AM_ABSY
_READB
_EOR
movzx edx, dh
_EXIT(edx)

ALIGN 32
_EOR_INDX:
AM_INDX                   ; 13
_READB                    ; ?
_EOR
_EXIT(6)

ALIGN 32
_EOR_INDY:
mov   edi, 5
AM_INDY                   ; 13
_READB                    ; ?
_EOR
_EXIT(edi)

; ***** AND

ALIGN 32
_AND_IMM:
AM_IMM
_AND
_EXIT(2)

ALIGN 32
_AND_ZP:
AM_ZP                     ; 5
_READB
_AND
_EXIT(3)

ALIGN 32
_AND_ZPX:
AM_ZPX
_READB
_AND
_EXIT(4)

ALIGN 32
_AND_ABS:
AM_ABS
_READB
_AND
_EXIT(4)

ALIGN 32
_AND_ABSX:
mov   dh, 4
AM_ABSX
_READB
_AND
movzx edx, dh
_EXIT(edx)

ALIGN 32
_AND_ABSY:
mov   dh, 4
AM_ABSY
_READB
_AND
movzx edx, dh
_EXIT(edx)

ALIGN 32
_AND_INDX:
AM_INDX                   ; 13
_READB                    ; ?
_AND
_EXIT(6)

ALIGN 32
_AND_INDY:
mov   edi, 5
AM_INDY                   ; 13
_READB                    ; ?
_AND
_EXIT(edi)

; ***** ADC

ALIGN 32
_ADC_IMM:
AM_IMM
_ADC
_EXIT(2)

ALIGN 32
_ADC_ZP:
AM_ZP                     ; 5
_READB
_ADC
_EXIT(3)

ALIGN 32
_ADC_ZPX:
AM_ZPX
_READB
_ADC
_EXIT(4)

ALIGN 32
_ADC_ABS:
AM_ABS
_READB
_ADC
_EXIT(4)

ALIGN 32
_ADC_ABSX:
mov   dh, 4
AM_ABSX
_READB
_ADC
movzx edx, dh
_EXIT(edx)

ALIGN 32
_ADC_ABSY:
mov   dh, 4
AM_ABSY
_READB
_ADC
movzx edx, dh
_EXIT(edx)

ALIGN 32
_ADC_INDX:
AM_INDX                   ; 13
_READB                    ; ?
_ADC
_EXIT(6)

ALIGN 32
_ADC_INDY:
mov   edi, 5
AM_INDY                   ; 13
_READB                    ; ?
_ADC
_EXIT(edi)

; ***** SBC

ALIGN 32
_SBC_IMM:
AM_IMM
_SBC
_EXIT(2)

ALIGN 32
_SBC_ZP:
AM_ZP                     ; 5
_READB
_SBC
_EXIT(3)

ALIGN 32
_SBC_ZPX:
AM_ZPX
_READB
_SBC
_EXIT(4)

ALIGN 32
_SBC_ABS:
AM_ABS
_READB
_SBC
_EXIT(4)

ALIGN 32
_SBC_ABSX:
mov   dh, 4
AM_ABSX
_READB
_SBC
movzx edx, dh
_EXIT(edx)

ALIGN 32
_SBC_ABSY:
mov   dh, 4
AM_ABSY
_READB
_SBC
movzx edx, dh
_EXIT(edx)

ALIGN 32
_SBC_INDX:
AM_INDX                   ; 13
_READB                    ; ?
_SBC
_EXIT(6)

ALIGN 32
_SBC_INDY:
mov   edi, 5
AM_INDY                   ; 13
_READB                    ; ?
_SBC
_EXIT(edi)