#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include "rom.h"
#include "sun4.h"
#include "memmap.h"

extern int memfd;

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char *sram;
static unsigned long nchrs;
static unsigned long nprgs;
static unsigned char prgmask;
static unsigned char chrmask;
static unsigned char command;
static unsigned char regs[1];
static unsigned char chrswitch[4];

extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

void InitSUN4(rominfo *ri)
{
    unsigned char *p;
    
    printf("Initializing Sunsoft FME-7 mapper...\n");

    prgs = ri->prgrom;
    chrs = ri->chrrom;
    sram = ri->sram;

    nchrs = ri->chr_pages;
    nprgs = ri->prg_pages;    
    
    nchrs <<= 2;

    prgmask = nprgs - 1;
    chrmask = nchrs - 1;
    
    printf("%d 8K PRG pages\n%d 1K CHR pages\n", nprgs, nchrs);

    mirror = ri->f.f.mirror;

    p = mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$8000 MMAP FAIL\n");

    p = mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1)*0x4000);
    if (p == (unsigned char *)-1)
	printf("$C000 MMAP FAIL\n");

    regs[0] = 0;

    chrswitch[0] = 0;
    chrswitch[1] = 1;
    chrswitch[2] = 2;
    chrswitch[3] = 3;

    vrom[0] = chrs;
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = vrom[0] + 0x800;
    vrom[3] = vrom[0] + 0xC00;
    vrom[4] = vrom[0] + 0x1000;
    vrom[5] = vrom[0] + 0x1400;
    vrom[6] = vrom[0] + 0x1800;
    vrom[7] = vrom[0] + 0x1C00;

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (ri->f.f.mirror)
	MM_Mirror_Horiz(nt1, nt2);
    else
	MM_Mirror_Vert(nt1, nt2);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char SUN4_Read_Mapper(unsigned long addr)
{
    return 0;
}

void Update_SUN4(void)
{
    vrom[0] = chrs + ((chrswitch[0] & chrmask) * 0x400);
    vrom[1] = vrom[0] + 0x400;
    vrom[2] = chrs + ((chrswitch[1] & chrmask) * 0x400);
    vrom[3] = vrom[2] + 0x400;
    vrom[4] = chrs + ((chrswitch[2] & chrmask) * 0x400);
    vrom[5] = vrom[4] + 0x400;
    vrom[6] = chrs + ((chrswitch[3] & chrmask) * 0x400);
    vrom[7] = vrom[6] + 0x400;

    switch(mirror)
    {
    case 0: MM_Mirror_Vert(nt1, nt2); break;
    case 1: MM_Mirror_Horiz(nt1, nt2); break;
    case 2: MM_Mirror_One(nt1); break;
    case 3: MM_Mirror_One(nt2); break;
    default: printf("Unknown mirror setting %d\n", mirror); break;
    }
    
    mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + ((regs[0] & prgmask)*0x4000));
}

void SUN4_Write_Mapper(unsigned long addr, unsigned char value)
{
    switch (addr)
    {
    case 0x8000: chrswitch[0] = value; break;
    case 0x9000: chrswitch[1] = value; break;
    case 0xA000: chrswitch[2] = value; break;
    case 0xB000: chrswitch[3] = value; break;
    case 0xE000: mirror = value; break;
    case 0xF000: regs[0] = value; break;
    default: printf("Sunsoft #4 write %x to %x\n", value, addr);
    }
    Update_SUN4();
}

unsigned char SUN4_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void SUN4_Write_CHR(unsigned long addr, unsigned char value)
{
    printf("WRITE TO CHR\n");
    return;
}
