#include <stdio.h>

#include <X11/X.h>
#include <X11/Xlib.h>

#include "xlibgfx.h"
#include "t24.h"
#include "pal.h"

extern unsigned char *framebuffer;
extern unsigned char *debugbuffer;
unsigned long palconv[69];
unsigned long black, red, blue, green;
unsigned long blankline[264];
unsigned long pallut[33];
unsigned long monolut[33];

void T24_InitLookups(void)
{
    XColor col;
    int i;

    printf("Initializing lookups\n");
    xcmap = XCreateColormap(xdisp, xwin, xvisual, 0);

    col.flags = DoRed | DoGreen | DoBlue;
    col.red = 0;
    col.green = 0;
    col.blue = 0;
    XAllocColor(xdisp, xcmap, &col);
    black = col.pixel;
    col.flags = DoRed | DoGreen | DoBlue;
    col.red = 0xFFFF;
    col.green = 0;
    col.blue = 0;
    XAllocColor(xdisp, xcmap, &col);
    red = col.pixel;
    col.flags = DoRed | DoGreen | DoBlue;
    col.red = 0;
    col.green = 0xFFFF;
    col.blue = 0;
    XAllocColor(xdisp, xcmap, &col);
    green = col.pixel;
    col.flags = DoRed | DoGreen | DoBlue;
    col.red = 0xFFFF;
    col.green = 0xFFFF;
    col.blue = 0xFFFF;
    XAllocColor(xdisp, xcmap, &col);
    blue = col.pixel;

    pallut[32] = red;
    
    for (i = 0; i < 64; i++)
    {
	col.flags = DoRed | DoGreen | DoBlue;
	col.red = nes_pal[i].r;
	col.blue = nes_pal[i].b;
	col.green = nes_pal[i].g;
	col.red += col.red << 8;
	col.green += col.green << 8;
	col.blue += col.blue << 8;
	if (XAllocColor(xdisp, xcmap, &col))
	{    
	    palconv[i] = col.pixel;
	}
    }
    palconv[64] = red;
    palconv[65] = red;
    palconv[66] = red;
    palconv[68] = red;    
    for (i = 0; i < 264; i++)
	blankline[i] = black;
}

void T24_Clearline(unsigned char line)
{
/*    unsigned long *buf = (unsigned long *)framebuffer + (line << 8);

    for (line = 0; line < 16; line++)
    {
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	*buf++ = black;
	}*/
    memcpy((unsigned long *)framebuffer + (line << 8) + (line << 3), blankline, 1056);
}

void T24_Monoline(unsigned char line, unsigned char *data)
{
    unsigned long *buf = (unsigned long *)framebuffer + (line << 8) + (line << 3);
    for (line = 0; line < 16; line++)
    {
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];	
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];	
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];	
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];
	*buf++ = monolut[*data++];	
	*buf++ = monolut[*data++];	
    }    
}

void T24_Expandline(unsigned char line, unsigned char *data)
{
    unsigned long *buf = (unsigned long *)framebuffer + (line << 8) + (line << 3);
    for (line = 0; line < 16; line++)
    {
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];	
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];	
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];	
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];
	*buf++ = pallut[*data++];	
	*buf++ = pallut[*data++];	
    }    
}

void T24_Updatepal(unsigned char *data)
{
    unsigned long *p = pallut;
    
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];

    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];

    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];

    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p++ = palconv[*data++];
    *p = palconv[*data];

    p = monolut;
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];

    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];

    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];

    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p++ = palconv[*data++ & 0x30];
    *p = palconv[*data & 0x30];    

}

void T24_DebugExpandline(unsigned char line, unsigned char *data)
{
}

unsigned char tbuf[4];
extern unsigned long decoderlut[16];

void T24_Decode_5bit(unsigned char line, unsigned char shift, unsigned char *p0,
		     unsigned char *p1, unsigned char *p2, unsigned char *p3, unsigned char *p4)
{
    unsigned long acc;
    unsigned char i = 32;
    unsigned long c;
    unsigned long *buf = (unsigned long *)framebuffer + (line << 3) + (line << 8) + 8 - shift;
    unsigned long *sb = (unsigned long *)tbuf;
    
    if (shift)
	i++;

    do
    {   /*
	  Note for the strange stuff here -- GCC's code generator was being really stupid,
	  this IS better than acc = decoderlut[*p4 >> 4]; and c = *p4 >> 4; acc = decoderlut[c];
	*/
	c = *p4;
	acc = decoderlut[c >> 4];
	acc <<= 1;
	c = *p3;
	acc |= decoderlut[c >> 4];
	acc <<= 1;
	c = *p2;
	acc |= decoderlut[c >> 4];
	acc <<= 1;
	c = *p1;
	acc |= decoderlut[c >> 4];
	acc <<= 1;
	c = *p0;
	acc |= decoderlut[c >> 4];
	
	*sb = acc;

	*buf++ = pallut[tbuf[0]];
	*buf++ = pallut[tbuf[1]];
	*buf++ = pallut[tbuf[2]];
	*buf++ = pallut[tbuf[3]];

	c = *p4;
	acc = decoderlut[c & 0xF];
	acc <<= 1;
	c = *p3;
	acc |= decoderlut[c & 0xF];
	acc <<= 1;
	c = *p2;
	acc |= decoderlut[c & 0xF];
	acc <<= 1;
	c = *p1;
	acc |= decoderlut[c & 0xF];
	acc <<= 1;
	c = *p0;
	acc |= decoderlut[c & 0xF];
	*sb = acc;	

	*p4++;
	*p3++;
	*p2++;
	*p1++;
	*p0++;
	
	*buf++ = pallut[tbuf[0]];
	*buf++ = pallut[tbuf[1]];
	*buf++ = pallut[tbuf[2]];
	*buf++ = pallut[tbuf[3]];

	i--;
    } while (i);
}
