#include <stdio.h>
#include <stdlib.h>

#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "unrom.h"
#include "memmap.h"

extern int memfd;

static int mirror;
static unsigned char *prgs;
static unsigned char nprgs;
extern unsigned char *nametable[4];

unsigned char *nt1;
unsigned char *nt2;

void InitUNROM(rominfo *ri)
{
    printf("Initializing UNROM mapper...\n");

    prgs = ri->prgrom;
    nprgs = ri->prg_pages;

    mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 1)*0x4000);
    
    vrom[0] = malloc(0x400);
    vrom[1] = malloc(0x400);
    vrom[2] = malloc(0x400);
    vrom[3] = malloc(0x400);
    vrom[4] = malloc(0x400);
    vrom[5] = malloc(0x400);
    vrom[6] = malloc(0x400);
    vrom[7] = malloc(0x400);    

    mirror = ri->f.f.mirror;
    
    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (mirror)
    {
	nametable[0] = nt1;
	nametable[1] = nt2;
	nametable[2] = nt1;
	nametable[3] = nt2;
    }
    else
    {
	nametable[0] = nt1;
	nametable[1] = nt1;
	nametable[2] = nt2;
	nametable[3] = nt2; 
    }

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char UNROM_Read_Mapper(unsigned long addr)
{
    return 0;
}

void UNROM_Write_Mapper(unsigned long addr, unsigned char value)
{
#ifdef DEBUG
    if (value > nprgs)
	printf("Bank switch outside of defined rom pages...\n");
#endif
    if (addr >= 0x8000)
	mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd,
	     (unsigned long)prgs + (value & (nprgs - 1))*0x4000);
}

unsigned char UNROM_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void UNROM_Write_CHR(unsigned long addr, unsigned char value)
{
    vrom[addr>>10][addr&0x3FF] = value;
}
