#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "vrc6.h"
#include "memmap.h"

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char nprgs;
static unsigned long nchrs;
static unsigned char irq_en = 0;
static unsigned char irq_reseten = 0;
static unsigned char irq_counter = 0;
static unsigned char irq_latch = 0;
static unsigned char irq_high = 0;
static unsigned char irq_state = 0;
static unsigned char chrpages[8];

extern unsigned char vrc6_audio;
extern unsigned char *nametable[4];
unsigned char *nt1;
unsigned char *nt2;

extern int memfd;
extern int zerofd;

void VRC6_Cycle(void)
{
/*    if (irq_en)
    {
	irq_counter++;
	if (irq_counter == 0x00)
	{
	    irq_high = 1;
//	    irq_counter = 1;
	}
	}*/
    if ((irq_state & 2) && (++irq_counter == 0))
    {
	irq_high = 1;
	irq_counter = irq_latch;
    }
}

extern short VCount;

int VRC6_Check(void)
{
    if (irq_high)
    {
//	printf("IRQ fire @ scanline %d\n", VCount);
	irq_high = 0;
	return 1;
    }
    return 0;
}

void InitVRC6(rominfo *ri)
{
    int i;
    void *p;
    
    printf("Initializing VRC6 mapper...\n");

    prgs = ri->prgrom;
    chrs = ri->chrrom;
    
    nchrs = ri->chr_pages << 3;
    nprgs = ri->prg_pages << 1;

    printf("%d 8K PRG pages\n%ld 1K CHR pages\n", nprgs, nchrs);
    
    mirror = ri->f.f.mirror;

    vrc6_audio = 1;
    
    printf("Setting $8000 to page 0 (%p), $A000 to page 1 (%p), \n"
	   "$C000 to page %d (%p), $E000 to page %d (%p)\n", prgs, prgs + 0x2000,
	   nprgs - 2, prgs + (nprgs - 2)*0x2000, nprgs - 1, prgs + (nprgs - 1)*0x2000);

    p = mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);
    if (p == (unsigned char *)-1)
	printf("$8000 MMAP FAIL\n");

    p = mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 2)*0x2000);
    if (p == (unsigned char *)-1)
	printf("$C000 MMAP FAIL\n");

    vrom[0] = chrs;
    vrom[1] = chrs + 0x400;
    vrom[2] = chrs + 0x800;    
    vrom[3] = chrs + 0xC00;
    vrom[4] = chrs + 0x1000;
    vrom[5] = chrs + 0x1400;
    vrom[6] = chrs + 0x1800;
    vrom[7] = chrs + 0x1C00;

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (ri->f.f.mirror)
	MM_Mirror_Vert(nt1, nt2);
    else
	MM_Mirror_Horiz(nt1, nt2);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char VRC6_Read_Mapper(unsigned long addr)
{
    return 0;
}

void Write_APU_Register(unsigned short addr, unsigned char value);

void VRC6_Write_Mapper(unsigned long addr, unsigned char value)
{
    switch (addr & 0xF003)
    {
    case 0x8000:
	mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
		 MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (value * 0x4000));
	break;
    case 0x9000:
    case 0x9001:
    case 0x9002:
    case 0xA000:
    case 0xA001:
    case 0xA002:
    case 0xB000:
    case 0xB001:
    case 0xB002:
	Write_APU_Register(addr, value);
	break;
    case 0xB003:
	switch (value & 0x7F)
	{
	case 0x20:
	    MM_Mirror_Vert(nt1, nt2);
	    break;
	case 0x24:
	case 0x04:
	    MM_Mirror_Horiz(nt1, nt2);
	    break;
	case 0x28:
	case 0x08:
	    MM_Mirror_One(nt1);
	    break;
	case 0x2C:
	    MM_Mirror_One(nt2);
	    break;	    
	}
	break;
    case 0xC000:
	mmap((unsigned char *)0xC000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (value * 0x2000));
	break;
    case 0xD000:
	chrpages[0] = value;
	vrom[0] = chrs + (value*0x400);
	break;
    case 0xD001:
	chrpages[1] = value;	
	vrom[1] = chrs + (value*0x400);
	break;
    case 0xD002:
	chrpages[2] = value;	
	vrom[2] = chrs + (value*0x400);
	break;
    case 0xD003:
	chrpages[3] = value;	
	vrom[3] = chrs + (value*0x400);
	break;
    case 0xE000:
	chrpages[4] = value;	
	vrom[4] = chrs + (value*0x400);
	break;
    case 0xE001:
	chrpages[5] = value;	
	vrom[5] = chrs + (value*0x400);
	break;
    case 0xE002:
	chrpages[6] = value;	
	vrom[6] = chrs + (value*0x400);
	break;
    case 0xE003:
	chrpages[7] = value;	
	vrom[7] = chrs + (value*0x400);
	break;
    case 0xF000:
	irq_latch = value;
	break;
    case 0xF001:
	irq_state = value;
	irq_counter = irq_latch;
	break;
    case 0xF002:
	irq_state = (irq_state << 1) | (irq_state & 1);
	break;
    default:
	printf("Unknown write: $%4x: $%x\n", addr, value);
    }
}

unsigned char VRC6_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void VRC6_Write_CHR(unsigned long addr, unsigned char value)
{
    return;
}
