#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "rom.h"
#include "vrc7.h"
#include "memmap.h"

static int mirror;
static unsigned char *chrs;
static unsigned char *prgs;
static unsigned char nprgs;
static unsigned long nchrs;

extern unsigned char *nametable[4];
unsigned char *nt1;
unsigned char *nt2;

extern int memfd;
extern int zerofd;

void InitVRC7(rominfo *ri)
{
    int i;
    void *p;

    printf("Initializing VRC7 mapper...\n");

    prgs = ri->prgrom;
    chrs = ri->chrrom;

    nchrs = ri->chr_pages << 3;
    nprgs = ri->prg_pages << 1;

    printf("%d 8K PRG pages\n%ld 1K CHR pages\n", nprgs, nchrs);

    mirror = ri->f.f.mirror;

    p = mmap((unsigned char *)0x8000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs);

    p = mmap((unsigned char *)0xC000, 0x4000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (nprgs - 2)*0x2000);

    vrom[0] = malloc(0x400);
    vrom[1] = malloc(0x400);
    vrom[2] = malloc(0x400);
    vrom[3] = malloc(0x400);
    vrom[4] = malloc(0x400);
    vrom[5] = malloc(0x400);
    vrom[6] = malloc(0x400);
    vrom[7] = malloc(0x400);

    nt1 = malloc(0x400);
    nt2 = malloc(0x400);

    if (ri->f.f.mirror)
	MM_Mirror_Vert(nt1, nt2);
    else
	MM_Mirror_Horiz(nt1, nt2);

    printf("RESET address: %.2x%.2x\n", ram[0xFFFD], ram[0xFFFC]);
    printf("NMI address: %.2x%.2x\n", ram[0xFFFB], ram[0xFFFA]);
    printf("IRQ address: %.2x%.2x\n", ram[0xFFFF], ram[0xFFFE]);
}

unsigned char VRC7_Read_Mapper(unsigned long addr)
{
    return 0;
}

void Write_APU_Register(unsigned short addr, unsigned char value);

void VRC7_Write_Mapper(unsigned long addr, unsigned char value)
{
    switch (addr & 0xF030)
    {
    case 0x8000:
	mmap((unsigned char *)0x8000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (value * 0x2000));
	break;
    case 0x8010:
	mmap((unsigned char *)0xA000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (value * 0x2000));
	break;
    case 0x9000:
	mmap((unsigned char *)0xC000, 0x2000, PROT_READ|PROT_WRITE,
	     MAP_PRIVATE|MAP_FIXED, memfd, (unsigned long)prgs + (value * 0x2000));
	break;
    case 0xA000:
    case 0xA010:
    case 0xB000:
    case 0xB010:
    case 0xC000:
    case 0xC010:
    case 0xD000:
    case 0xD010:
	printf("Messing with CHR thingy\n");
	break;
    case 0xE000:
	printf("Messing with mirror\n");
	break;
    case 0xE010:
    case 0xF000:
    case 0xF010:
	printf("Messing with IRQ\n");
	break;
    case 0x9010:
    case 0x9030:
	Write_APU_Register(addr, value);
	break;
    default:
	printf("Unknown write: $%4x: $%x\n", addr, value);
    }    
}

unsigned char VRC7_Read_CHR(unsigned long addr)
{
    return vrom[addr >> 10][addr & 0x3FF];
}

void VRC7_Write_CHR(unsigned long addr, unsigned char value)
{
    vrom[addr >> 10][addr & 0x3FF] = value;
}

