#include <stdio.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

#include "pal.h"

#include "t24.h"
#include "t16.h"
#include "p8.h"

#define SCR_HEIGHT 240
#define SCR_WIDTH  256
#define DEBUG_WIDTH 512
#define DEBUG_HEIGHT 256
#define DEBUG2_WIDTH 512
#define DEBUG2_HEIGHT 480

Status mit_shm;
XImage *image;
XImage *image2;
XImage *debugimage;
XImage *debugimage2;
XShmSegmentInfo shminfo;
XShmSegmentInfo shminfo2;
XShmSegmentInfo debugshminfo;
XShmSegmentInfo debugshminfo2;

unsigned short *framebuffer = NULL;
unsigned short *framebuffer1 = NULL;
unsigned short *framebuffer2 = NULL;
unsigned short *debugbuffer = NULL;
unsigned short *debugbuffer2 = NULL;

int bytes_per_line = 0;
int dbytes_per_line = 0;
int d2bytes_per_line = 0;
int screen_height = 0;
int scrswitch = 0;

#define SCR_SIZE ( bytes_per_line * screen_height )

Display *xdisp;
int      xscreen;
Visual  *xvisual;
GC       xgc;
Colormap xcmap;
Window   xwin;
int      xplanes;

int d1, d2;
Bool d3;

typedef struct
{
    unsigned char class;
    unsigned char depth;
    void (*InitLookups)(void);
    void (*Expandline)(unsigned char line, unsigned char *data);
    void (*Clearline)(unsigned char line);
    void (*Updatepal)(unsigned char *data);
    void (*Decode_5bit)(unsigned char line, unsigned char shift, unsigned char *p0,
			unsigned char *p1, unsigned char *p2, unsigned char *p3, unsigned char *p4 );
} backend;

backend backends[] =
{
    {TrueColor, 16, T16_InitLookups, T16_Expandline, T16_Clearline, T16_Updatepal, NULL},
    {PseudoColor, 8,  P8_InitLookups, P8_Expandline, P8_Clearline, NULL, NULL},
    {TrueColor, 24, T24_InitLookups, T24_Expandline, T24_Clearline, T24_Updatepal, T24_Decode_5bit},
    {0xFF, 0xFF, NULL, NULL}
};

void (*Expandline)(unsigned char line, unsigned char *data);
void (*DebugExpandline)(unsigned char line, unsigned char *data);
void (*Clearline)(unsigned char line);
void (*Updatepal)(unsigned char *data);
void (*Decode_5bit)(unsigned char line, unsigned char shift, unsigned char *p0,
		    unsigned char *p1, unsigned char *p2, unsigned char *p3, unsigned char *p4 );

int killimage(void)
{
    XShmDetach( xdisp,&shminfo );
    XShmDetach( xdisp,&shminfo2 );
    XDestroyImage( image );
    XDestroyImage( image2 );    
    shmdt( shminfo.shmaddr );
    shmdt( shminfo2.shmaddr );
    shmctl( shminfo.shmid,IPC_RMID,0 );
    shmctl( shminfo2.shmid,IPC_RMID,0 );
    return 0;
}

int updateimage()
{
    if (scrswitch)
    {
	XShmPutImage(xdisp, xwin, xgc, image2, 8, 0,
		     0, 0, SCR_WIDTH, SCR_HEIGHT, 0);
	framebuffer = framebuffer1;
    }
    else
    {
	XShmPutImage(xdisp, xwin, xgc, image, 8, 0,
		     0, 0, SCR_WIDTH, SCR_HEIGHT, 0);
	framebuffer = framebuffer2;
    }
    scrswitch ^= 1;
    return 0;
}

int kill_XDisplay();

int init_XDisplay()
{
    XGCValues xgcvals;
    unsigned long xgcmask;
    backend *p;
    int i;

    xdisp = XOpenDisplay( NULL );
    xscreen = DefaultScreen( xdisp );
    xvisual = DefaultVisual( xdisp, xscreen );
    xplanes = DefaultDepth( xdisp, xscreen );
    xgcvals.graphics_exposures = 0;
    xgcmask = GCGraphicsExposures;
    xgc = XCreateGC( xdisp, RootWindow(xdisp, xscreen), xgcmask, &xgcvals );
    xwin = XCreateSimpleWindow( xdisp, RootWindow(xdisp, xscreen), 0, 0,
				SCR_WIDTH, SCR_HEIGHT, 0, 0, 0);
    XMapWindow( xdisp, xwin );
    XFlush( xdisp );
    XSelectInput( xdisp, xwin, KeyPressMask | KeyReleaseMask );
    mit_shm = XShmQueryVersion( xdisp, &d1, &d2, &d3 );
    image=XShmCreateImage( xdisp,
			   xvisual,
			   xplanes,
			   ZPixmap,
			   NULL,
			   &shminfo,
			   SCR_WIDTH + 8,
			   SCR_HEIGHT );

    image2=XShmCreateImage( xdisp,
			   xvisual,
			   xplanes,
			   ZPixmap,
			   NULL,
			   &shminfo2,
			   SCR_WIDTH + 8,
			   SCR_HEIGHT );

    bytes_per_line = image->bytes_per_line;
    screen_height = image->height;
    shminfo.shmid = shmget(IPC_PRIVATE, SCR_SIZE, IPC_CREAT|0777);
    shminfo.shmaddr = image->data = shmat(shminfo.shmid,0,0);
    shminfo.readOnly = False;
    XShmAttach(xdisp,&shminfo);
    shminfo2.shmid = shmget(IPC_PRIVATE, SCR_SIZE, IPC_CREAT|0777);
    shminfo2.shmaddr = image2->data = shmat(shminfo2.shmid,0,0);
    shminfo2.readOnly = False;
    XShmAttach(xdisp,&shminfo2);
    printf("X attached...\n");
    framebuffer1 = (unsigned short *)image->data;
    framebuffer2 = (unsigned short *)image2->data;
    memset(framebuffer1, WhitePixel(xdisp, xscreen), SCR_SIZE);
    memset(framebuffer2, WhitePixel(xdisp, xscreen), SCR_SIZE);
    shmctl(shminfo.shmid, IPC_RMID, NULL);
    shmctl(shminfo2.shmid, IPC_RMID, NULL);    
    printf("Updating main image\n");
    updateimage();
    XSync(xdisp, False);

    framebuffer = framebuffer1;
    
    p = backends;
    
    while (p->class != 0xFF)
    {
	if ((p->class == xvisual->class) && (p->depth == xplanes))
	{
	    p->InitLookups();
	    Expandline = p->Expandline;
	    Clearline = p->Clearline;
	    Updatepal = p->Updatepal;
	    Decode_5bit = p->Decode_5bit;
	    printf("Expandline: %p\n", Expandline);
	    printf("Clearline: %p\n", Clearline);
	    printf("Decode_5bit: %p\n", Decode_5bit);
	    break;
	}
	p++;
    }
    
    return 0;
}

int kill_XDisplay()
{
    killimage();
    XDestroyWindow( xdisp, xwin );
    XFreeGC( xdisp, xgc );
    XCloseDisplay( xdisp );
    return 0;
}
